package com.huawei.dli.sdk;

import com.huawei.dli.sdk.common.Column;
import com.huawei.dli.sdk.common.DataType;
import com.huawei.dli.sdk.common.DataTypeInfo;
import com.huawei.dli.sdk.common.DecimalTypeInfo;
import com.huawei.dli.sdk.common.Row;
import com.huawei.dli.sdk.common.utils.DateTimeUtils;
import com.huawei.dli.sdk.common.utils.ObsClientUtils;
import com.huawei.dli.sdk.exception.DLIException;

import com.obs.services.model.ObsObject;

import org.apache.avro.Conversions;
import org.apache.avro.LogicalTypes;
import org.apache.avro.Schema;
import org.apache.avro.file.DataFileStream;
import org.apache.avro.generic.GenericDatumReader;
import org.apache.avro.generic.GenericFixed;
import org.apache.avro.generic.GenericRecord;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.nio.ByteBuffer;
import java.sql.Timestamp;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.stream.Collectors;

/**
 * Read avro file from OBS, the avro files are generated by Spark.
 * Currently only support primitive sql type:
 *   boolean, tinyint, smallint/short, int, bigint/long, float, double, decimal(p, s), string, date, timestamp
 */
public class AvroObsReader implements Reader {
    private static final Logger logger = LoggerFactory.getLogger(AvroObsReader.class);

    private static final int MAX_RECONNECT_NUM = 5;

    private final DownloadJob downloadJob;

    private final ObsClientUtils obsClientUtils;

    private final String ak;

    private final String sk;

    private String bucketName;

    private String objectPrefix;

    private final Conversions.DecimalConversion decimalConversion = new Conversions.DecimalConversion();

    private final Schema nullSchema = Schema.create(Schema.Type.NULL);

    private List<ObsObject> objects;

    private int totalIter = 0;

    private int nextIterIdx = 0;

    private Iterator<Object> iterator;

    private int readRecordsPerIter = 0;

    private boolean hasRenameCol = true;

    protected boolean closed = false;

    @Deprecated
    public AvroObsReader(
        DownloadJob downloadJob,
        String endpoint,
        String ak,
        String sk,
        String bucketName,
        String objectPrefix) {
        this.downloadJob = downloadJob;
        this.obsClientUtils = new ObsClientUtils(endpoint);
        this.ak = ak;
        this.sk = sk;
        this.bucketName = bucketName;
        this.objectPrefix = objectPrefix;
    }

    public AvroObsReader(
        DownloadJob downloadJob,
        String endpoint,
        String ak,
        String sk,
        List<ObsObject> objects) {
        this.downloadJob = downloadJob;
        this.obsClientUtils = new ObsClientUtils(endpoint);
        this.ak = ak;
        this.sk = sk;
        this.objects = objects;
    }

    @Override
    public void openReader() throws IOException {
        if (objects == null) {
            objects = obsClientUtils.listObjects(ak, sk, bucketName, objectPrefix);
        }
        totalIter = objects.size();
        logger.info("object size: {}", totalIter);

        iterator = getNextIterator();
    }

    private Iterator<Object> getNextIterator() {
        if (totalIter > 0 && nextIterIdx < totalIter) {
            readRecordsPerIter = 0;
            return createIterator(false);
        } else {
            return null;
        }
    }

    private Iterator<Object> createIterator(boolean recreate) {
        int currIterIdx = recreate ? nextIterIdx - 1 : nextIterIdx++;
        ObsObject object = objects.get(currIterIdx);
        logger.info("{} iterator for object: {}", (recreate ? "recreate" : "create"), object.getObjectKey());
        InputStream inputStream = obsClientUtils.getObjectStream(ak, sk, object.getBucketName(), object.getObjectKey());
        try {
            return new DataFileStream<>(inputStream, new GenericDatumReader<>()).iterator();
        } catch (IOException e) {
            logger.error("failed to create iterator", e);
            throw new IllegalStateException("Unexpected error as create iterator");
        }
    }

    @Override
    public Row read() throws DLIException, IOException {
        if (closed) {
            throw new IllegalStateException("The reader state is closed");
        }

        for (int times = 0; times < MAX_RECONNECT_NUM; times++) {
            try {
                if (!iterator.hasNext()) {
                    logger.info("read {} records from pre iterator", readRecordsPerIter);
                    if (nextIterIdx < totalIter) {
                        iterator = getNextIterator();
                    }
                }
                if (iterator != null && iterator.hasNext()) {
                    GenericRecord values = (GenericRecord) iterator.next();
                    Row row = downloadJob.newRow();
                    updateRow(values, row);
                    readRecordsPerIter++;
                    return row;
                } else {
                    return null;
                }
            } catch (Exception e) {
                logger.warn("get record failed, may be connection closed by obs, so reconnect it!!!", e);
                reconnectAndSkipHandled();
            }
        }
        throw new DLIException("Failed to read value with retry");
    }

    /**
     * A flag used to determine whether to read AVRO data based on the name of c$i.
     * Do not set this value unless necessary.
     *
     * @param hasRenameCol If rename the column name to c$i when write data.
     */
    public void setHasRenameCol(boolean hasRenameCol) {
        this.hasRenameCol = hasRenameCol;
    }

    private void updateRow(GenericRecord record, Row row) throws DLIException {
        List<Column> columns = row.getSchema().getColumns();
        List<Schema.Field> avroFields = record.getSchema().getFields();
        for (int i = 0; i < columns.size(); i++) {
            Column column = columns.get(i);
            String colNameInFile = column.getColumnName();
            if (hasRenameCol) {
                colNameInFile = "c" + i; // As avro names should start with [A-Za-z_], so we rename it to c$i.
            }
            Object data = null;
            if (record.hasField(colNameInFile)) {
                data = record.get(colNameInFile);
            } else {
                logger.error("missing data for column: {}", column.getColumnName());
            }
            Schema.Field avroField = avroFields.get(i);
            updateCol(row, i, column.getType(), column.getDataTypeInfo(), data, avroField);
        }
    }

    private void updateCol(Row row, int i, DataType dataType, DataTypeInfo dataTypeInfo, Object data,
        Schema.Field avroField) throws DLIException {
        switch (dataType) {
            case BOOLEAN:
                row.setBoolean(i, (data == null) ? null : (boolean) data);
                break;
            case TINYINT:
                row.setTinyint(i, (data == null) ? null : Byte.parseByte(String.valueOf(data)));
                break;
            case SMALLINT:
            case SHORT:
                row.setShort(i, (data == null) ? null : Short.parseShort(String.valueOf(data)));
                break;
            case INT:
                row.setInt(i, (data == null) ? null : (int) data);
                break;
            case LONG:
                row.setLong(i, (data == null) ? null : (long) data);
                break;
            case BIGINT:
                row.setBigInteger(i, (data == null) ? null : new BigInteger(String.valueOf(data)));
                break;
            case FLOAT:
                row.setFloat(i, (data == null) ? null : (float) data);
                break;
            case DOUBLE:
                row.setDouble(i, (data == null) ? null : (double) data);
                break;
            case DECIMAL:
                DecimalTypeInfo decimalTypeInfo = (DecimalTypeInfo) dataTypeInfo;
                updateDecimalCol(row, i, data, decimalTypeInfo, avroField);
                break;
            case STRING:
                row.setString(i, (data == null) ? null : String.valueOf(data));
                break;
            case DATE:
                updateDateCol(row, i, data, avroField);
                break;
            case TIMESTAMP:
                updateTimestampCol(row, i, data, avroField);
                break;
            default:
                throw new IllegalArgumentException("Unsupported data type: " + dataType);
        }
    }

    private void updateDecimalCol(Row row, int i, Object data, DecimalTypeInfo d, Schema.Field avroField)
        throws DLIException {
        // On our datasource dli-avro use string type for write decimal data on avro,
        // use Object.toString() to get the value.
        // But on spark-avro from Spark 2.4 use avro fixed type for write decimal date on avro.
        if (data == null) {
            row.setBigDecimal(i, null);
        } else {
            Schema fieldTypeSchema = getFieldTypeSchema(avroField, "decimal");
            Schema fieldFullSchema = avroField.schema(); // a UNION type
            if (fieldTypeSchema.getType() == Schema.Type.STRING) {
                row.setBigDecimal(i, new BigDecimal(String.valueOf(data)));
            } else if (fieldTypeSchema.getType() == Schema.Type.FIXED) {
                row.setBigDecimal(i, decimalConversion.fromFixed((GenericFixed) data, fieldFullSchema,
                    LogicalTypes.decimal(d.getPrecision(), d.getScale())));
            } else if (fieldTypeSchema.getType() == Schema.Type.BYTES) {
                row.setBigDecimal(i, decimalConversion.fromBytes((ByteBuffer) data, fieldFullSchema,
                    LogicalTypes.decimal(d.getPrecision(), d.getScale())));
            } else {
                throw new IllegalArgumentException("Unexpected avro type: " + fieldTypeSchema.getName()
                    + " for sql decimal type.");
            }
        }
    }

    private void updateDateCol(Row row, int i, Object data, Schema.Field avroField) throws DLIException {
        // On our datasource dli-avro use long type for write date data on avro,
        // use Date.getTime to get the value.
        // But on spark-avro from Spark 2.4 use int type for write date data on avro.
        if (data == null) {
            row.setDate(i, null);
        } else {
            Schema fieldTypeSchema = getFieldTypeSchema(avroField, "date");
            if (fieldTypeSchema.getType() == Schema.Type.LONG) {
                row.setDate(i, new Date((long) data));
            } else if (fieldTypeSchema.getType() == Schema.Type.INT) {
                row.setDate(i, DateTimeUtils.toJavaDate((int) data));
            } else {
                throw new IllegalArgumentException("Unexpected avro type: " + fieldTypeSchema.getName()
                    + " for sql date type.");
            }
        }
    }

    private void updateTimestampCol(Row row, int i, Object data, Schema.Field avroField) throws DLIException {
        // On our datasource dli-avro use long type for write timestamp data on avro,
        // use Timestamp.getTime to get the value.
        // But on spark-avro from Spark 2.4 use long type with different avro logical type
        // for write timestamp data on avro.
        if (data == null) {
            row.setTimestamp(i, null);
        } else {
            Schema fieldTypeSchema = getFieldTypeSchema(avroField, "timestamp");
            if (fieldTypeSchema.getLogicalType() == null
                || fieldTypeSchema.getLogicalType() instanceof LogicalTypes.TimestampMillis) {
                row.setTimestamp(i, new Timestamp((long) data));
            } else if (fieldTypeSchema.getLogicalType() instanceof LogicalTypes.TimestampMicros) {
                row.setTimestamp(i, DateTimeUtils.toJavaTimestamp((long) data));
            } else {
                throw new IllegalArgumentException("Unexpected avro logical type: "
                    + fieldTypeSchema.getLogicalType().getName() + " for sql timestamp type.");
            }
        }
    }

    private Schema getFieldTypeSchema(Schema.Field avroField, String sqlType) {
        // Schema.getType is UNION type, so need to getTypes and filter null type.
        List<Schema> fieldTypeSchema = avroField.schema().getTypes().stream()
            .filter(schema -> !schema.equals(nullSchema))
            .collect(Collectors.toList());
        if (fieldTypeSchema.size() != 1) {
            throw new IllegalArgumentException("Unexpected avro type: "
                + fieldTypeSchema.stream().map(Schema::getName).collect(Collectors.joining())
                + " for sql " + sqlType + " type.");
        }
        return fieldTypeSchema.get(0);
    }

    private void reconnectAndSkipHandled() {
        iterator = createIterator(true);
        logger.warn("reconnect, and skip records: {}", readRecordsPerIter);
        // skip all records have read
        long startTime = System.currentTimeMillis();
        for (int i = 0; i < readRecordsPerIter; i++) {
            if (iterator.hasNext()) {
                iterator.next();
            } else {
                throw new IllegalStateException(
                    String.format("Unexpected error, pre read records: %d, but now only read records: %d",
                        readRecordsPerIter, (i + 1)));
            }
        }
        long endTime = System.currentTimeMillis();
        logger.info("skip records cost: {}ms", (endTime - startTime));
    }

    @Override
    public void close() throws IOException {
        if (closed) {
            throw new IllegalStateException("The reader state is closed");
        }
        logger.info("read {} records from pre iterator", readRecordsPerIter);
        obsClientUtils.closeObsClient();
    }

    @Override
    public boolean isClosed() {
        return closed;
    }

    @Override
    public boolean hasNext() {
        for (int times = 0; times < MAX_RECONNECT_NUM; times++) {
            try {
                return (iterator != null && iterator.hasNext()) || (nextIterIdx < totalIter);
            } catch (Exception e) {
                logger.warn("get record failed, may be connection closed by obs, so reconnect it!!!", e);
                reconnectAndSkipHandled();
            }
        }
        throw new IllegalStateException("Should not reach this");
    }
}
