package com.huawei.dli.sdk;

import com.huawei.dli.sdk.common.Column;
import com.huawei.dli.sdk.common.DataType;
import com.huawei.dli.sdk.common.DecimalTypeInfo;
import com.huawei.dli.sdk.common.Row;
import com.huawei.dli.sdk.common.utils.DateTimeUtils;
import com.huawei.dli.sdk.common.utils.ObsClientUtils;
import com.huawei.dli.sdk.exception.DLIException;

import com.obs.services.ObsClient;

import org.apache.avro.Conversions;
import org.apache.avro.LogicalTypes;
import org.apache.avro.Schema;
import org.apache.avro.SchemaBuilder;
import org.apache.avro.file.DataFileWriter;
import org.apache.avro.generic.GenericData;
import org.apache.avro.generic.GenericDatumWriter;
import org.apache.avro.generic.GenericRecord;
import org.apache.avro.io.DatumWriter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.math.RoundingMode;
import java.sql.Timestamp;
import java.util.Date;
import java.util.List;
import java.util.UUID;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Executor;
import java.util.concurrent.Executors;
import java.util.stream.Collectors;

/**
 * On dli-avro, TINYINT/SMALLINT/INT is int, DECIMAL is string, DATE/TIMESTAMP is long when write data,
 * but don't support tinyint/smallint when read data, so don't use it.
 * If we upgrade spark to 2.4 or fix this bug, then replace ObsWriter with this.
 *
 * On spark-avro from Spark 2.4 TINYINT/SMALLINT/INT is int, DECIMAL is fixed,
 * DATE is int with logical type, TIMESTAMP is long with logical type.
 */
public class AvroObsWriter extends Writer {
    private static final Logger logger = LoggerFactory.getLogger(AvroObsWriter.class);

    private static final long MAX_SIZE_PER_FILE = 134_217_728; // 128MB

    private static final Executor EXECUTOR = Executors.newCachedThreadPool();

    private static final int ONE_BYTE = 1;

    private static final int FOUR_BYTE = 4;

    private static final int EIGHT_BYTE = 8;

    private final UploadJob uploadJob;

    private final ObsClientUtils obsClientUtils;

    private final String bucketName;

    private final String objectPrefix;

    private final Conversions.DecimalConversion decimalConversion = new Conversions.DecimalConversion();

    private final Schema nullSchema = Schema.create(Schema.Type.NULL);

    private Schema schema;

    private DataFileWriter<GenericRecord> dataFileWriter;

    private OutputStream outputStream;

    private CountDownLatch cdl;

    private long writtenBytes = 0;

    protected boolean closed = false;

    public AvroObsWriter(
        UploadJob uploadJob,
        String endpoint,
        String ak,
        String sk,
        String bucketName,
        String objectPrefix) {
        this.uploadJob = uploadJob;
        this.obsClientUtils = new ObsClientUtils(endpoint, ak, sk);
        this.bucketName = bucketName;
        this.objectPrefix = objectPrefix;
        init(uploadJob.tblSchema.getColumns());
    }

    private void init(List<Column> columns) {
        if (schema == null) {
            SchemaBuilder.FieldAssembler<Schema> assembler = SchemaBuilder.record("topLevelRecord").fields();
            for (Column column : columns) {
                // for static partition column don't write data
                if (uploadJob.staticPartCols.contains(column.getColumnName())) {
                    continue;
                }
                SchemaBuilder.TypeBuilder<Schema> builder = SchemaBuilder.builder();
                switch (column.getType()) {
                    case BOOLEAN:
                        assembler.name(column.getColumnName())
                            .type(Schema.createUnion(builder.booleanType(), nullSchema))
                            .noDefault();
                        break;
                    case TINYINT:
                    case SMALLINT:
                    case SHORT:
                    case INT:
                        assembler.name(column.getColumnName())
                            .type(Schema.createUnion(builder.intType(), nullSchema))
                            .noDefault();
                        break;
                    case LONG:
                    case BIGINT:
                        assembler.name(column.getColumnName())
                            .type(Schema.createUnion(builder.longType(), nullSchema))
                            .noDefault();
                        break;
                    case FLOAT:
                        assembler.name(column.getColumnName())
                            .type(Schema.createUnion(builder.floatType(), nullSchema))
                            .noDefault();
                        break;
                    case DOUBLE:
                        assembler.name(column.getColumnName())
                            .type(Schema.createUnion(builder.doubleType(), nullSchema))
                            .noDefault();
                        break;
                    case DECIMAL:
                        DecimalTypeInfo d = (DecimalTypeInfo) column.getDataTypeInfo();
                        int fixedSize = computeMinBytesForPrecision(d.getPrecision());
                        assembler.name(column.getColumnName())
                            .type(Schema.createUnion(
                                LogicalTypes.decimal(d.getPrecision(), d.getScale())
                                    .addToSchema(SchemaBuilder.fixed(column.getColumnName()).size(fixedSize)),
                                nullSchema))
                            .noDefault();
                        break;
                    case STRING:
                        assembler.name(column.getColumnName())
                            .type(Schema.createUnion(builder.stringType(), nullSchema))
                            .noDefault();
                        break;
                    case DATE:
                        assembler.name(column.getColumnName())
                            .type(Schema.createUnion(
                                LogicalTypes.date().addToSchema(builder.intType()),
                                nullSchema))
                            .noDefault();
                        break;
                    case TIMESTAMP:
                        assembler.name(column.getColumnName())
                            .type(Schema.createUnion(
                                LogicalTypes.timestampMicros().addToSchema(builder.longType()),
                                nullSchema))
                            .noDefault();
                        break;
                    default:
                        throw new IllegalArgumentException("Unsupported data type: " + column.getType());
                }
            }
            schema = assembler.endRecord();
            DatumWriter<GenericRecord> datumWriter = new GenericDatumWriter<>(schema);
            dataFileWriter = new DataFileWriter<>(datumWriter);
        }
    }

    private int computeMinBytesForPrecision(int precision) {
        int numBytes = 1;
        while (Math.pow(2.0, 8 * numBytes - 1) < Math.pow(10.0, precision)) {
            numBytes += 1;
        }
        return numBytes;
    }

    @Override
    public synchronized void write(Row row) throws DLIException {
        if (closed) {
            throw new IllegalStateException("The writer state is closed");
        }

        if (outputStream == null) {
            try {
                logger.info("init piped input and output stream");
                PipedInputStream inputStream = new PipedInputStream();
                outputStream = new PipedOutputStream(inputStream);
                dataFileWriter.create(schema, outputStream);
                cdl = new CountDownLatch(1);
                EXECUTOR.execute(writeData(inputStream, cdl));
            } catch (IOException e) {
                logger.error("failed to create output stream for writer", e);
                throw new DLIException("Unexpected error while write data");
            }
        }

        GenericRecord record = new GenericData.Record(schema);
        List<Column> columns = row.getSchema().getColumns();
        List<Schema.Field> avroFields = record.getSchema().getFields();
        for (int i = 0; i < columns.size(); i++) {
            String colName = columns.get(i).getColumnName();
            // for static partition column don't write data
            if (uploadJob.staticPartCols.contains(colName)) {
                continue;
            }
            DataType dataType = columns.get(i).getType();
            Schema.Field avroField = avroFields.get(i);
            switch (dataType) {
                case BOOLEAN:
                    updateRecord(record, colName, row.getBoolean(i), ONE_BYTE);
                    break;
                case TINYINT:
                    Byte bValue = row.getTinyint(i);
                    Integer b2IntValue = bValue == null ? null : Integer.valueOf(bValue);
                    updateRecord(record, colName, b2IntValue, FOUR_BYTE);
                    break;
                case SMALLINT:
                case SHORT:
                    Short sValue = row.getShort(i);
                    Integer s2IntValue = sValue == null ? null : Integer.valueOf(sValue);
                    updateRecord(record, colName, s2IntValue, FOUR_BYTE);
                    break;
                case INT:
                    updateRecord(record, colName, row.getInt(i), FOUR_BYTE);
                    break;
                case LONG:
                    updateRecord(record, colName, row.getLong(i), EIGHT_BYTE);
                    break;
                case BIGINT:
                    BigInteger bIntValue = row.getBigInteger(i);
                    Long bigIntLongValue = bIntValue == null ? null : bIntValue.longValue();
                    updateRecord(record, colName, bigIntLongValue, EIGHT_BYTE);
                    break;
                case FLOAT:
                    updateRecord(record, colName, row.getFloat(i), FOUR_BYTE);
                    break;
                case DOUBLE:
                    updateRecord(record, colName, row.getDouble(i), EIGHT_BYTE);
                    break;
                case DECIMAL:
                    BigDecimal bDecValue = row.getBigDecimal(i);
                    if (bDecValue == null) {
                        updateRecord(record, colName, null, 0);
                    } else {
                        DecimalTypeInfo decTypeInfo = (DecimalTypeInfo) columns.get(i).getDataTypeInfo();
                        BigDecimal bDecValueUpdated = bDecValue.setScale(decTypeInfo.getScale(), RoundingMode.HALF_UP);
                        List<Schema> fieldSchema = avroField.schema().getTypes().stream()
                            .filter(schema -> !schema.equals(nullSchema))
                            .collect(Collectors.toList());
                        if (fieldSchema.size() != 1) {
                            throw new IllegalArgumentException("Unexpected avro field type: "
                                + fieldSchema.stream().map(Schema::getName).collect(Collectors.joining())
                                + " for sql decimal type.");
                        }
                        updateRecord(
                            record,
                            colName,
                            decimalConversion.toFixed(
                                bDecValueUpdated, fieldSchema.get(0), fieldSchema.get(0).getLogicalType()),
                            fieldSchema.get(0).getFixedSize());
                    }
                    break;
                case STRING:
                    String strValue = row.getString(i);
                    updateRecord(record, colName, strValue, strValue == null ? 0 : strValue.getBytes().length);
                    break;
                case DATE:
                    Date dateValue = row.getDate(i);
                    Integer dateIntValue = dateValue == null ? null : DateTimeUtils.fromJavaDate(dateValue);
                    updateRecord(record, colName, dateIntValue, FOUR_BYTE);
                    break;
                case TIMESTAMP:
                    Timestamp tsValue = row.getTimestamp(i);
                    Long tsLongValue = tsValue == null ? null : DateTimeUtils.fromJavaTimestamp(tsValue);
                    updateRecord(record, colName, tsLongValue, EIGHT_BYTE);
                    break;
                default:
                    throw new IllegalArgumentException("Unsupported data type: " + dataType.getName());
            }
        }

        try {
            dataFileWriter.append(record);
            if (writtenBytes > MAX_SIZE_PER_FILE) {
                logger.info("current file reach expected size, change another one");
                dataFileWriter.flush();
                dataFileWriter.close();
                cdl.await(); // Wait read thread finish it's work, or else will cause NPE.
                outputStream = null;
                writtenBytes = 0;
            }
        } catch (IOException e) {
            logger.error("failed to write data", e);
            throw new DLIException("Unexpected error while write data");
        } catch (InterruptedException e) {
            logger.error("thread interrupted while write data", e);
            throw new DLIException("Unexpected error while write data");
        }
    }

    private void updateRecord(GenericRecord record, String key, Object value, int len) {
        if (value == null) {
            record.put(key, null);
        } else {
            record.put(key, value);
            writtenBytes += len;
        }
    }

    private Runnable writeData(InputStream inputStream, CountDownLatch cdl) {
        return () -> {
            ObsClient obsClient = obsClientUtils.getObsClient();
            String objKey = objectPrefix + String.format("/part-%s.avro", UUID.randomUUID().toString());
            logger.info("init writer on object: {}", objKey);
            obsClient.putObject(bucketName, objKey, inputStream);
            try {
                logger.info("all data have witten to obs, close the input stream");
                inputStream.close();
            } catch (IOException e) {
                logger.error("failed to close input stream", e);
            } catch (Exception e) {
                logger.error("unexpected error", e);
            } finally {
                cdl.countDown();
            }
        };
    }

    @Override
    public synchronized void flush() throws DLIException {
        if (closed) {
            throw new IllegalStateException("The writer state is closed");
        }

        try {
            dataFileWriter.flush();
        } catch (IOException e) {
            logger.error("failed to flush data", e);
            throw new DLIException("Unexpected error while flush");
        }
    }

    @Override
    public synchronized void close() throws DLIException {
        if (closed) {
            throw new IllegalStateException("The writer state is closed");
        }

        try {
            logger.info("all data have been written, close the writer");
            flush();
            dataFileWriter.close();
            cdl.await(); // Wait read thread finish it's work, or else will cause NPE.
            outputStream = null;
            writtenBytes = 0;
        } catch (IOException e) {
            logger.error("failed to close stream for writer", e);
            throw new DLIException("Unexpected error while close");
        } catch (InterruptedException e) {
            logger.error("thread interrupted while wait close", e);
            throw new DLIException("Unexpected error while close");
        }
        obsClientUtils.closeObsClient();
        closed = true;
    }

    @Override
    public boolean isClosed() {
        return closed;
    }
}
