package com.huawei.dli.sdk;

import com.huawei.dli.restapi.ApiException;
import com.huawei.dli.restapi.ApiResponse;
import com.huawei.dli.restapi.model.GlobalErrorResponse;
import com.huawei.dli.sdk.common.IHttpExecAdpter;
import com.huawei.dli.sdk.exception.DLIException;

import com.google.gson.Gson;
import com.google.gson.JsonSyntaxException;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.List;
import java.util.function.Function;

public abstract class BaseJob {
    private static final Logger LOGGER = LoggerFactory.getLogger(BaseJob.class);
    public static final String JOB_TRANSACTION_ERROR_CODE = "dli.0202";

    private String transactionId;
    protected DLIClient client;


    public BaseJob(DLIClient client) {
        this.client = client;
    }

    public BaseJob(DLIClient client, String transactionId) {
        this.client = client;
        this.transactionId = transactionId;
    }

    protected abstract String getTransactionKeyName();

    public <T, V> T executeWithRetry(
        IHttpExecAdpter<T> exec,
        Class<T> resDtoClass,
        Function<T, GlobalErrorResponse> errorFetcher,
        Function<T, String> transactionConflictManage) throws DLIException, InterruptedException {
        int retryTimes = this.client.getDLIInfo().getRetryTimes();
        boolean needRetry = this.client.getDLIInfo().isUseRetry();
        retryTimes = needRetry ? retryTimes : 0;
        // retry mode, the times should more than 0
        if (needRetry && retryTimes <= 0) {
            LOGGER.error("mode retry times should more than 0 in retry mode.");
            throw new DLIException("mode retry times should more than 0 in retry mode.");
        }

        for (int i = 0; i <= retryTimes; ++i) {
            try {
                ApiResponse<T> res = executeCall(exec, resDtoClass, errorFetcher, transactionConflictManage);
                if (res.getStatusCode() >= 200 && res.getStatusCode() < 300) {
                    return res.getData();
                }
                GlobalErrorResponse error = errorFetcher.apply(res.getData());
                apiExceptionRetryManage(
                    res.getStatusCode(),
                    error.getErrorCode(),
                    error.getErrorMsg(),
                    i,
                    false);
            } catch (ApiException e) {
                apiExceptionRetryManage(e.getCode(), null, e.getMessage(), i, true);
            }
        }
        return null;
    }

    protected <T> ApiResponse<T> executeCall(
        IHttpExecAdpter<T> exec,
        Class<T> resDtoClass,
        Function<T, GlobalErrorResponse> errorFetcher,
        Function<T, String> transactionConflictManage) throws DLIException, ApiException {
        try {
            return exec.exec();
        } catch (ApiException e) {
            if (StringUtils.containsIgnoreCase(e.getResponseBody(), "dli")
                || StringUtils.containsIgnoreCase(e.getResponseBody(), "apig")) {
                try {
                    T resDto = new Gson().fromJson(e.getResponseBody(), resDtoClass);
                    GlobalErrorResponse errorDto = resDto != null ? errorFetcher.apply(resDto) : null;
                    if (errorDto != null && transactionConflictManage != null
                        && StringUtils.equalsAnyIgnoreCase(errorDto.getErrorCode(), Job.JOB_TRANSACTION_ERROR_CODE)) {
                        LOGGER.warn("Spark already submitted with same transactionId, errorMsg "
                            + errorDto.getErrorMsg());
                        String conflictTip = transactionConflictManage.apply(resDto);
                        if (StringUtils.isNotBlank(conflictTip)) {
                            throw new DLIException(conflictTip);
                        }
                        return new ApiResponse<T>(200, null, resDto);
                    }
                    if (errorDto != null
                        && (StringUtils.containsIgnoreCase(errorDto.getErrorCode(), "dli.")
                        || StringUtils.containsIgnoreCase(errorDto.getErrorCode(), "apigw."))) {
                        LOGGER.error("Execute spark job failed: ", e);
                        return new ApiResponse<T>(e.getCode(), null, resDto);
                    }
                } catch (JsonSyntaxException ex) {
                    LOGGER.warn("Request response not json: " + e.getMessage());
                }
            }
            LOGGER.error("Execute spark job failed: ", e);
            throw e;
        }
    }

    protected void apiExceptionRetryManage(
        int statusCode,
        String serviceCode,
        String errorMsg,
        int currentTimes,
        boolean mustRetry) throws DLIException, InterruptedException {
        if (!client.getDLIInfo().isUseRetry()) {
            LOGGER.info("Job not need to retry, return error");
            throw new DLIException(serviceCode + ": " + errorMsg);
        }
        int retryTimes = this.client.getDLIInfo().getRetryTimes();
        LOGGER.warn("Submit spark job failed with httpStatus: {}, errorCode: {} , responseMsg: {}, retryTimes: {}",
            statusCode, serviceCode, errorMsg, retryTimes);
        // call rest api the service return error, maybe API-GW or DLI return error
        if (!mustRetry && !isServiceInsideError(statusCode, serviceCode)) {
            throw new DLIException(serviceCode + ": " + errorMsg);
        }
        if (currentTimes < retryTimes) {
            // this is not the last times, sleep and retry
            Thread.sleep(this.client.getDLIInfo().getRetryIntervals() * 1000);
        } else {
            throw new DLIException(String.format(
                "Submit spark job failed with httpStatus: %s, error: %s", statusCode, serviceCode + ": " + errorMsg));
        }
    }

    protected boolean isServiceInsideError(int statusCode, String serviceCode) {
        if (statusCode >= 200 && statusCode < 300) {
            return false;
        }
        if (statusCode / 100 == 5) {
            return true;
        }

        List<String> errorKeyWords = Arrays.asList("\"error_code\":\"dli.0999\"", "\"error_code\":\"apigw.0201\"",
            "\"error_code\":\"apigw.0202\"", "\"error_code\":\"apigw.0203\"", "\"error_code\":\"apigw.0401\"",
            "\"error_code\":\"apigw.0402\"", "\"error_code\":\"apigw.0404\"");
        for (String errorKeyWord : errorKeyWords) {
            if (StringUtils.containsIgnoreCase(errorKeyWord, serviceCode)) {
                return true;
            }
        }
        return false;
    }

    public String getTransactionId() {
        return transactionId;
    }

    public void setTransactionId(String transactionId) {
        this.transactionId = transactionId;
    }
}
