package com.huawei.dli.sdk;

import com.huawei.dli.restapi.ApiException;
import com.huawei.dli.restapi.ApiResponse;
import com.huawei.dli.restapi.model.GlobalErrorResponse;
import com.huawei.dli.restapi.model.MessageInfo;
import com.huawei.dli.restapi.model.SessionInfo;
import com.huawei.dli.restapi.model.SessionLog;
import com.huawei.dli.restapi.model.SessionState;
import com.huawei.dli.sdk.common.IHttpExecAdpter;
import com.huawei.dli.sdk.common.SparkJobInfo;
import com.huawei.dli.sdk.common.SparkJobStatus;
import com.huawei.dli.sdk.exception.DLIException;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.UUID;

public class BatchJob extends SparkJob {
    private SparkJobInfo jobInfo;

    private static final Logger LOGGER = LoggerFactory.getLogger(BatchJob.class);

    public BatchJob(Cluster cluster, SparkJobInfo jobInfo) {
        this(cluster, jobInfo, null);
    }

    public BatchJob(Cluster cluster, SparkJobInfo jobInfo, String transactionId) {
        super(cluster, transactionId);
        this.jobInfo = jobInfo;
    }

    @Override
    public void submit() throws DLIException {
        asyncSubmit();
        super.checkJobRunningState();
        LOGGER.info("Job submit synchronous, jobId: " + this.getJobId());
    }

    @Override
    public void asyncSubmit() throws DLIException {
        if (this.getJobId() != null) {
            LOGGER.error("Job had been submitted, don't submit repeatedly");
            throw new DLIException("Job had been submitted, don't submit repeatedly");
        }
        if (!getCluster().getClusterName().equals(jobInfo.getClusterName())) {
            LOGGER.error("Input 2 cluster name, please check");
            throw new DLIException("Input 2 cluster name, please check");
        }
        try {
            SessionInfo resp = submitWithRetry();
            this.setJobId(resp.getId());
            LOGGER.info("Job submit asynchronous, jobId: " + resp.getId());
        } catch (InterruptedException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e.getMessage());
        }
    }

    @Override
    public SparkJobStatus getStatus() throws DLIException {
        try {
            SessionState state = client.getRestApi().getBatchState(getProjectId(), getJobId(), client.getAuthToken());
            return convertSessionState(state.getState());
        } catch (ApiException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e.getMessage(), e.getCode(), e.getResponseBody());
        }
    }

    @Override
    public List<String> getLog(int from, int size) throws DLIException {
        return getBatchLog(from, size, null, 0);
    }

    @Override
    public List<String> getLogByJobId(String jobId, int from, int size) throws DLIException {
        this.setJobId(jobId);
        return getBatchLog(from, size, null, 0);
    }

    @Override
    public List<String> getDriverLogByJobId(String jobId, int from, int size, int index) throws DLIException {
        this.setJobId(jobId);
        return getBatchLog(from, size, "driver", index);
    }

    @Override
    public List<String> getDriverLog(int from, int size, int index) throws DLIException {
        return getBatchLog(from, size, "driver", index);
    }

    @Override
    protected String getTransactionKeyName() {
        return "spark.dli.job.transaction.id";
    }

    private SessionInfo submitWithRetry() throws DLIException, InterruptedException {
        String transactionId =
            StringUtils.isNotBlank(getTransactionId()) ? getTransactionId() : UUID.randomUUID().toString();
        if (jobInfo.getConf() == null || !jobInfo.getConf().containsKey(getTransactionKeyName())) {
            jobInfo.addConfItem(getTransactionKeyName(), transactionId);
        }
        Object reqJson = jobInfo.toJsonObj();

        return executeWithRetry(
            new IHttpExecAdpter<SessionInfo>() {
                @Override
                public ApiResponse<SessionInfo> exec() throws DLIException, ApiException {
                    return client.getRestApi().createBatchWithHttpInfo(
                        getProjectId(),
                        reqJson, client.getAuthToken(),
                        null);
                }
            },
            SessionInfo.class,
            (SessionInfo res) ->
                new GlobalErrorResponse().errorCode(res.getErrorCode()).errorMsg(res.getErrorMsg()),
            (SessionInfo dto) -> {
                dto.setErrorCode(null);
                dto.setErrorMsg(null);
                return null;
            }
        );
    }

    private List<String> getBatchLog(int from, int size, String type, int index) throws DLIException {
        if (this.getStatus().equals(SparkJobStatus.STARTING) || this.getStatus().equals(SparkJobStatus.RUNNING)) {
            LOGGER.error("Job is running, can't get log");
            throw new DLIException("Job is running, can't get log");
        }
        try {
            SessionLog sessionLog = client.getRestApi()
                .getBatchLog(getProjectId(), this.getJobId(), client.getAuthToken(),
                    from, size, type, index);
            return sessionLog.getLog();
        } catch (ApiException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e.getMessage(), e.getCode(), e.getResponseBody());
        }
    }

    /**
     * delete batchJob (recommend)
     *
     * @return MessageInfo
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public MessageInfo delBatch() throws DLIException {
        String batchId = this.getJobId();
        return delBatch(batchId);
    }

    /**
     * delete batchJob by batchId (deprecated)
     *
     * @param batchId batch id
     * @return MessageInfo
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public MessageInfo delBatch(String batchId) throws DLIException {
        try {
            if (batchId == null || batchId.equals("")) {
                LOGGER.error("No batchJob in progress, delete failed");
                throw new DLIException("No batchJob in progress, delete failed");
            }
            MessageInfo msg = new MessageInfo();
            SparkJobStatus status = this.getStatus();
            if (status.equals(SparkJobStatus.ERROR) || status.equals(SparkJobStatus.DEAD)
                || status.equals(SparkJobStatus.SUCCESS)) {
                LOGGER.info("batchJob's SparkJobStatus can't be deleted");
                msg.setMsg("batchJob's SparkJobStatus can't be deleted");
                return msg;
            }
            LOGGER.info("batchJob delete succeed, batchId: " + batchId);
            return client.getRestApi().delBatch(getProjectId(), batchId, client.getAuthToken());
        } catch (ApiException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e.getMessage(), e.getCode(), e.getResponseBody());
        }
    }
}
