package com.huawei.dli.sdk;

import com.huawei.dli.restapi.ApiException;
import com.huawei.dli.restapi.model.ColumnInfo;
import com.huawei.dli.restapi.model.ColumnResult;
import com.huawei.dli.restapi.model.ResponseInfo;
import com.huawei.dli.restapi.model.TableInfo;
import com.huawei.dli.restapi.model.TableListInfo;
import com.huawei.dli.restapi.model.TableListResult;
import com.huawei.dli.sdk.common.Column;
import com.huawei.dli.sdk.common.CsvFormatInfo;
import com.huawei.dli.sdk.common.DataTypeInfo;
import com.huawei.dli.sdk.common.StorageType;
import com.huawei.dli.sdk.common.TableSchema;
import com.huawei.dli.sdk.common.TableType;
import com.huawei.dli.sdk.common.type.TypeInfoUtils;
import com.huawei.dli.sdk.exception.DLIException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;

public class Database {

    private static final int MAX_PAGE_SIZE = 100;

    private static final Logger LOGGER = LoggerFactory.getLogger(Database.class);

    private String databaseName;

    private String description = "";

    private String owner;

    private int tableCount;

    private DLIClient client;

    private String enterpriseProjectId;

    private String resourceId;

    public Database(String databaseName, String description, String owner, int tableCount, DLIClient client) {
        this.databaseName = databaseName;
        this.description = description;
        this.owner = owner;
        this.tableCount = tableCount;
        this.client = client;
    }

    public String getDatabaseName() {
        return databaseName;
    }

    public String getDescription() {
        return description;
    }

    public String getOwner() {
        return owner;
    }

    public int getTableCount() {
        return tableCount;
    }

    public String getProjectId() {
        return client.getAuthentication().getProjectId();
    }

    public String getEnterpriseProjectId() {
        return enterpriseProjectId;
    }

    public void setEnterpriseProjectId(String enterpriseProjectId) {
        this.enterpriseProjectId = enterpriseProjectId;
    }

    public String getResourceId() {
        return resourceId;
    }

    public void setResourceId(String resourceId) {
        this.resourceId = resourceId;
    }

    /**
     * Delete the current database.
     * If there is a table in the database to be deleted,
     * all the tables in it must to be deleted first.
     *
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public void deleteDatabase() throws DLIException {
        if (databaseName == null) {
            LOGGER.error("Database name can not be null.");
            throw new DLIException("Database name can not be null.");
        }
        try {
            ResponseInfo response = client.getRestApi()
                .deleteDatabase(getProjectId(), databaseName, client.getAuthToken());
            if (!response.getIsSuccess()) {
                LOGGER.error(response.getMessage());
                throw new DLIException(response.getMessage());
            }
        } catch (ApiException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e.getMessage(), e.getCode(), e.getResponseBody());
        }
    }

    /**
     * List all tables in the current database.
     *
     * @return List tables
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public List<Table> listAllTables() throws DLIException {
        return this.listAllTables(null, false, null);
    }

    /**
     * List all tables in the current database.
     * Filter by keyword.
     *
     * @param keyword The keyword of the filter table name.
     * @return List tables
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public List<Table> listAllTables(String keyword) throws DLIException {
        return this.listAllTables(keyword, true, null);
    }

    /**
     * List all tables in the current database.
     *
     * @param withDetail Whether to get the details of the table (owner, size, etc.).
     * @return List tables
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public List<Table> listAllTables(boolean withDetail) throws DLIException {
        return this.listAllTables(null, withDetail, null);
    }

    /**
     * List all tables in the current database.
     * Filter by table type.
     *
     * @param type Filter by table type.(EXTERNAL,MANAGED,VIEW)
     * @return List tables
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public List<Table> listAllTables(TableType type) throws DLIException {
        return this.listAllTables(null, true, type);
    }

    /**
     * List all tables in the current database.
     * Filter by keyword.
     *
     * @param keyword    The keyword of the filter table name.
     * @param withDetail Whether to get the details of the table (owner, size, etc.).
     * @return List tables
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public List<Table> listAllTables(String keyword, boolean withDetail) throws DLIException {
        return this.listAllTables(keyword, withDetail, null);
    }

    /**
     * List all tables in the current database.
     *
     * @param keyword    The keyword of the filter table name.
     * @param withDetail Whether to get the details of the table (owner, size, etc.).
     * @param type       Filter by table type.(EXTERNAL,MANAGED,VIEW)
     * @return list of tables
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public List<Table> listAllTables(String keyword, boolean withDetail, TableType type) throws DLIException {
        List<Table> tables = new ArrayList<Table>();
        for (int i = 1; ; i++) {
            List<Table> tableList = listAllTables(keyword, withDetail, MAX_PAGE_SIZE, i, type);
            tables.addAll(tableList);
            if (tableList.size() < MAX_PAGE_SIZE) {
                break;
            }
        }
        return tables;
    }

    /**
     * List tables by page in the current database.
     *
     * @param keyword     The keyword of the filter table name.
     * @param withDetail  Whether to get the details of the table (owner, size, etc.).
     * @param pageSize    The maximum number of tables displayed on each page
     * @param currentPage The current page number
     * @param type        Filter by table type.(EXTERNAL,MANAGED,VIEW)
     * @return List of tables with total count
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public PageTables listTableByPage(String keyword, boolean withDetail,
        Integer pageSize, Integer currentPage, TableType type) throws DLIException {
        if (databaseName == null) {
            LOGGER.error("Database name can not be null.");
            throw new DLIException("Database name can not be null.");
        }
        PageTables pageTables = new PageTables();
        try {
            TableListResult result =
                client.getRestApi()
                    .listAllTables(getProjectId(), databaseName, client.getAuthToken(), keyword, withDetail,
                        pageSize, currentPage, validateAndConvertToInternalType(type));
            List<Table> tables = new ArrayList<>();
            if (result.getIsSuccess()) {
                pageTables.setTotalCount(result.getTableCount());
                List<TableListInfo> tableListInfos = result.getTables();
                for (TableListInfo tableListInfo : tableListInfos) {
                    String tableName = tableListInfo.getTableName();
                    String dataLocation = tableListInfo.getDataLocation();
                    String tableType = tableListInfo.getTableType();
                    if (withDetail) {
                        long tableSize = tableListInfo.getTableSize();
                        String ownerInfo = tableListInfo.getOwner();
                        String location = tableListInfo.getLocation();
                        long lastAccessTime = tableListInfo.getLastAccessTime();
                        long createTime = tableListInfo.getCreateTime();
                        List<String> partitionColumns = tableListInfo.getPartitionColumns();
                        Table table = new Table(this, tableName, ownerInfo, createTime,
                            lastAccessTime, location, dataLocation, tableSize,
                            TableType.fromName(tableType), partitionColumns, tableListInfo.getDataType());
                        tables.add(table);
                    } else {
                        Table table = new Table(this, tableName, dataLocation, TableType.fromName(tableType));
                        tables.add(table);
                    }
                }
                pageTables.setTables(tables);
                return pageTables;
            }
            LOGGER.error(result.getMessage());
            throw new DLIException(result.getMessage());
        } catch (ApiException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e.getMessage(), e.getCode(), e.getResponseBody());
        }
    }

    /**
     * List all tables in the current database.
     *
     * @param keyword     The keyword of the filter table name.
     * @param withDetail  Whether to get the details of the table (owner, size, etc.).
     * @param pageSize    The maximum number of tables displayed on each page
     * @param currentPage The current page number
     * @param type        Filter by table type.(EXTERNAL,MANAGED,VIEW)
     * @return List of tables
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    private List<Table> listAllTables(String keyword, boolean withDetail,
        Integer pageSize, Integer currentPage, TableType type) throws DLIException {
        return listTableByPage(keyword, withDetail, pageSize, currentPage, type).getTables();
    }

    /**
     * SDK tableType is difference from rest api,this should convert to the real type in metadata.
     * EXTERNAL -> MANAGED_TABLE, MANAGED -> EXTERNAL_TABLE, VIEW -> VIRTUAL_VIEW
     *
     * @param tableType table type enum
     * @return table type string
     */
    private static String validateAndConvertToInternalType(TableType tableType) {
        if (null == tableType) {
            return null;
        } else if (TableType.EXTERNAL.equals(tableType)) {
            return "EXTERNAL_TABLE";
        } else if (TableType.MANAGED.equals(tableType)) {
            return "MANAGED_TABLE";
        } else if (TableType.VIEW.equals(tableType)) {
            return "VIRTUAL_VIEW";
        } else {
            LOGGER.error("Invalid table type for list tables: " + tableType);
            throw new IllegalArgumentException("Invalid table type for list tables: " + tableType);
        }
    }

    /**
     * Get the table in the current database by table name.
     *
     * @param tableName table name
     * @return The table in this database
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public Table getTable(String tableName) throws DLIException {
        List<Table> tables = listAllTables(tableName, true, null, null, null);
        for (Table table : tables) {
            if (table.getTableName().equalsIgnoreCase(tableName)) {
                return table;
            }
        }
        LOGGER.error("There is no Table named " + tableName);
        throw new DLIException("There is no Table named " + tableName);
    }

    /**
     * Create DLI table.
     *
     * @param tableName   table name
     * @param description table description
     * @param columns     table columns
     * @param sortColumns An list of column names to use for sorting in DLI table.
     * @return DLI table
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public Table createDLITable(String tableName, String description, List<Column> columns,
        List<String> sortColumns) throws DLIException {
        if (databaseName == null) {
            LOGGER.error("Database name can not be null.");
            throw new DLIException("Database name can not be null.");
        }
        TableInfo tableInfo = getTableInfo(tableName, "DLI", description, columns);
        if (sortColumns != null) {
            List<String> sorts = new ArrayList<String>();
            for (String sortColumn : sortColumns) {
                sorts.add(sortColumn);
            }
            tableInfo.setSortColumns(sorts);
        }
        return createTable(tableInfo);
    }

    private TableInfo getTableInfo(String tblName, String dataLoc, String tblDesc, List<Column> columns) {
        TableInfo tblInfo = new TableInfo();
        tblInfo.setTableName(tblName);
        tblInfo.setDataLocation(dataLoc);
        tblInfo.setDescription(tblDesc);
        List<ColumnInfo> colInfoList = new ArrayList<>();
        for (Column column : columns) {
            ColumnInfo c = new ColumnInfo();
            c.setColumnName(column.getColumnName());
            c.setDescription(column.getDescription());
            c.setType(column.getDataTypeInfo().toString());
            c.setIsPartitionColumn(column.isPartitionColumn());
            colInfoList.add(c);
        }
        tblInfo.setColumns(colInfoList);
        return tblInfo;
    }

    private Table createTable(TableInfo tblInfo) throws DLIException {
        ResponseInfo response = null;
        try {
            response = client.getRestApi().createTable(getProjectId(), databaseName, tblInfo, client.getAuthToken());
            if (response.getIsSuccess()) {
                return getTable(tblInfo.getTableName());
            }
            LOGGER.error(response.getMessage());
            throw new DLIException(response.getMessage());
        } catch (ApiException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e.getMessage(), e.getCode(), e.getResponseBody());
        }
    }

    /**
     * Create OBS table.
     *
     * @param tableName     table name
     * @param description   table description
     * @param columns       table columns
     * @param storageType   the type of data
     * @param dataPath      table data storage path
     * @param csvFormatInfo csv file format info.
     * @return OBS table
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public Table createObsTable(String tableName, String description, List<Column> columns,
        StorageType storageType, String dataPath,
        CsvFormatInfo csvFormatInfo) throws DLIException {
        if (databaseName == null) {
            LOGGER.error("Database name can not be null.");
            throw new DLIException("Database name can not be null.");
        }
        TableInfo tableInfo = getTableInfo(tableName, "OBS", description, columns);
        tableInfo.setDataType(storageType.getName());
        tableInfo.setDataPath(dataPath);
        if (csvFormatInfo != null) {
            tableInfo.setWithColumnHeader(csvFormatInfo.isWithColumnHeader());
            tableInfo.setDelimiter(csvFormatInfo.getDelimiter());
            tableInfo.setQuoteChar(csvFormatInfo.getQuoteChar());
            tableInfo.setEscapeChar(csvFormatInfo.getEscapeChar());
            tableInfo.setDateFormat(csvFormatInfo.getDateFormat());
            tableInfo.setTimestampFormat(csvFormatInfo.getTimestampFormat());
        }
        return createTable(tableInfo);
    }

    /**
     * Create DLI/OBS table
     *
     * @param tableName     table name
     * @param dataLocation  DLI or OBS
     * @param description   table description
     * @param columns       table columns
     * @param storageType   the type of data,only OBS table configuration,DLI table set null.
     * @param dataPath      table data storage path,only OBS table configuration,DLI table set null.
     * @param csvFormatInfo csv file format info,only OBS table configuration,DLI table set null.
     * @param sortColumns   An list of column names to use for sorting in DLI table,OBS table set null.
     * @return table
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public Table createTable(String tableName, String dataLocation, String description, List<Column> columns,
        StorageType storageType, String dataPath, CsvFormatInfo csvFormatInfo,
        List<String> sortColumns) throws DLIException {
        if ("DLI".equals(dataLocation)) {
            LOGGER.info("Create DLITable succeed");
            return createDLITable(tableName, description, columns, sortColumns);
        }
        if ("OBS".equals(dataLocation)) {
            LOGGER.info("Create ObsTable succeed");
            return createObsTable(tableName, description, columns, storageType, dataPath, csvFormatInfo);
        }
        LOGGER.error("data_location is illegal");
        throw new DLIException("data_location is illegal");
    }

    public TableSchema getTableDetail(String tableName) throws DLIException {
        TableSchema tableSchema = new TableSchema();
        List<Column> columns = new ArrayList<Column>();
        try {
            ColumnResult response = this.getClient().getRestApi().describeTable(this.getProjectId(),
                    this.getDatabaseName(), tableName, this.getClient().getAuthToken());
            if (response.getIsSuccess()) {
                List<ColumnInfo> columnList = response.getColumns();
                for (ColumnInfo columnInfo : columnList) {
                    String columnType = columnInfo.getType();
                    DataTypeInfo dataTypeInfo = TypeInfoUtils.getDataTypeInfo(columnType);
                    Column column = new Column(
                            columnInfo.getColumnName(),
                            dataTypeInfo,
                            columnInfo.getDescription(),
                            columnInfo.getIsPartitionColumn() == null ? false : columnInfo.getIsPartitionColumn());
                    columns.add(column);
                }
                tableSchema.setColumns(columns);
                List<String> sortColumns = response.getSortColumns();
                tableSchema.setSortColumns(sortColumns);
                String createTableSql = response.getCreateTableSql();
                tableSchema.setCreateTableSql(createTableSql);
                String tableComment = response.getTableComment();
                tableSchema.setTableComment(tableComment);
                return tableSchema;
            }
            LOGGER.error(response.getMessage());
            throw new DLIException(response.getMessage());
        } catch (ApiException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e.getMessage(), e.getCode(), e.getResponseBody());
        }
    }

    public DLIClient getClient() {
        return client;
    }

    @Override
    public String toString() {
        return "Database{"
            + "databaseName='" + databaseName + '\''
            + ", description='" + description + '\''
            + ", owner='" + owner + '\''
            + '}';
    }

}
