package com.huawei.dli.sdk;

import com.huawei.dli.sdk.common.BadRecordPolicy;
import com.huawei.dli.sdk.common.Column;
import com.huawei.dli.sdk.common.CompressType;
import com.huawei.dli.sdk.common.DLIInfo;
import com.huawei.dli.sdk.common.DownloadJobParam;
import com.huawei.dli.sdk.common.ExportMode;
import com.huawei.dli.sdk.common.JobStatus;
import com.huawei.dli.sdk.common.PartitionSpec;
import com.huawei.dli.sdk.common.StorageType;
import com.huawei.dli.sdk.common.utils.ObsClientUtils;
import com.huawei.dli.sdk.exception.DLIException;

import com.obs.services.model.ObsObject;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringJoiner;
import java.util.stream.Collectors;

public class DownloadJob extends TunnelJob {
    private static final Logger logger = LoggerFactory.getLogger(DownloadJob.class);

    private SQLJob job;

    private FileFormat fileFormat;

    private boolean hasRenameCol;

    private boolean isReady = false;

    private List<ObsObject> objects;

    private String transactionId;

    private BadRecordPolicy badRecordPolicy = BadRecordPolicy.EXCEPTION;

    /**
     * Create a download job to download query result
     *
     * @param queue the queue to running the download spark job
     * @param job   the query sql job
     * @throws DLIException the exception if create download job failed
     */
    public DownloadJob(Queue queue, SQLJob job) throws DLIException {
        this(queue, job, 0);
    }

    /**
     * Create a download job to download query result
     *
     * @param queue the queue to running the download spark job
     * @param job   the query sql job
     * @param limit the limit number of download result
     * @throws DLIException the exception if create download job failed
     */
    public DownloadJob(Queue queue, SQLJob job, Integer limit) throws DLIException {
        prepare(queue, job.getResultSchema());
        this.job = job;
        this.fileFormat = FileFormat.JSON;
        this.hasRenameCol = false;
        String dataPath = String.format("obs://%s/%s", bucketName, objectPrefix);
        String exportJobId = this.job.asyncExportResult(dataPath, StorageType.JSON, CompressType.NONE,
            ExportMode.ERRORIFEXISTS, queue.getQueueName(), false, limit, "utf-8");
        this.job.setJobId(exportJobId);
    }

    /**
     * Create a download job to download data from DLI.
     *
     * @param queue    the queue to running the download spark job
     * @param jobParam the job param include dbName, tableName,fileFormat,partitions
     * @throws DLIException throw exception if create download job failed
     */
    public DownloadJob(Queue queue, DownloadJobParam jobParam) throws DLIException {
        if (jobParam.getBadRecordPolicy() != null) {
            this.badRecordPolicy = jobParam.getBadRecordPolicy();
        }
        if (StringUtils.isNotBlank(jobParam.getTransactionId())) {
            this.transactionId = jobParam.getTransactionId();
        }
        if (jobParam.getPartitionSpecs() != null && jobParam.getPartitionSpecs().size() > 0) {
            initWithPartitions(queue, jobParam.getDbName(), jobParam.getTblName(), jobParam.getPartitionSpecs(),
                jobParam.getFileFormat() == null ? FileFormat.AVRO : jobParam.getFileFormat());
        } else {
            initWithNoPartitions(queue, jobParam.getDbName(), jobParam.getTblName(),
                jobParam.getFileFormat() == null ? FileFormat.AVRO : jobParam.getFileFormat());
        }
    }

    /**
     * Create a download job to download data from DLI.
     *
     * @param queue   the queue to running the download spark job
     * @param dbName  the database name which the data to download
     * @param tblName the table name which the data to download
     * @throws DLIException throw exception if create download job failed
     */
    public DownloadJob(Queue queue, String dbName, String tblName) throws DLIException {
        this.initWithNoPartitions(queue, dbName, tblName, FileFormat.AVRO);
    }

    /**
     * Create a download job to download data from DLI.
     *
     * @param queue          the queue to running the download spark job
     * @param dbName         the database name which the data to download
     * @param tblName        the table name which the data to download
     * @param partitionSpecs the multi partition spec for a partitioned table
     * @throws DLIException throw exception if create download job failed
     */
    public DownloadJob(Queue queue, String dbName, String tblName, List<String> partitionSpecs) throws DLIException {
        this(queue, dbName, tblName,
            partitionSpecs.stream().map(PartitionSpec::new).collect(Collectors.toList()), FileFormat.AVRO);
    }

    /**
     * Create a download job to download data from DLI.
     *
     * @param queue          the queue to running the download spark job
     * @param dbName         the database name which the data to download
     * @param tblName        the table name which the data to download
     * @param partitionSpecs the multi partition spec for a partitioned table
     * @throws DLIException throw exception if create download job failed
     */
    private DownloadJob(Queue queue, String dbName, String tblName, List<PartitionSpec> partitionSpecs,
        FileFormat fileFormat) throws DLIException {
        this.initWithPartitions(queue, dbName, tblName, partitionSpecs, fileFormat);
    }

    private void initWithPartitions(Queue queue, String dbName, String tblName, List<PartitionSpec> partitionSpecs,
        FileFormat fileFormat) throws DLIException {
        prepare(queue, dbName, tblName, "download");
        if (partCols.size() == 0) {
            throw new DLIException("It is not a partitioned table, please check.");
        }
        if (partitionSpecs == null) {
            throw new DLIException("Invalid partition spec or there are unexpected partition column, please check.");
        }

        StringJoiner selectColumns = new StringJoiner(", ");
        List<Column> columns = tblSchema.getColumns();
        for (int i = 0; i < columns.size(); i++) {
            selectColumns.add(String.format("`%s` AS `c%d`", columns.get(i).getColumnName(), i));
        }

        StringBuilder finalWhereClause = new StringBuilder();
        for (int num = 0; num < partitionSpecs.size(); num++) {
            PartitionSpec partitionSpec = partitionSpecs.get(num);
            if (partitionSpec == null || !validPartSpec(partitionSpec)) {
                throw new DLIException("Invalid partition spec or there are unexpected partition column,"
                    + " please check.");
            }
            String whereClause = partitionSpec.keys().stream()
                .map(partKey -> String.format("`%s`='%s'", partKey, partitionSpec.get(partKey)))
                .collect(Collectors.joining(" AND "));
            // add 'or' condition for multi partition
            if (partitionSpecs.size() != 1) {
                if (num + 1 <= partitionSpecs.size()) {
                    whereClause = "(" + whereClause + ")";
                }
                if (num + 1 != partitionSpecs.size()) {
                    whereClause = whereClause + " OR ";
                }
            }
            finalWhereClause.append(whereClause);
        }
        String exportSql = String.format(
            "INSERT OVERWRITE DIRECTORY 'obs://%s/%s' USING %s "
                + "OPTIONS(timestampFormat 'yyyy-MM-dd HH:mm:ss.SSSZZ', nullValue '%s') "
                + "SELECT %s FROM `%s`.`%s` WHERE %s ",
            bucketName, objectPrefix, fileFormat, nullValue, selectColumns.toString(), dbName, tblName,
            finalWhereClause.toString());
        this.job = new SQLJob(queue, exportSql);
        this.job.setTransactionId(transactionId);
        Map<String, Object> conf = new HashMap<>();
        conf.put("spark.sql.datasource.csv.treatEmptyStringAsNull", "false");
        List<Map<String, Object>> confList = new ArrayList<>();
        confList.add(conf);
        this.job.setConf(confList);
        this.fileFormat = fileFormat;
        this.hasRenameCol = true;
        this.job.asyncSubmit();
    }

    @Deprecated
    public DownloadJob(Queue queue, String dbName, String tblName, BadRecordPolicy badRecordPolicy)
        throws DLIException {
        this(queue, dbName, tblName);
        this.badRecordPolicy = badRecordPolicy;
    }

    private void initWithNoPartitions(Queue queue, String dbName, String tblName, FileFormat fileFormat)
        throws DLIException {
        prepare(queue, dbName, tblName, "download");
        StringJoiner selectColumns = new StringJoiner(", ");
        List<Column> columns = tblSchema.getColumns();
        for (int i = 0; i < columns.size(); i++) {
            selectColumns.add(String.format("`%s` AS `c%d`", columns.get(i).getColumnName(), i));
        }
        String exportSql = String.format(
            "INSERT OVERWRITE DIRECTORY 'obs://%s/%s' USING %s "
                + "OPTIONS(timestampFormat 'yyyy-MM-dd HH:mm:ss.SSSZZ', nullValue '%s') "
                + "SELECT %s FROM `%s`.`%s` ",
            bucketName, objectPrefix, fileFormat, nullValue, selectColumns.toString(), dbName, tblName);
        this.job = new SQLJob(queue, exportSql);
        this.job.setTransactionId(transactionId);
        Map<String, Object> conf = new HashMap<>();
        conf.put("spark.sql.datasource.csv.treatEmptyStringAsNull", "false");
        List<Map<String, Object>> confList = new ArrayList<>();
        confList.add(conf);
        this.job.setConf(confList);
        this.fileFormat = fileFormat;
        this.hasRenameCol = true;
        this.job.asyncSubmit();
    }

    public DownloadJob(Queue queue, String dbName, String tblName, String partitionSpec) throws DLIException {
        this(queue, dbName, tblName, new PartitionSpec(partitionSpec));
    }

    @Deprecated
    public DownloadJob(Queue queue, String dbName, String tblName, String partitionSpec,
        BadRecordPolicy badRecordPolicy) throws DLIException {
        this(queue, dbName, tblName, partitionSpec);
        this.badRecordPolicy = badRecordPolicy;
    }

    /**
     * Create a download job to download data from DLI.
     *
     * @param queue         the queue to running the download spark job
     * @param dbName        the database name which the data to download
     * @param tblName       the table name which the data to download
     * @param partitionSpec the partition spec for a partitioned table
     * @throws DLIException throw exception if create download job failed
     */
    public DownloadJob(Queue queue, String dbName, String tblName, PartitionSpec partitionSpec) throws DLIException {
        this(queue, dbName, tblName, partitionSpec, FileFormat.AVRO);
    }

    @Deprecated
    public DownloadJob(Queue queue, String dbName, String tblName, PartitionSpec partitionSpec,
        BadRecordPolicy badRecordPolicy) throws DLIException {
        this(queue, dbName, tblName, partitionSpec);
        this.badRecordPolicy = badRecordPolicy;
    }

    private DownloadJob(Queue queue, String dbName, String tblName, PartitionSpec partitionSpec, FileFormat fileFormat)
        throws DLIException {
        this(queue, dbName, tblName, Collections.singletonList(partitionSpec), fileFormat);
    }

    public String[] emptyRowValues() {
        return new String[this.tblSchema.getColumns().size()];
    }

    public BadRecordPolicy getBadRecordPolicy() {
        return badRecordPolicy;
    }

    @Deprecated
    public void setMaxCharsPerColumn(int maxCharsPerColumn) {
        /* do nothing */
    }

    public synchronized Reader createReader() throws DLIException {
        if (!isReady) {
            throw new DLIException("There is not ready for create reader, please check status first.");
        }

        DLIInfo dliInfo = job.getQueue().getClient().getDLIInfo();
        return createInternalReader(dliInfo, objects);
    }

    private void initForReader() {
        DLIInfo dliInfo = job.getQueue().getClient().getDLIInfo();
        ObsClientUtils obsClientUtils = new ObsClientUtils(dliInfo.getObsEndPoint());
        objects = obsClientUtils.listObjects(dliInfo.getAccessKey(), dliInfo.getSecretKey(), bucketName, objectPrefix);
    }

    private Reader createInternalReader(DLIInfo dliInfo, List<ObsObject> objectsToReader) throws DLIException {
        try {
            Reader reader;
            switch (fileFormat) {
                case CSV:
                    reader = new ObsReader(
                        this,
                        dliInfo.getObsEndPoint(),
                        dliInfo.getAccessKey(),
                        dliInfo.getSecretKey(),
                        objectsToReader);
                    ((ObsReader) reader).setNullValue(nullValue);
                    break;
                case JSON:
                    reader = new JsonObsReader(
                        this,
                        dliInfo.getObsEndPoint(),
                        dliInfo.getAccessKey(),
                        dliInfo.getSecretKey(),
                        objectsToReader);
                    if (hasRenameCol) {
                        ((JsonObsReader) reader).setHasRenameCol(true);
                    }
                    break;
                case AVRO:
                    reader = new AvroObsReader(
                        this,
                        dliInfo.getObsEndPoint(),
                        dliInfo.getAccessKey(),
                        dliInfo.getSecretKey(),
                        objectsToReader);
                    if (hasRenameCol) {
                        ((AvroObsReader) reader).setHasRenameCol(true);
                    }
                    break;
                default:
                    throw new DLIException("Unexpected inner file format: " + fileFormat.getName());
            }
            reader.openReader();
            return reader;
        } catch (IOException e) {
            logger.error("Failed to open reader", e);
            throw new DLIException(e);
        }
    }

    /**
     * Get readers to read data in concurrent, the reader num is calculate by DLI with the expectedMaxConcurrent,
     * that is the reader num is less or equal expectedMaxConcurrent if the data size is small.
     *
     * @param expectedMaxConcurrent the max concurrent
     * @return the reader list
     * @throws DLIException the exception while create readers
     */
    public synchronized List<Reader> createReaders(int expectedMaxConcurrent) throws DLIException {
        if (!isReady) {
            throw new DLIException("There is not ready for create reader, please check status first.");
        }
        int realConcurrent = Math.min(expectedMaxConcurrent, objects.size());
        List<Reader> readers = new ArrayList<>(realConcurrent);
        DLIInfo dliInfo = job.getQueue().getClient().getDLIInfo();
        int step = (int) Math.ceil((double) objects.size() / realConcurrent);
        for (int i = 0; i < objects.size(); ) {
            readers.add(createInternalReader(dliInfo, copyList(objects, i, i + step)));
            i += step;
        }
        return readers;
    }

    private List<ObsObject> copyList(List<ObsObject> originList, int start, int end) {
        List<ObsObject> newList = new ArrayList<>(end - start);
        int fEnd = Math.min(end, originList.size());
        for (int i = start; i < fEnd; i++) {
            newList.add(originList.get(i));
        }
        return newList;
    }

    public synchronized String getInternalJobId() throws DLIException {
        if (job != null) {
            return job.getJobId();
        }
        throw new DLIException("There is not job have been created, please check.");
    }

    public synchronized String getDownloadStatus() throws DLIException {
        if (job != null) {
            JobStatus status = job.getStatus();
            logger.info("Download job internal status: {}", status.getName());
            if (status == JobStatus.FINISHED) {
                isReady = true;
                initForReader();
            }
            return status.getName();
        }
        throw new DLIException("There is not job have been created, please check.");
    }

    public synchronized long getResultCount() throws DLIException {
        if (job != null) {
            return job.getRowCount();
        }
        throw new DLIException("There is not job have been created, please check.");
    }
}
