package com.huawei.dli.sdk;

import com.huawei.dli.restapi.ApiException;
import com.huawei.dli.restapi.model.ImportJobRequest;
import com.huawei.dli.restapi.model.ResponseJobInfo;
import com.huawei.dli.sdk.common.CsvFormatInfo;
import com.huawei.dli.sdk.common.PartitionSpec;
import com.huawei.dli.sdk.common.StorageType;
import com.huawei.dli.sdk.exception.DLIException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import lombok.Getter;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

@Getter
public class ImportJob extends Job {

    private static final StorageType STORAGE_TYPE_DEFAULT = StorageType.CSV;

    private static final Logger LOGGER = LoggerFactory.getLogger(ImportJob.class);

    private final String databaseName;

    private final String tableName;

    private final String dataPath;

    private StorageType storageType;

    private CsvFormatInfo csvFormatInfo;

    private Boolean isOverWrite = false;

    private PartitionSpec partitionSpec = null;

    public ImportJob(Queue queue, String databaseName, String tableName, String dataPath) {
        super(queue);
        this.databaseName = databaseName;
        this.tableName = tableName;
        this.dataPath = dataPath;
        this.storageType = STORAGE_TYPE_DEFAULT;
    }

    public ImportJob(
        Queue queue,
        String databaseName,
        String tableName,
        String dataPath,
        String partitionSpec,
        Boolean isOverWrite) {
        super(queue);
        this.databaseName = databaseName;
        this.tableName = tableName;
        this.dataPath = dataPath;
        this.storageType = STORAGE_TYPE_DEFAULT;
        if (partitionSpec == null) {
            this.partitionSpec = null;
        } else {
            this.partitionSpec = new PartitionSpec(partitionSpec);
        }
        this.isOverWrite = isOverWrite;
    }

    public ImportJob(
        Queue queue,
        String databaseName,
        String tableName,
        String dataPath,
        PartitionSpec partitionSpec,
        Boolean isOverWrite) {
        super(queue);
        this.databaseName = databaseName;
        this.tableName = tableName;
        this.dataPath = dataPath;
        this.storageType = STORAGE_TYPE_DEFAULT;
        this.partitionSpec = partitionSpec;
        this.isOverWrite = isOverWrite;
    }

    public void setStorageType(StorageType storageType) {
        this.storageType = storageType;
    }

    public void setCsvFormatInfo(CsvFormatInfo csvFormatInfo) {
        this.csvFormatInfo = csvFormatInfo;
    }

    public void setOverWrite(Boolean overWrite) {
        isOverWrite = overWrite;
    }

    public void setPartitionSpec(PartitionSpec partitionSpec) {
        this.partitionSpec = partitionSpec;
    }

    /**
     * Submit import job
     *
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public void submit() throws DLIException {
        asyncSubmit();
        super.checkJobRunningStatus(getJobId());
    }

    private void setImportJobConfig(ImportJobRequest jobRequest) {
        if (super.getConf() != null && !super.getConf().isEmpty()) {
            List<String> confStr = new ArrayList<>();
            for (Map<String, Object> confMap : super.getConf()) {
                Iterator<Map.Entry<String,Object>> confEntry = confMap.entrySet().iterator();
                while (confEntry.hasNext()) {
                    Map.Entry<String,Object> confKV = confEntry.next();
                    if (confKV.getKey() != null && confKV.getValue() != null) {
                        confStr.add(confKV.getKey().concat("=").concat((String) confKV.getValue()));
                    }
                }
            }
            jobRequest.setConf(confStr);
        }
    }

    public void asyncSubmit() throws DLIException {
        if (getJobId() != null) {
            LOGGER.error("Job had been submitted, don't submit repeatedly");
            throw new DLIException("Job had been submitted, don't submit repeatedly");
        }
        ImportJobRequest jobRequest = new ImportJobRequest();
        jobRequest.setDataPath(dataPath);
        jobRequest.setDataType(storageType.getName());
        jobRequest.setDatabaseName(databaseName);
        jobRequest.setTableName(tableName);
        if (!"default".equals(getQueue().getQueueName())) {
            jobRequest.setQueueName(getQueue().getQueueName());
        }
        CsvFormatInfo formatInfo = getCsvFormatInfo();
        if (formatInfo != null) {
            jobRequest.setWithColumnHeader(formatInfo.isWithColumnHeader());
            jobRequest.setDelimiter(formatInfo.getDelimiter());
            jobRequest.setQuoteChar(formatInfo.getQuoteChar());
            jobRequest.setEscapeChar(formatInfo.getEscapeChar());
            jobRequest.setDateFormat(formatInfo.getDateFormat());
            jobRequest.setTimestampFormat(formatInfo.getTimestampFormat());
        }

        if (partitionSpec != null) {
            jobRequest.setPartitionSpec(partitionSpec.getKv());
        }
        jobRequest.setOverwrite(isOverWrite);

        setImportJobConfig(jobRequest);

        ResponseJobInfo response = null;
        try {
            response = client.getRestApi().importData(getProjectId(), jobRequest, client.getAuthToken());
            if (!response.getIsSuccess()) {
                LOGGER.error(response.getMessage());
                throw new DLIException(response.getMessage());
            }
            super.setJobId(response.getJobId());
        } catch (ApiException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e.getMessage(), e.getCode(), e.getResponseBody());
        }
    }
}
