package com.huawei.dli.sdk;

import com.huawei.dli.restapi.ApiException;
import com.huawei.dli.restapi.model.ResponseInfo;
import com.huawei.dli.restapi.model.StatusResult;
import com.huawei.dli.sdk.common.JobResultInfo;
import com.huawei.dli.sdk.common.JobStatus;
import com.huawei.dli.sdk.common.JobType;
import com.huawei.dli.sdk.exception.DLIException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.Map;

public abstract class Job extends BaseJob {

    public static final long JOB_TIMEOUT_SECOND = 300;

    private static final Logger LOGGER = LoggerFactory.getLogger(Job.class);

    private String jobId;

    private Queue queue;

    private long jobTimeout = JOB_TIMEOUT_SECOND;

    private List<Map<String, Object>> conf;

    private JobType jobType;

    private Map<String, String> tags;

    public Job(Queue queue) {
        super(queue.getClient());
        this.queue = queue;
    }

    @Override
    protected String getTransactionKeyName() {
        return "spark.sql.dli.job.transaction.id";
    }

    public String getJobId() {
        return jobId;
    }

    public Queue getQueue() {
        return queue;
    }

    public void setJobId(String jobId) {
        this.jobId = jobId;
    }

    public String getProjectId() {
        return client.getAuthentication().getProjectId();
    }

    public JobType getJobType() {
        return jobType;
    }

    protected void setJobType(JobType jobType) {
        this.jobType = jobType;
    }

    public Map<String, String> getTags() {
        return tags;
    }

    public void setTags(Map<String, String> tags) {
        this.tags = tags;
    }

    /**
     * Cancel RUNNING or LAUNCHING status job.
     *
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public void cancelJob() throws DLIException {
        if (jobId == null) {
            LOGGER.error("No job in progress,cancel failed");
            throw new DLIException("No job in progress,cancel failed");
        }
        JobStatus status = getStatus();
        if (JobStatus.FINISHED.equals(status) || JobStatus.FAILED.equals(status)) {
            LOGGER.error("The job has finished,can not be canceled");
            throw new DLIException("The job has finished,can not be canceled");
        }
        if (JobStatus.CANCELLED.equals(status)) {
            LOGGER.error("The job has canceled");
            throw new DLIException("The job has canceled");
        }
        try {
            ResponseInfo response = client.getRestApi().cancelJob(getProjectId(), jobId, client.getAuthToken());
            if (!response.getIsSuccess()) {
                LOGGER.error(response.getMessage());
                throw new DLIException(response.getMessage());
            }
        } catch (ApiException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e.getMessage(), e.getCode(), e.getResponseBody());
        }
    }

    public abstract void submit() throws DLIException;

    /**
     * Get the latest job status
     *
     * @return JobStatus
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public JobStatus getStatus() throws DLIException {
        JobResultInfo jobResultInfo = getStatusResult(jobId);
        if (jobResultInfo != null) {
            return jobResultInfo.getJobStatus();
        }
        throw new DLIException("Get job status failed, please check job:" + jobId);
    }

    /**
     * Get this job resultInfo: detail,statement,status...
     *
     * @param jobId the job id
     * @return JobResultInfo
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    protected JobResultInfo getStatusResult(String jobId) throws DLIException {
        try {
            StatusResult response = client.getRestApi().getJobStatus(getProjectId(), jobId, client.getAuthToken());
            if (response.getIsSuccess()) {
                return new JobResultInfo(response);
            }
            LOGGER.error(response.getMessage());
            throw new DLIException(response.getMessage());
        } catch (ApiException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e.getMessage(), e.getCode(), e.getResponseBody());
        }
    }

    /**
     * Set job timeout, default 300s
     *
     * @param jobTimeout job time out in second
     */
    public void setJobTimeout(int jobTimeout) {
        this.jobTimeout = jobTimeout;
    }

    public long getJobTimeout() {
        return jobTimeout;
    }

    public List<Map<String, Object>> getConf() {
        return conf;
    }

    public void setConf(List<Map<String, Object>> conf) {
        this.conf = conf;
    }

    /**
     * Check job running status by jobId after submit(),
     * when job status is FINISHED,FAILED,CANCELLED,mean job is over.
     * When exceeded the job timeout,end of the submit() method,
     * but job is running in the server.
     * You can use getJobId() to get jobId and use the jobId to execute
     * queryJobResultInfo(String jobId) and cancelJob(String jobId) method from DLIClient.
     *
     * @param jobId the job id
     *
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    protected void checkJobRunningStatus(String jobId) throws DLIException {
        long startTime = System.currentTimeMillis();
        long waitTime = 1000L;
        JobResultInfo jobResultInfo = null;
        while (true) {
            try {
                jobResultInfo = getStatusResult(jobId);
                if (System.currentTimeMillis() - startTime > 5 * 60 * 1000L) {
                    waitTime = 10000L;
                }
            } catch (DLIException ex) {
                if (ex.getMessage().contains("APIGW.0308")) {
                    if (waitTime < 10000L) {
                        waitTime += 1000L;
                    }
                    try {
                        Thread.sleep(waitTime);
                    } catch (InterruptedException e) {
                        LOGGER.error("msg", e);
                        throw new DLIException(e);
                    }
                    continue;
                } else {
                    throw ex;
                }
            }
            if (checkJobFinish(startTime, jobResultInfo, waitTime)) {
                return;
            }
        }
    }

    protected boolean checkJobFinish(long startTime, JobResultInfo jobResultInfo) throws DLIException {
        return checkJobFinish(startTime, jobResultInfo, 1000L);
    }

    protected boolean checkJobFinish(long startTime, JobResultInfo jobResultInfo, Long waitTime) throws DLIException {
        if (jobResultInfo == null) {
            throw new DLIException("There is no result of job, please check");
        }
        if ((JobStatus.FINISHED).equals(jobResultInfo.getJobStatus())) {
            return true;
        }
        if ((JobStatus.FAILED).equals(jobResultInfo.getJobStatus())) {
            LOGGER.error(jobResultInfo.getMessage());
            throw new DLIException(jobResultInfo.getMessage());
        }
        if ((JobStatus.CANCELLED).equals(jobResultInfo.getJobStatus())) {
            LOGGER.error(jobResultInfo.getMessage());
            throw new DLIException(jobResultInfo.getMessage());
        }
        try {
            Thread.sleep(waitTime);
        } catch (InterruptedException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e);
        }
        long endTime = System.currentTimeMillis();
        long timeout = endTime - startTime;
        if (getJobTimeout() != -1 && timeout >= getJobTimeout() * 1000) {
            LOGGER.error("Timed out for Job " + jobResultInfo.getJobStatus() + " status: " + timeout + "ms."
                + "The Job id is " + getJobId());
            throw new DLIException(
                "Timed out for Job " + jobResultInfo.getJobStatus() + " status: " + timeout + "ms."
                    + "The Job id is " + getJobId());
        }
        return false;
    }
}
