package com.huawei.dli.sdk;

import com.huawei.dli.sdk.common.Column;
import com.huawei.dli.sdk.common.DataType;
import com.huawei.dli.sdk.common.DataTypeInfo;
import com.huawei.dli.sdk.common.Row;
import com.huawei.dli.sdk.common.type.ArrayTypeInfo;
import com.huawei.dli.sdk.common.type.MapTypeInfo;
import com.huawei.dli.sdk.common.type.StructTypeInfo;
import com.huawei.dli.sdk.common.type.TypeUtils;
import com.huawei.dli.sdk.common.utils.ObsClientUtils;
import com.huawei.dli.sdk.exception.DLIException;

import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import com.obs.services.model.ObsObject;

import org.apache.commons.io.IOUtils;
import org.apache.commons.io.LineIterator;
import org.apache.commons.lang3.time.FastDateFormat;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.nio.charset.StandardCharsets;
import java.sql.Timestamp;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.TimeZone;

public class JsonObsReader implements Reader {
    private static final Logger logger = LoggerFactory.getLogger(JsonObsReader.class);

    private static final FastDateFormat FAST_DATE_FORMAT =
        FastDateFormat.getInstance("yyyy-MM-dd", Locale.US);

    private static final ThreadLocal<DateFormat> TIMESTAMP_FORMAT =
        ThreadLocal.withInitial(() -> new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSSZZ"));

    private static final FastDateFormat FAST_TIMESTAMP_FORMAT =
        FastDateFormat.getInstance("yyyy-MM-dd HH:mm:ss.SSSZZ", TimeZone.getDefault(), Locale.US);

    // the default value on Spark, if our export-result api support set timestampFormat option, then remove this
    private static final FastDateFormat DEFAULT_FAST_TIMESTAMP_FORMAT =
        FastDateFormat.getInstance("yyyy-MM-dd'T'HH:mm:ss.SSSXXX", TimeZone.getDefault(), Locale.US);

    private static final int MAX_RECONNECT_NUM = 5;

    private final DownloadJob downloadJob;

    private final ObsClientUtils obsClientUtils;

    private final String ak;

    private final String sk;

    private String bucketName;

    private String objectPrefix;

    private List<ObsObject> objects;

    private int totalIter = 0;

    private int nextIterIdx = 0;

    private LineIterator iterator;

    private int readRecordsPerIter = 0;

    private boolean hasRenameCol = false;

    protected boolean closed = false;

    @Deprecated
    public JsonObsReader(
        DownloadJob downloadJob,
        String endpoint,
        String ak,
        String sk,
        String bucketName,
        String objectPrefix) {
        this.downloadJob = downloadJob;
        this.obsClientUtils = new ObsClientUtils(endpoint);
        this.ak = ak;
        this.sk = sk;
        this.bucketName = bucketName;
        this.objectPrefix = objectPrefix;
    }

    public JsonObsReader(
        DownloadJob downloadJob,
        String endpoint,
        String ak,
        String sk,
        List<ObsObject> objects) {
        this.downloadJob = downloadJob;
        this.obsClientUtils = new ObsClientUtils(endpoint);
        this.ak = ak;
        this.sk = sk;
        this.objects = objects;
    }

    @Override
    public void openReader() throws IOException {
        if (objects == null) {
            objects = obsClientUtils.listObjects(ak, sk, bucketName, objectPrefix);
        }
        totalIter = objects.size();
        logger.info("object size: {}", totalIter);

        iterator = getNextIterator();
    }

    private LineIterator getNextIterator() {
        if (totalIter > 0 && nextIterIdx < totalIter) {
            readRecordsPerIter = 0;
            return createIterator(false);
        } else {
            return null;
        }
    }

    private LineIterator createIterator(boolean recreate) {
        int currIterIdx = recreate ? nextIterIdx - 1 : nextIterIdx++;
        ObsObject object = objects.get(currIterIdx);
        logger.info("{} iterator for object: {}", (recreate ? "recreate" : "create"), object.getObjectKey());
        InputStream inputStream = obsClientUtils.getObjectStream(ak, sk, object.getBucketName(), object.getObjectKey());
        try {
            return IOUtils.lineIterator(inputStream, StandardCharsets.UTF_8);
        } catch (IOException e) {
            logger.error("failed to create iterator", e);
            throw new IllegalStateException("Unexpected error as create iterator");
        }
    }

    @Override
    public Row read() throws DLIException, IOException {
        if (closed) {
            throw new IllegalStateException("The reader state is closed");
        }

        for (int times = 0; times < MAX_RECONNECT_NUM; times++) {
            try {
                if (!iterator.hasNext()) {
                    logger.info("read {} records from pre iterator", readRecordsPerIter);
                    if (nextIterIdx < totalIter) {
                        iterator = getNextIterator();
                    }
                }
                if (iterator != null && iterator.hasNext()) {
                    String values = iterator.next();
                    Row row = downloadJob.newRow();
                    updateRow(values, row);
                    readRecordsPerIter++;
                    return row;
                } else {
                    return null;
                }
            } catch (Exception e) {
                logger.warn("get record failed, may be connection closed by obs, so reconnect it!!!", e);
                reconnectAndSkipHandled();
            }
        }
        throw new DLIException("Failed to read value with retry");
    }

    /**
     * A flag used to determine whether to read JSON data based on the name of c$i.
     * Do not set this value unless necessary.
     *
     * @param hasRenameCol If rename the column name to c$i when write data.
     */
    public void setHasRenameCol(boolean hasRenameCol) {
        this.hasRenameCol = hasRenameCol;
    }

    private void updateRow(String jsonStr, Row row) throws DLIException {
        JsonObject jsonObject = JsonParser.parseString(jsonStr).getAsJsonObject();
        List<Column> columns = row.getSchema().getColumns();
        for (int i = 0; i < columns.size(); i++) {
            Column column = columns.get(i);
            String colNameInFile = column.getColumnName();
            if (hasRenameCol) {
                colNameInFile = "c" + i;
            }
            JsonElement jsonElement = null;
            if (jsonObject.has(colNameInFile)) {
                jsonElement = jsonObject.get(colNameInFile);
            } else {
                // Use Spark write data, if col's value is null, will not write into file!!!
                logger.debug("missing data for column: {}", column.getColumnName());
            }
            updateCol(row, i, column.getType(), column.getDataTypeInfo(), jsonElement);
        }
    }

    private void updateCol(Row row, int i, DataType dataType, DataTypeInfo dataTypeInfo, JsonElement element)
        throws DLIException {
        switch (dataType) {
            case BOOLEAN:
                Boolean boolData = (element == null || element.isJsonNull()) ? null : element.getAsBoolean();
                row.setBoolean(i, boolData);
                break;
            case TINYINT:
                Byte byteData = (element == null || element.isJsonNull()) ? null : element.getAsByte();
                row.setTinyint(i, byteData);
                break;
            case SMALLINT:
            case SHORT:
                Short shortData = (element == null || element.isJsonNull()) ? null : element.getAsShort();
                row.setShort(i, shortData);
                break;
            case INT:
                Integer intData = (element == null || element.isJsonNull()) ? null : element.getAsInt();
                row.setInt(i, intData);
                break;
            case LONG:
                Long longData = (element == null || element.isJsonNull()) ? null : element.getAsLong();
                row.setLong(i, longData);
                break;
            case BIGINT:
                BigInteger bigIntData = (element == null || element.isJsonNull()) ? null : element.getAsBigInteger();
                row.setBigInteger(i, bigIntData);
                break;
            case FLOAT:
                Float floatData = (element == null || element.isJsonNull()) ? null : element.getAsFloat();
                row.setFloat(i, floatData);
                break;
            case DOUBLE:
                Double doubleData = (element == null || element.isJsonNull()) ? null : element.getAsDouble();
                row.setDouble(i, doubleData);
                break;
            case DECIMAL:
                BigDecimal decimalData = (element == null || element.isJsonNull()) ? null : element.getAsBigDecimal();
                row.setBigDecimal(i, decimalData);
                break;
            case STRING:
                row.setString(i, (element == null || element.isJsonNull()) ? null : element.getAsString());
                break;
            case DATE:
                if (element == null || element.isJsonNull()) {
                    row.setDate(i, null);
                } else {
                    String dtStr = element.getAsString();
                    try {
                        Date dateData = FAST_DATE_FORMAT.parse(dtStr);
                        row.setDate(i, dateData);
                    } catch (ParseException e) {
                        logger.error("failed to parse date data: {}", dtStr, e);
                        throw new DLIException("Failed to parse date data: " + dtStr);
                    }
                }
                break;
            case TIMESTAMP:
                if (element == null || element.isJsonNull()) {
                    row.setTimestamp(i, null);
                } else {
                    String tsStr = element.getAsString();
                    try {
                        // Spark use org.apache.commons.lang3.time.FastDateFormat to format date and timestamp!!!
                        Timestamp tsData = new Timestamp(DEFAULT_FAST_TIMESTAMP_FORMAT.parse(tsStr).getTime());
                        row.setTimestamp(i, tsData);
                    } catch (ParseException e) {
                        try {
                            Timestamp tsData = new Timestamp(FAST_TIMESTAMP_FORMAT.parse(tsStr).getTime());
                            row.setTimestamp(i, tsData);
                        } catch (ParseException e1) {
                            logger.debug("failed to parse timestamp data: {} by FastDateFormat,"
                                + " try to use java.text.DateFormat", tsStr, e);
                            try {
                                Timestamp tsData = new Timestamp(TIMESTAMP_FORMAT.get().parse(tsStr).getTime());
                                row.setTimestamp(i, tsData);
                            } catch (ParseException e2) {
                                logger.error("failed to parse timestamp data: {} by java.text.DateFormat", tsStr, e2);
                                throw new DLIException("Failed to parse timestamp data: " + tsStr);
                            }
                        }
                    }
                }
                break;
            case ARRAY:
                if (element == null || element.isJsonNull()) {
                    row.setArray(i, null);
                } else {
                    row.setArray(i, TypeUtils.makeArray((ArrayTypeInfo) dataTypeInfo, element.toString()));
                }
                break;
            case MAP:
                if (element == null || element.isJsonNull()) {
                    row.setMap(i, null);
                } else {
                    row.setMap(i, TypeUtils.makeMap((MapTypeInfo) dataTypeInfo, element.toString()));
                }
                break;
            case STRUCT:
                if (element == null || element.isJsonNull()) {
                    row.setStruct(i, null);
                } else {
                    row.setStruct(i, TypeUtils.makeStruct((StructTypeInfo) dataTypeInfo, element.toString()));
                }
                break;
            default:
                throw new IllegalArgumentException("Unsupported data type: " + dataType.getName());
        }
    }

    private void reconnectAndSkipHandled() {
        iterator = createIterator(true);
        logger.warn("reconnect, and skip records: {}", readRecordsPerIter);
        // skip all records have read
        long startTime = System.currentTimeMillis();
        for (int i = 0; i < readRecordsPerIter; i++) {
            if (iterator.hasNext()) {
                iterator.next();
            } else {
                throw new IllegalStateException(
                    String.format("Unexpected error, pre read records: %d, but now only read records: %d",
                        readRecordsPerIter, (i + 1)));
            }
        }
        long endTime = System.currentTimeMillis();
        logger.info("skip records cost: {}ms", (endTime - startTime));
    }

    @Override
    public void close() throws IOException {
        if (closed) {
            throw new IllegalStateException("The reader state is closed");
        }
        logger.info("read {} records from pre iterator", readRecordsPerIter);
        obsClientUtils.closeObsClient();
    }

    @Override
    public boolean isClosed() {
        return closed;
    }

    @Override
    public boolean hasNext() {
        for (int times = 0; times < MAX_RECONNECT_NUM; times++) {
            try {
                return (iterator != null && iterator.hasNext()) || (nextIterIdx < totalIter);
            } catch (Exception e) {
                logger.warn("get record failed, may be connection closed by obs, so reconnect it!!!", e);
                reconnectAndSkipHandled();
            }
        }
        throw new IllegalStateException("Should not reach this");
    }
}
