package com.huawei.dli.sdk;

import com.huawei.dli.sdk.common.BadRecordPolicy;
import com.huawei.dli.sdk.common.Column;
import com.huawei.dli.sdk.common.DataType;
import com.huawei.dli.sdk.common.Row;
import com.huawei.dli.sdk.common.utils.ObsClientUtils;
import com.huawei.dli.sdk.exception.DLIException;

import com.obs.services.model.ObsObject;
import com.univocity.parsers.common.ParsingContext;
import com.univocity.parsers.common.ResultIterator;
import com.univocity.parsers.csv.CsvParser;
import com.univocity.parsers.csv.CsvParserSettings;
import com.univocity.parsers.csv.UnescapedQuoteHandling;

import org.apache.commons.lang3.time.FastDateFormat;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.nio.charset.StandardCharsets;
import java.sql.Timestamp;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.TimeZone;

public class ObsReader implements Reader {
    private static final Logger logger = LoggerFactory.getLogger(ObsReader.class);

    private static final FastDateFormat FAST_DATE_FORMAT =
        FastDateFormat.getInstance("yyyy-MM-dd", Locale.US);

    private static final ThreadLocal<DateFormat> TIMESTAMP_FORMAT =
        ThreadLocal.withInitial(() -> new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSSZZ"));

    private static final FastDateFormat FAST_TIMESTAMP_FORMAT =
        FastDateFormat.getInstance("yyyy-MM-dd HH:mm:ss.SSSZZ", TimeZone.getDefault(), Locale.US);

    private static final int MAX_RECONNECT_NUM = 5;

    private final DownloadJob downloadJob;

    private final ObsClientUtils obsClientUtils;

    private final String ak;

    private final String sk;

    private String bucketName;

    private String objectPrefix;

    private List<ObsObject> objects;

    private int totalIter = 0;

    private int nextIterIdx = 0;

    private String nullValue = Row.NULL_VALUE_FOR_CSV;

    private CsvParserSettings parserSettings;

    private ResultIterator<String[], ParsingContext> iterator;

    private int readRecordsPerIter = 0;

    protected boolean closed = false;

    @Deprecated
    public ObsReader(
        DownloadJob downloadJob,
        String endpoint,
        String ak,
        String sk,
        String bucketName,
        String objectPrefix) {
        this.downloadJob = downloadJob;
        this.obsClientUtils = new ObsClientUtils(endpoint);
        this.ak = ak;
        this.sk = sk;
        this.bucketName = bucketName;
        this.objectPrefix = objectPrefix;
    }

    public ObsReader(
        DownloadJob downloadJob,
        String endpoint,
        String ak,
        String sk,
        List<ObsObject> objects) {
        this.downloadJob = downloadJob;
        this.obsClientUtils = new ObsClientUtils(endpoint);
        this.ak = ak;
        this.sk = sk;
        this.objects = objects;
    }

    @Deprecated
    public void setMaxCharsPerColumn(int maxCharsPerColumn) {
        /* do nothing */
    }

    public void setNullValue(String nullValue) {
        this.nullValue = nullValue;
    }

    public void openReader() throws IOException {
        if (objects == null) {
            objects = obsClientUtils.listObjects(ak, sk, bucketName, objectPrefix);
        }
        totalIter = objects.size();
        logger.info("object size: {}", totalIter);

        // See org.apache.spark.sql.execution.datasources.csv.CSVOptions.asParserSettings on Spark
        parserSettings = new CsvParserSettings();
        parserSettings.getFormat().setDelimiter(',');
        parserSettings.getFormat().setQuote('\"');
        parserSettings.getFormat().setQuoteEscape('\\');
        parserSettings.getFormat().setComment('\u0000');
        parserSettings.getFormat().setLineSeparator("\n");
        parserSettings.setIgnoreLeadingWhitespaces(false);
        parserSettings.setIgnoreTrailingWhitespaces(false);
        parserSettings.setReadInputOnSeparateThread(false);
        parserSettings.setSkipEmptyLines(true);
        parserSettings.setNullValue(nullValue);
        parserSettings.setEmptyValue("");
        parserSettings.setMaxCharsPerColumn(-1);
        parserSettings.setUnescapedQuoteHandling(UnescapedQuoteHandling.STOP_AT_DELIMITER);
        iterator = getNextIterator();
    }

    private ResultIterator<String[], ParsingContext> getNextIterator() {
        if (totalIter > 0 && nextIterIdx < totalIter) {
            readRecordsPerIter = 0;
            return createIterator(false);
        } else {
            return null;
        }
    }

    private ResultIterator<String[], ParsingContext> createIterator(boolean recreate) {
        int currIterIdx = recreate ? nextIterIdx - 1 : nextIterIdx++;
        ObsObject object = objects.get(currIterIdx);
        logger.info("{} iterator for object: {}", (recreate ? "recreate" : "create"), object.getObjectKey());
        InputStream inputStream = obsClientUtils.getObjectStream(ak, sk, object.getBucketName(), object.getObjectKey());
        return new CsvParser(parserSettings).iterate(inputStream, StandardCharsets.UTF_8).iterator();
    }

    @Override
    public Row read() throws DLIException, IOException {
        if (closed) {
            throw new IllegalStateException("The reader state is closed");
        }

        if (!iterator.hasNext()) {
            logger.info("read {} records from pre iterator", readRecordsPerIter);
            if (nextIterIdx < totalIter) {
                iterator = getNextIterator();
            }
        }
        if (iterator != null && iterator.hasNext()) {
            String[] values = readWithRetry();
            Row row = downloadJob.newRow();
            updateRow(values, row);
            return row;
        } else {
            return null;
        }
    }

    private String[] readWithRetry() throws DLIException {
        String[] values;
        for (int times = 0; times < MAX_RECONNECT_NUM; times++) {
            try {
                values = iterator.next();
                readRecordsPerIter++;
                return values;
            } catch (Exception e) {
                logger.warn("get record failed, may be connection closed by obs, so reconnect it!!!", e);
                reconnectAndSkipHandled();
            }
        }
        throw new DLIException("Failed to read value with retry");
    }

    private void reconnectAndSkipHandled() throws DLIException {
        iterator = createIterator(true);
        logger.warn("reconnect, and skip records: {}", readRecordsPerIter);
        // skip all records have read
        long startTime = System.currentTimeMillis();
        for (int i = 0; i < readRecordsPerIter; i++) {
            if (iterator.hasNext()) {
                iterator.next();
            } else {
                throw new DLIException(
                    String.format("Unexpected error, pre read records: %d, but now only read records: %d",
                        readRecordsPerIter, (i + 1)));
            }
        }
        long endTime = System.currentTimeMillis();
        logger.info("skip records cost: {}ms", (endTime - startTime));
    }

    private void updateRow(String[] values, Row row) throws DLIException {
        if (values.length != row.getSchema().getColumns().size()) {
            String msg = String.format("Data value size: %s not match column size: %s",
                values.length, row.getSchema().getColumns().size());
            logger.error(msg + ", and data after parse details:");
            for (String data : values) {
                logger.error("\t{}", data);
            }
            if (downloadJob.getBadRecordPolicy() == BadRecordPolicy.SET_NULL) {
                values = downloadJob.emptyRowValues();
            } else {
                throw new DLIException(msg);
            }
        }
        List<Column> columns = row.getSchema().getColumns();
        for (int i = 0; i < columns.size(); i++) {
            Column column = columns.get(i);
            updateCol(row, i, column.getType(), values[i]);
        }
    }

    private void updateCol(Row row, int i, DataType dataType, String data) throws DLIException {
        switch (dataType) {
            case BOOLEAN:
                Boolean boolData = isNull(data) ? null : Boolean.valueOf(data);
                row.setBoolean(i, boolData);
                break;
            case TINYINT:
                Byte byteData = isNull(data) ? null : Byte.valueOf(data);
                row.setTinyint(i, byteData);
                break;
            case SMALLINT:
            case SHORT:
                Short shortData = isNull(data) ? null : Short.valueOf(data);
                row.setShort(i, shortData);
                break;
            case INT:
                Integer intData = isNull(data) ? null : Integer.valueOf(data);
                row.setInt(i, intData);
                break;
            case LONG:
                Long longData = isNull(data) ? null : Long.valueOf(data);
                row.setLong(i, longData);
                break;
            case BIGINT:
                BigInteger bigIntData = isNull(data) ? null : new BigInteger(data);
                row.setBigInteger(i, bigIntData);
                break;
            case FLOAT:
                Float floatData = isNull(data) ? null : Float.valueOf(data);
                row.setFloat(i, floatData);
                break;
            case DOUBLE:
                Double doubleData = isNull(data) ? null : Double.valueOf(data);
                row.setDouble(i, doubleData);
                break;
            case DECIMAL:
                BigDecimal decimalData = isNull(data) ? null : new BigDecimal(data);
                row.setBigDecimal(i, decimalData);
                break;
            case STRING:
                row.setString(i, isNull(data) ? null : data);
                break;
            case DATE:
                try {
                    Date dateData = isNull(data) ? null : FAST_DATE_FORMAT.parse(data);
                    row.setDate(i, dateData);
                } catch (ParseException e) {
                    logger.error("failed to parse date data: {}", data, e);
                    throw new DLIException("Failed to parse date data: " + data);
                }
                break;
            case TIMESTAMP:
                if (isNull(data)) {
                    row.setTimestamp(i, null);
                } else {
                    try {
                        // Spark use org.apache.commons.lang3.time.FastDateFormat to format date and timestamp!!!
                        Timestamp tsData = new Timestamp(FAST_TIMESTAMP_FORMAT.parse(data).getTime());
                        row.setTimestamp(i, tsData);
                    } catch (ParseException e) {
                        logger.debug(
                            "failed to parse timestamp data: {} by FastDateFormat, try to use java.text.DateFormat",
                            data, e);
                        try {
                            Timestamp tsData = new Timestamp(TIMESTAMP_FORMAT.get().parse(data).getTime());
                            row.setTimestamp(i, tsData);
                        } catch (ParseException e1) {
                            logger.error("failed to parse timestamp data: {} by java.text.DateFormat", data, e);
                            throw new DLIException("Failed to parse timestamp data: " + data);
                        }
                    }
                }
                break;
            default:
                throw new IllegalArgumentException("Unsupported data type: " + dataType.getName());
        }
    }

    private boolean isNull(String data) {
        return data == null || data.equals(nullValue);
    }

    @Override
    public void close() throws IOException {
        if (closed) {
            throw new IllegalStateException("The reader state is closed");
        }
        logger.info("read {} records from pre iterator", readRecordsPerIter);
        obsClientUtils.closeObsClient();
    }

    @Override
    public boolean isClosed() {
        return closed;
    }

    public boolean hasNext() {
        return (iterator != null && iterator.hasNext()) || (nextIterIdx < totalIter);
    }
}
