package com.huawei.dli.sdk;

import com.huawei.dli.sdk.common.Column;
import com.huawei.dli.sdk.common.DataBlock;
import com.huawei.dli.sdk.common.DataType;
import com.huawei.dli.sdk.common.Row;
import com.huawei.dli.sdk.common.utils.IterableInputStream;
import com.huawei.dli.sdk.common.utils.ObsClientUtils;
import com.huawei.dli.sdk.exception.DLIException;

import com.google.gson.JsonObject;
import com.obs.services.ObsClient;
import com.obs.services.exception.ObsException;
import com.obs.services.model.AppendObjectRequest;
import com.obs.services.model.AppendObjectResult;

import org.apache.commons.lang3.exception.ExceptionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.sql.Timestamp;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.UUID;

public class ObsWriter extends Writer {
    private static final Logger logger = LoggerFactory.getLogger(ObsWriter.class);

    private static final ThreadLocal<DateFormat> DATE_FORMAT =
        ThreadLocal.withInitial(() -> new SimpleDateFormat("yyyy-MM-dd"));

    private static final ThreadLocal<DateFormat> TIMESTAMP_FORMAT =
        ThreadLocal.withInitial(() -> new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSSZZ"));

    private static final int MIN_DATA_BLOCK_SIZE = 8_388_608; // 8MB

    private static final int MAX_DATA_BLOCK_SIZE = 134_217_728; // 128MB

    private static final int MAX_RECONNECT_NUM = 3;

    private static final long MAX_SIZE_PER_FILE = MAX_DATA_BLOCK_SIZE;

    private final UploadJob uploadJob;

    private final ObsClientUtils obsClientUtils;

    private final String bucketName;

    private final String objectPrefix;

    private final TunnelJob.FileFormat fileFormat;

    private final ArrayList<DataBlock> dataBlocks = new ArrayList<>();

    private AppendObjectRequest appendObjReq;

    private AppendObjectResult appendObjResult;

    private DataBlock dataBlock = null;

    private int dataBlockSize = 33_554_432; // 32MB

    private boolean closed = false;

    public ObsWriter(
        UploadJob uploadJob,
        String endpoint,
        String ak,
        String sk,
        String bucketName,
        String objectPrefix,
        TunnelJob.FileFormat fileFormat) {
        this.uploadJob = uploadJob;
        this.obsClientUtils = new ObsClientUtils(endpoint, ak, sk);
        this.bucketName = bucketName;
        this.objectPrefix = objectPrefix;
        this.fileFormat = fileFormat;
    }

    @Override
    public synchronized void write(Row row) throws DLIException {
        if (closed) {
            throw new IllegalStateException("The writer state is closed");
        }

        String rowStr;
        if (fileFormat == TunnelJob.FileFormat.CSV) {
            rowStr = toCsvString(row);
        } else {
            rowStr = toJsonString(row);
        }
        byte[] rowByteArr = rowStr.getBytes(StandardCharsets.UTF_8);

        if (rowByteArr.length > this.dataBlockSize) {
            logger.info("create new data block as row length: {} exceed data block size: {}",
                rowByteArr.length, this.dataBlockSize);
            DataBlock dataBlockInfo = new DataBlock(rowByteArr.length);
            dataBlockInfo.appendData(rowByteArr);
            this.dataBlocks.add(dataBlockInfo);
            flush();
            return;
        }

        if (this.dataBlock == null) {
            logger.info("create new data block as data block not init");
            this.dataBlock = new DataBlock(this.dataBlockSize);
            this.dataBlocks.add(this.dataBlock);
        }

        try {
            this.dataBlock.appendData(rowByteArr);
        } catch (NegativeArraySizeException e) {
            logger.info("create new data block as current data block is overflows");
            this.dataBlock = new DataBlock(this.dataBlockSize);
            this.dataBlocks.add(this.dataBlock);
            this.dataBlock.appendData(rowByteArr);
            flush();
        }
    }

    private String toCsvString(Row row) {
        StringBuilder strBuilder = new StringBuilder();
        List<Column> columns = row.getSchema().getColumns();
        int size = columns.size();
        boolean needAddLn = true;
        for (int i = 0; i < size; i++) {
            // for static partition column don't write data
            if (uploadJob.staticPartCols.contains(columns.get(i).getColumnName())) {
                continue;
            }
            DataType dataType = columns.get(i).getType();
            Object data = row.getCols().get(i);
            String colStr = getColStrValue(dataType, data);
            if (colStr != null) {
                if (dataType == DataType.STRING) {
                    strBuilder.append("\"").append(colStr).append("\"");
                } else {
                    strBuilder.append(colStr);
                }
            }
            if (colStr == null && dataType == DataType.STRING) {
                strBuilder.append(Row.NULL_VALUE_FOR_CSV);
            }
            if (i != size - 1) {
                strBuilder.append(",");
            } else {
                strBuilder.append("\n");
                needAddLn = false;
            }
        }
        if (needAddLn && strBuilder.toString().endsWith(",")) {
            String str = strBuilder.toString();
            str = str.substring(0, str.length() - 1);
            return str + "\n";
        }
        return strBuilder.toString();
    }

    private String toJsonString(Row row) throws DLIException {
        JsonObject jsonObj = new JsonObject();
        List<Column> columns = row.getSchema().getColumns();
        for (int i = 0; i < columns.size(); i++) {
            // for static partition column don't write data
            if (uploadJob.staticPartCols.contains(columns.get(i).getColumnName())) {
                continue;
            }
            String colName = columns.get(i).getColumnName();
            DataType dataType = columns.get(i).getType();
            switch (dataType) {
                case BOOLEAN:
                    jsonObj.addProperty(colName, row.getBoolean(i));
                    break;
                case TINYINT:
                    jsonObj.addProperty(colName, row.getTinyint(i));
                    break;
                case SMALLINT:
                case SHORT:
                    jsonObj.addProperty(colName, row.getShort(i));
                    break;
                case INT:
                    jsonObj.addProperty(colName, row.getInt(i));
                    break;
                case BIGINT:
                    jsonObj.addProperty(colName, row.getBigInteger(i));
                    break;
                case LONG:
                    jsonObj.addProperty(colName, row.getLong(i));
                    break;
                case FLOAT:
                    jsonObj.addProperty(colName, row.getFloat(i));
                    break;
                case DOUBLE:
                    jsonObj.addProperty(colName, row.getDouble(i));
                    break;
                case DECIMAL:
                    jsonObj.addProperty(colName, row.getBigDecimal(i));
                    break;
                case STRING:
                    // TIPS: For invisible char need encode, so use gson to gen json string.
                    jsonObj.addProperty(colName, row.getString(i));
                    break;
                case DATE:
                case TIMESTAMP:
                    jsonObj.addProperty(colName, getColStrValue(dataType, row.getCols().get(i)));
                    break;
                default:
                    throw new IllegalArgumentException("Unsupported data type: " + dataType.getName());
            }
        }
        return jsonObj.toString() + "\n";
    }

    private String getColStrValue(DataType dataType, Object data) {
        switch (dataType) {
            case BOOLEAN:
            case TINYINT:
            case SMALLINT:
            case SHORT:
            case INT:
            case LONG:
            case BIGINT:
            case FLOAT:
            case DOUBLE:
            case DECIMAL:
                return data == null ? null : String.valueOf(data);
            case STRING:
                return data == null ? null :
                    ((String) data).replaceAll("\\\\", "\\\\\\\\").replaceAll("\"", "\\\\\"");
            case DATE:
                return data == null ? null : DATE_FORMAT.get().format((Date) data);
            case TIMESTAMP:
                return data == null ? null : TIMESTAMP_FORMAT.get().format((Timestamp) data);
            default:
                throw new IllegalArgumentException("Unsupported data type: " + dataType.getName());
        }
    }

    @Override
    public synchronized void flush() throws DLIException {
        if (closed) {
            throw new IllegalStateException("The writer state is closed");
        }
        if (dataBlocks.size() > 1) {
            writeDataBlocks();
        } else {
            logger.info("no need to flush data as not full");
        }
    }

    @Override
    public synchronized void close() throws DLIException {
        if (closed) {
            throw new IllegalStateException("The writer state is closed");
        }

        writeDataBlocks();
        obsClientUtils.closeObsClient();
        closed = true;
    }

    private void writeDataBlocks() {
        if (dataBlocks.isEmpty()) {
            logger.info("there is not data blocks to write");
            return;
        }

        ObsClient obsClient = obsClientUtils.getObsClient();
        Iterator<DataBlock> iterator = dataBlocks.iterator();
        InputStream inputStream = new IterableInputStream<>(iterator, DataBlock::getData, DataBlock::getLength);

        if (appendObjReq == null || appendObjResult == null) {
            String objKey = objectPrefix + String.format("/part-%s.%s",
                UUID.randomUUID().toString(), fileFormat.getName());
            logger.info("flush {} data blocks to {}", dataBlocks.size(), objKey);
            appendObjReq = new AppendObjectRequest();
            appendObjReq.setBucketName(bucketName);
            appendObjReq.setObjectKey(objKey);
            appendObjReq.setInput(inputStream);
            appendObjResult = appendObjectWithRetry(obsClient, appendObjReq);
        } else {
            logger.info("append {} data blocks to {}", dataBlocks.size(), appendObjReq.getObjectKey());
            appendObjReq.setPosition(appendObjResult.getNextPosition());
            appendObjReq.setInput(inputStream);
            appendObjResult = appendObjectWithRetry(obsClient, appendObjReq);
            if (appendObjResult.getNextPosition() > MAX_SIZE_PER_FILE) {
                logger.info("current file size has reached the excepted value");
                appendObjReq = null;
                appendObjResult = null;
            }
        }

        this.dataBlock = null;
        dataBlocks.clear();
    }

    @Override
    public boolean isClosed() {
        return closed;
    }

    void setDataBlockSize(int dataBlockSize) {
        if (dataBlockSize < MIN_DATA_BLOCK_SIZE || dataBlockSize > MAX_DATA_BLOCK_SIZE) {
            throw new IllegalArgumentException("Unexpected data block size, should be ib [8, 128]MB.");
        }
        this.dataBlockSize = dataBlockSize;
    }

    private AppendObjectResult appendObjectWithRetry(ObsClient obsClient, AppendObjectRequest appendObjReq) {
        for (int times = 0; times < MAX_RECONNECT_NUM; times++) {
            try {
                return obsClient.appendObject(appendObjReq);
            } catch (ObsException e) {
                if (ExceptionUtils.getStackTrace(e).contains("Connection reset")
                    || ExceptionUtils.getStackTrace(e).contains("Broken pipe")) {
                    logger.warn("append object failed, may be connection closed by obs, so reconnect it!", e);
                    continue;
                }
                throw e;
            }
        }
        return null;
    }
}
