package com.huawei.dli.sdk;

import com.huawei.dli.sdk.common.Row;
import com.huawei.dli.sdk.common.TableSchema;
import com.huawei.dli.sdk.exception.DLIException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;

public class ResultSet {
    private static final int MAX_PAGE_SIZE = 100;

    private static final Logger LOGGER = LoggerFactory.getLogger(ResultSet.class);

    private List<Row> rows;

    private int localIdx = 0;

    private SQLJob sqlJob;

    private int rowIdx = 0;

    private int currentPage = 1;

    private boolean isClosed = false;

    private Row currentRow;

    public ResultSet(SQLJob sqlJob) {
        this.sqlJob = sqlJob;
    }

    public ResultSet(SQLJob sqlJob, List<Row> rows) {
        this.sqlJob = sqlJob;
        this.rows = rows;
    }

    private void checkClosed() throws DLIException {
        if (isClosed) {
            LOGGER.error("The result set has been closed");
            throw new DLIException("The result set has been closed");
        }
    }

    /**
     * Moves the cursor forward one row from its current position.
     * A <code>ResultSet</code> cursor is initially positioned
     * before the first row; the first call to the method
     * <code>next</code> makes the first row the current row; the
     * second call makes the second row the current row, and so on.
     * <p>
     * When a call to the <code>next</code> method returns <code>false</code>,
     * the cursor is positioned after the last row.
     *
     * @return boolean
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public boolean next() throws DLIException {
        checkClosed();
        // for Query sql, is sync mode, the rows is null, need to get result
        // for DDL/DCL sql, the rows is not null
        if (rows == null) {
            rows = sqlJob.getJobResult(MAX_PAGE_SIZE, currentPage);
            ++currentPage;
        }

        if (rowIdx >= sqlJob.getRowCount()) {
            close();
            return false;
        }

        // get next page
        if (localIdx >= rows.size()) {
            rows = sqlJob.getJobResult(MAX_PAGE_SIZE, currentPage);
            ++currentPage;
            localIdx = 0;
        }

        currentRow = rows.get(localIdx);
        ++localIdx;
        ++rowIdx;
        return true;
    }

    private void close() {
        if (isClosed) {
            return;
        }
        isClosed = true;
    }

    /**
     * Get the current row form ResultSet
     *
     * @return Row
     */
    public Row getRow() {
        return currentRow;
    }

    /**
     * Get row count of the ResultSet
     *
     * @return int
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public int getRowCount() throws DLIException {
        return this.sqlJob.getRowCount();
    }

    /**
     * Get the Schema of the ResultSet
     *
     * @return TableSchema
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public TableSchema getSchema() throws DLIException {
        return this.sqlJob.getResultSchema();
    }
}
