package com.huawei.dli.sdk;

import com.huawei.dli.restapi.ApiException;
import com.huawei.dli.restapi.ApiResponse;
import com.huawei.dli.restapi.model.ExportCommonInfo;
import com.huawei.dli.restapi.model.GlobalErrorResponse;
import com.huawei.dli.restapi.model.JobTagEntityInfo;
import com.huawei.dli.restapi.model.JobResult;
import com.huawei.dli.restapi.model.ResponseJobInfo;
import com.huawei.dli.restapi.model.SQLJobRequest;
import com.huawei.dli.restapi.model.StatusResult;
import com.huawei.dli.sdk.common.Column;
import com.huawei.dli.sdk.common.CompressType;
import com.huawei.dli.sdk.common.DataTypeInfo;
import com.huawei.dli.sdk.common.EnumSyncInfo;
import com.huawei.dli.sdk.common.ExportMode;
import com.huawei.dli.sdk.common.IHttpExecAdpter;
import com.huawei.dli.sdk.common.JobResultInfo;
import com.huawei.dli.sdk.common.JobStatus;
import com.huawei.dli.sdk.common.JobType;
import com.huawei.dli.sdk.common.Row;
import com.huawei.dli.sdk.common.StorageType;
import com.huawei.dli.sdk.common.TableSchema;
import com.huawei.dli.sdk.common.type.TypeInfoUtils;
import com.huawei.dli.sdk.common.type.TypeUtils;
import com.huawei.dli.sdk.exception.DLIException;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.UUID;

public class SQLJob extends Job {

    private static final int MAX_RESULT_COUNT = 1000;

    private static final int MAX_PAGE_SIZE = 1000;

    private static final Logger LOGGER = LoggerFactory.getLogger(SQLJob.class);

    private String sql;

    private String currentdb;

    private int rowCount = -1;

    private TableSchema resultSchema;

    /**
     * Row is used to cache the synchronous return job result when jobType is DDL or DCL.
     */
    private List<Row> rows;

    /**
     * This is the job id of a "exportResult" job.
     * When call the exportResult method to export the query result of a "QUERY" job,
     * it need to save the job id of the "exportResult" job.
     * This job id is used in get exportResult status or cancel exportResult job.
     */
    private String exportJobId;

    private String jobMode;

    public String getJobMode() {
        return jobMode;
    }

    private void setJobMode(String jobMode) {
        this.jobMode = jobMode;
    }


    public SQLJob(Queue queue, String sql) {
        super(queue);
        this.sql = sql;
    }

    public SQLJob(Queue queue, String currentdb, String sql) {
        super(queue);
        this.currentdb = currentdb;
        this.sql = sql;
    }

    public SQLJob(Queue queue, String currentdb, String sql, String transactionId) {
        super(queue);
        this.currentdb = currentdb;
        this.sql = sql;
        this.setTransactionId(transactionId);
    }

    public String getSql() {
        return sql;
    }

    public void setSql(String sql) {
        this.sql = sql;
    }

    public String getCurrentdb() {
        return currentdb;
    }

    /**
     * Submit a SQL job, which returns a <code>ResultSet</code> object.
     *
     * @return ResultSet
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public ResultSet submitQuery() throws DLIException {
        submit();
        // Query SQL is sync mode rows is null.
        // DDL/DCL is async mode row is not null
        if (this.rows == null) {
            return new ResultSet(this);
        } else {
            return new ResultSet(this, rows);
        }
    }

    /**
     * Submit SQL job
     *
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public void submit() throws DLIException {
        if (this.getJobId() != null) {
            LOGGER.error("Job had been submitted, don't submit repeatedly");
            throw new DLIException("Job had been submitted, don't submit repeatedly");
        }

        // if the user set use retry and this sql is query, use retry mode
        submit(true);
        LOGGER.info("Success execute sql job in retry mode, jobId: " + getJobId());
    }

    public void asyncSubmit() throws DLIException {
        if (this.getJobId() != null) {
            LOGGER.error("Job had been submitted, don't submit repeatedly");
            throw new DLIException("Job had been submitted, don't submit repeatedly");
        }
        try {
            submit(false);
            LOGGER.info("AsyncSubmit job succeed, jobId: " + getJobId());
        } catch (Exception e) {
            LOGGER.error("msg", e);
            throw new DLIException(e);
        }
    }

    /**
     * Submit SQL job in retry mode
     *
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    private ResponseJobInfo submit(boolean isWaitResponse) throws DLIException {
        try {
            ResponseJobInfo response = submitSql();
            processAndWaitResponse(response, isWaitResponse);
            return response;
        } catch (InterruptedException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e);
        }
    }

    /**
     * Wait the sql job finish in retry mode
     *
     * @param response response of the submit job of this SQL
     * @throws DLIException         If fail to call the API, e.g. server error or cannot deserialize the response body
     * @throws InterruptedException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    private void processAndWaitResponse(
        ResponseJobInfo response,
        boolean isWait) throws DLIException, InterruptedException {
        JobType jobType = null;
        try {
            jobType = JobType.valueOf(response.getJobType());
        } catch (IllegalArgumentException e) {
            jobType = JobType.UNKNOWN;
        }
        this.setJobType(jobType);
        this.setJobMode(response.getJobMode());
        super.setJobId(response.getJobId());
        if (isWait) {
            checkJobStatusWithRetry();
        }

        /**
         * If jobType is DDL or DCL,when the jobType return synchronous job result,
         * such as:"show databases",get the schema and row list in the response,
         * and stored in member variable rows.
         * You can use getJobResult() get row list.
         */
        boolean hasResult = response.getSchema() != null && !response.getSchema().isEmpty()
            && response.getRows() != null;
        if ((JobType.DDL.equals(jobType) && hasResult) || (JobType.DCL.equals(jobType) && hasResult)) {
            rows = buildRowList(response.getSchema(), response.getRows());
            this.rowCount = rows.size();
        }
    }

    private void checkJobStatusWithRetry() throws DLIException, InterruptedException {
        long startTime = System.currentTimeMillis();
        while (true) {
            JobResultInfo jobResultInfo = getJobStatusWithRetry(this.getJobId());
            if (jobResultInfo == null || checkJobFinish(startTime, jobResultInfo)) {
                return;
            }
        }
    }

    @Override
    protected JobResultInfo getStatusResult(String jobId) throws DLIException {
        if (this.client.getDLIInfo().isUseRetry()) {
            try {
                return getJobStatusWithRetry(jobId);
            } catch (InterruptedException e) {
                LOGGER.error("Thread interrupt", e);
                throw new DLIException(e.getMessage());
            }
        } else {
            return super.getStatusResult(jobId);
        }
    }

    /**
     * Call the get job status rest api with retry
     *
     * @param jobId the job id of the SQL job
     * @return JobResultInfo
     * @throws DLIException         If fail to call the API, e.g. server error or cannot deserialize the response body
     * @throws InterruptedException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    private JobResultInfo getJobStatusWithRetry(String jobId) throws DLIException, InterruptedException {
        StatusResult status = executeWithRetry(
            new IHttpExecAdpter<StatusResult>() {
                @Override
                public ApiResponse<StatusResult> exec() throws DLIException, ApiException {
                    return client.getRestApi().getJobStatusWithHttpInfo(getProjectId(),
                        jobId, client.getAuthToken());
                }
            },
            StatusResult.class,
            (StatusResult res) ->
                new GlobalErrorResponse().errorCode(res.getErrorCode()).errorMsg(res.getErrorMsg()),
            null
        );
        return status != null ? new JobResultInfo(status) : null;
    }

    private ResponseJobInfo submitSql() throws DLIException, InterruptedException {
        SQLJobRequest sqlJobRequest = buildSqlJobRequest();
        String transactionId =
            StringUtils.isNotBlank(getTransactionId()) ? getTransactionId() : UUID.randomUUID().toString();
        if (sqlJobRequest.getConf() == null || !sqlJobRequest.getConf().contains(getTransactionKeyName())) {
            sqlJobRequest.addConfItem(getTransactionKeyName() + "=" + transactionId);
        }
        return executeWithRetry(
            new IHttpExecAdpter<ResponseJobInfo>() {
                @Override
                public ApiResponse<ResponseJobInfo> exec() throws DLIException, ApiException {
                    return client.getRestApi().executeWithHttpInfo(
                        getProjectId(),
                        sqlJobRequest,
                        client.getAuthToken());
                }
            },
            ResponseJobInfo.class,
            (ResponseJobInfo res) ->
                new GlobalErrorResponse().errorCode(res.getErrorCode()).errorMsg(res.getErrorMsg()),
            (ResponseJobInfo dto) -> {
                if (EnumSyncInfo.isSync(dto.getJobMode())) {
                    return "Sync sql already submitted with same transactionId: " + getTransactionId();
                } else {
                    dto.setIsSuccess(true);
                    dto.setErrorMsg(null);
                    dto.setErrorCode(null);
                    return null;
                }
            }
        );
    }

    private SQLJobRequest buildSqlJobRequest() {
        SQLJobRequest sqlJobRequest = new SQLJobRequest();
        sqlJobRequest.setCurrentdb(currentdb);
        sqlJobRequest.setSql(sql);
        sqlJobRequest.setQueueName(getQueue().getQueueName());
        List<Map<String, Object>> confs = getConf();
        if (confs != null) {
            List<String> confList = new ArrayList<String>();
            boolean isAsync = false;
            for (Map<String, Object> conf : confs) {
                for (Map.Entry<String, Object> entry : conf.entrySet()) {
                    Object value = entry.getValue();
                    if ("dli.sql.sqlasync.enabled".equals(entry.getKey())) {
                        value = "true"; // all sql run should be in async
                        isAsync = true;
                    }
                    String confStr = entry.getKey() + "=" + value;
                    confList.add(confStr);
                }
            }
            if (!isAsync) {
                confList.add("dli.sql.sqlasync.enabled=true");
            }
            sqlJobRequest.setConf(confList);
        } else {
            List<String> confList = new ArrayList<>();
            confList.add("dli.sql.sqlasync.enabled=true");
            sqlJobRequest.setConf(confList);
        }
        Map<String, String> tags = getTags();
        if (tags != null) {
            List<JobTagEntityInfo> tagsList = new ArrayList<>();
            for (Map.Entry<String, String> tag : tags.entrySet()) {
                JobTagEntityInfo jobTagEntityInfo = new JobTagEntityInfo();
                jobTagEntityInfo.setKey(tag.getKey());
                jobTagEntityInfo.setValue(tag.getValue());
                tagsList.add(jobTagEntityInfo);
            }
            sqlJobRequest.setTags(tagsList);
        }
        return sqlJobRequest;
    }

    /**
     * If job type is "QUEUE",you can get this job result.
     * And support part of the DDL,DCL jobType.Such as:"show databases","desc tbName"
     *
     * @return List rows
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public List<Row> getJobResult() throws DLIException {
        if ((JobType.DDL.equals(getJobType()) && rows != null)
            || (JobType.DCL.equals(getJobType()) && rows != null)) {
            return rows;
        }
        List<Row> rowsInfo = new ArrayList<Row>();
        for (int i = 1; i <= MAX_RESULT_COUNT / MAX_PAGE_SIZE; i++) {
            List<Row> rowList = getJobResult(MAX_PAGE_SIZE, i);
            rowsInfo.addAll(rowList);
            if (rowList.size() < MAX_PAGE_SIZE) {
                break;
            }
        }
        return rowsInfo;
    }

    /**
     * If job type is "QUEUE",you can get this job result
     * Just get top 1000
     *
     * @return list of Row
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public List<Row> previewJobResult() throws DLIException {
        if (this.getJobId() == null) {
            LOGGER.error("Job ID is null, preview job result failed");
            throw new DLIException("Job ID is null, preview job result failed");
        }

        if (this.client.getDLIInfo().isUseRetry()) {
            try {
                return getJobResultWithRetry(0, 0, true);
            } catch (InterruptedException e) {
                LOGGER.error("msg", e);
                throw new DLIException(e);
            }
        } else {
            return previewJobResultNoRetry();
        }
    }

    /**
     * If job type is "QUEUE",you can get this job result.
     * Only "QUEUE" type job can getJobResult by page.
     *
     * @param pageSize    The maximum number of jobs displayed on each page, in the range: [1, 100].
     * @param currentPage The current page number
     * @return list of Row
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public List<Row> getJobResult(Integer pageSize, Integer currentPage) throws DLIException {
        if (this.getJobId() == null) {
            LOGGER.error("Job ID is null, get job result failed");
            throw new DLIException("Job ID is null, get job result failed");
        }

        if (this.client.getDLIInfo().isUseRetry()) {
            try {
                return getJobResultWithRetry(pageSize, currentPage, false);
            } catch (InterruptedException e) {
                LOGGER.error("msg", e);
                throw new DLIException(e);
            }
        } else {
            return getJobResultNoRetry(pageSize, currentPage);
        }
    }

    private List<Row> getJobResultNoRetry(Integer pageSize, Integer currentPage) throws DLIException {
        try {
            String queueName;
            if (this.getQueue().getQueueName().equals("default")) {
                queueName = null;
            } else {
                queueName = this.getQueue().getQueueName();
            }
            JobResult response = client.getRestApi().getJobResult(getProjectId(), getJobId(),
                client.getAuthToken(), pageSize, currentPage, queueName);
            if (response.getIsSuccess()) {
                this.rowCount = response.getRowCount();
                return buildRowList(response.getSchema(), response.getRows());
            }
            LOGGER.error(response.getMessage());
            throw new DLIException(response.getMessage());
        } catch (ApiException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e.getMessage(), e.getCode(), e.getResponseBody());
        }
    }

    private List<Row> previewJobResultNoRetry() throws DLIException {
        try {
            String queueName;
            if (this.getQueue().getQueueName().equals("default")) {
                queueName = null;
            } else {
                queueName = this.getQueue().getQueueName();
            }
            JobResult response = client.getRestApi().previewJobResult(getProjectId(), getJobId(),
                client.getAuthToken(), queueName);
            if (response.getIsSuccess()) {
                this.rowCount = response.getRowCount();
                return buildRowList(response.getSchema(), response.getRows());
            }
            LOGGER.error(response.getMessage());
            throw new DLIException(response.getMessage());
        } catch (ApiException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e.getMessage(), e.getCode(), e.getResponseBody());
        }
    }

    private List<Row> getJobResultWithRetry(
        Integer pageSize,
        Integer currentPage,
        boolean isPreview) throws DLIException, InterruptedException {
        String queueName;
        if (this.getQueue().getQueueName().equals("default")) {
            queueName = null;
        } else {
            queueName = this.getQueue().getQueueName();
        }

        JobResult result = executeWithRetry(
            new IHttpExecAdpter<JobResult>() {
                @Override
                public ApiResponse<JobResult> exec() throws DLIException, ApiException {
                    if (isPreview) {
                        return client.getRestApi()
                            .previewJobResultWithHttpInfo(getProjectId(), getJobId(), client.getAuthToken(), queueName);
                    } else {
                        return client.getRestApi().getJobResultWithHttpInfo(getProjectId(), getJobId(),
                            client.getAuthToken(), pageSize, currentPage, queueName);
                    }
                }
            },
            JobResult.class,
            (JobResult res) ->
                new GlobalErrorResponse().errorCode(res.getErrorCode()).errorMsg(res.getErrorMsg()),
            null
        );
        if (result != null && result.getIsSuccess()) {
            this.rowCount = result.getRowCount();
            return buildRowList(result.getSchema(), result.getRows());
        }
        return null;
    }

    /**
     * Convert the List<Map<String, String>> schemas from the response to List<Column>.
     *
     * @param schemas
     * @return
     */
    private List<Column> getColumns(List<Map<String, String>> schemas) {
        List<Column> columns = new ArrayList<Column>();
        for (Map<String, String> schema : schemas) {
            for (Map.Entry<String, String> entry : schema.entrySet()) {
                String columnName = entry.getKey();
                String columnType = entry.getValue();
                DataTypeInfo dataTypeInfo = TypeInfoUtils.getDataTypeInfo(columnType);
                Column column = new Column(columnName, dataTypeInfo);
                columns.add(column);
            }
        }
        return columns;
    }

    /**
     * Convert the List<Map<String, String>> schemaResponse,List<List<String>> rowsResponse
     * from the response to List<Row>.
     *
     * @param schemaResponse schema resp
     * @param rowsResponse   row resp
     * @return list of Row
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    private List<Row> buildRowList(List<Map<String, String>> schemaResponse,
        List<List<String>> rowsResponse) throws DLIException {
        List<Row> rowsInfo = new ArrayList<Row>();
        TableSchema tableSchema = new TableSchema();
        List<Column> columns = getColumns(schemaResponse);
        tableSchema.setColumns(columns);
        for (List<String> rowResponse : rowsResponse) {
            Row row = new Row(tableSchema);
            int index = 0;
            List<Object> cols = new ArrayList<Object>();
            for (String colStr : rowResponse) {
                DataTypeInfo typeInfo = columns.get(index).getDataTypeInfo();
                if (colStr != null) {
                    Object dataObj = TypeUtils.transform(colStr, typeInfo);
                    cols.add(index, dataObj);
                } else {
                    cols.add(index, null);
                }
                index++;
            }
            row.setCols(cols);
            rowsInfo.add(row);
        }
        this.resultSchema = tableSchema;
        return rowsInfo;
    }

    /**
     * Export QUEUE type job result.
     *
     * @param dataPath     data path
     * @param storageType  StorageType
     * @param compressType CompressType
     * @param exportMode   ExportMode
     * @param queueName    queue name
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public void exportResult(String dataPath, StorageType storageType, CompressType compressType,
        ExportMode exportMode, String queueName) throws DLIException {
        exportResult(dataPath, storageType, compressType, exportMode, queueName, false, 0);
    }

    /**
     * @param dataPath         data path
     * @param storageType      StorageType
     * @param compressType     CompressType
     * @param exportMode       ExportMode
     * @param queueName        queue name
     * @param withColumnHeader if with col header
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public void exportResult(String dataPath, StorageType storageType, CompressType compressType,
        ExportMode exportMode, String queueName, boolean withColumnHeader) throws DLIException {
        exportResult(dataPath, storageType, compressType, exportMode, queueName, withColumnHeader, 0, "utf-8");
    }

    /**
     * @param dataPath         data path
     * @param storageType      StorageType
     * @param compressType     CompressType
     * @param exportMode       ExportMode
     * @param queueName        queue name
     * @param withColumnHeader if with col header
     * @param limitNum         limit num
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public void exportResult(String dataPath, StorageType storageType, CompressType compressType,
        ExportMode exportMode, String queueName, boolean withColumnHeader, Integer limitNum)
        throws DLIException {
        asyncExportResult(dataPath, storageType, compressType,
            exportMode, queueName, withColumnHeader, limitNum, "utf-8");
        super.checkJobRunningStatus(exportJobId);
    }

    /**
     * @param dataPath         data path
     * @param storageType      StorageType
     * @param compressType     CompressType
     * @param exportMode       ExportMode
     * @param queueName        queue name
     * @param withColumnHeader if with col header
     * @param limitNum         limit num
     * @param encodingType     encodingType
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public void exportResult(String dataPath, StorageType storageType, CompressType compressType,
        ExportMode exportMode, String queueName, boolean withColumnHeader,
        Integer limitNum, String encodingType)
        throws DLIException {
        asyncExportResult(dataPath, storageType, compressType,
            exportMode, queueName, withColumnHeader, limitNum, encodingType);
        super.checkJobRunningStatus(exportJobId);
    }

    String asyncExportResult(String dataPath, StorageType storageType, CompressType compressType,
        ExportMode exportMode, String queueName, boolean withColumnHeader, Integer limitNum, String encodingType)
        throws DLIException {
        if (getJobId() == null) {
            LOGGER.error("Job ID is null,export result failed");
            throw new DLIException("Job ID is null,export result failed");
        }
        ExportCommonInfo exportInfo = new ExportCommonInfo();
        exportInfo.setDataPath(dataPath);
        exportInfo.setCompress(compressType.getName());
        exportInfo.setDataType(storageType.getName());
        exportInfo.setExportMode(exportMode.getName());
        exportInfo.setWithColumnHeader(withColumnHeader);
        exportInfo.setEncodingType(encodingType);
        limitNum = limitNum == null ? 0 : limitNum;
        exportInfo.setLimitNum(limitNum);
        if (!"default".equals(queueName)) {
            exportInfo.setQueueName(queueName);
        }
        ResponseJobInfo response = null;
        try {
            response = client.getRestApi()
                .storeJobResult(getProjectId(), getJobId(), client.getAuthToken(), exportInfo);
            if (!response.getIsSuccess()) {
                LOGGER.error(response.getMessage());
                throw new DLIException(response.getMessage());
            }
            exportJobId = response.getJobId();
            return exportJobId;
        } catch (ApiException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e.getMessage(), e.getCode(), e.getResponseBody());
        }
    }

    /**
     * Export QUEUE type job result with a iterator
     *
     * @return Reader
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public Reader exportResult() throws DLIException {
        return exportResult(null);
    }

    public Reader exportResult(Integer limit) throws DLIException {
        DownloadJob downloadJob = new DownloadJob(getQueue(), this, limit);
        checkJobRunningStatus(exportJobId);
        downloadJob.getDownloadStatus(); // to trigger tunnel ready
        return downloadJob.createReader();
    }

    /**
     * If execute exportResult,you can get exportJob status.
     *
     * @return JobStatus:LAUNCHING,FINISHED...
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public JobStatus getExportJobStatus() throws DLIException {
        if (null == exportJobId) {
            LOGGER.error("No export job in progress, get export status failed");
            throw new DLIException("No export job in progress, get export status failed");
        }
        JobResultInfo jobResultInfo = super.getStatusResult(exportJobId);
        return jobResultInfo.getJobStatus();
    }

    /**
     * If exportResult is RUNNING or LAUNCHING,you can cancel it.
     *
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public void cancelExportJob() throws DLIException {
        if (exportJobId == null) {
            LOGGER.error("No export job in progress,cancel failed");
            throw new DLIException("No export job in progress,cancel failed");
        }
        JobStatus status = getExportJobStatus();
        if (JobStatus.FINISHED.equals(status) || JobStatus.FAILED.equals(status)) {
            LOGGER.error("The job has finished,can not be canceled");
            throw new DLIException("The job has finished,can not be canceled");
        }
        if (JobStatus.CANCELLED.equals(status)) {
            LOGGER.error("The job has canceled");
            throw new DLIException("The job has canceled");
        }
        client.cancelJob(exportJobId);
        LOGGER.info("Cancel export job succeed, exportJobId: " + exportJobId);
    }

    public int getRowCount() throws DLIException {
        if (rowCount == -1) {
            getJobResult(MAX_PAGE_SIZE, 1);
        }
        return rowCount;
    }

    public TableSchema getResultSchema() throws DLIException {
        if (resultSchema == null) {
            getJobResult(MAX_PAGE_SIZE, 1);
        }
        return resultSchema;
    }

    public ResultSet getJobResultSet() {
        if (this.rows == null) {
            return new ResultSet(this);
        } else {
            return new ResultSet(this, rows);
        }
    }
}
