package com.huawei.dli.sdk;

import com.huawei.dli.sdk.common.SparkJobStatus;
import com.huawei.dli.sdk.exception.DLIException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;

public abstract class SparkJob extends BaseJob {
    private static final long JOB_TIMEOUT_SECOND = -1;

    private static final Logger LOGGER = LoggerFactory.getLogger(SparkJob.class);

    private String jobId;

    private Cluster cluster;

    private long jobTimeout = JOB_TIMEOUT_SECOND;

    public SparkJob(Cluster cluster) {
        super(cluster.getClient());
        this.cluster = cluster;
    }

    public SparkJob(Cluster cluster, String transactionId) {
        super(cluster.getClient(), transactionId);
        this.cluster = cluster;
    }

    public String getJobId() {
        return jobId;
    }

    // package access, only can set by self.
    void setJobId(String jobId) {
        this.jobId = jobId;
    }

    public String getProjectId() {
        return client.getAuthentication().getProjectId();
    }

    public Cluster getCluster() {
        return cluster;
    }

    public long getJobTimeout() {
        return jobTimeout;
    }

    public void setJobTimeout(int jobTimeout) {
        this.jobTimeout = jobTimeout;
    }

    /**
     * Submit a Spark job. If timeout is -1, not check running status.
     *
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public abstract void submit() throws DLIException;

    /**
     * Submit a Spark job in async.
     *
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public abstract void asyncSubmit() throws DLIException;

    /**
     * Get a Spark job status, it may be starting, running, dead or success.
     *
     * @return SparkJobStatus
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public abstract SparkJobStatus getStatus() throws DLIException;

    /**
     * Get a Spark job client log when job is finished. If not finished will throw an exception.
     *
     * @param from Start line of the log to be displayed. By default, the last 100 lines of the log
     *             are displayed. If a log file contains fewer than 100 lines, line 0 is the start line.
     * @param size Number of obtained logs.
     * @return the logs
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public abstract List<String> getLog(int from, int size) throws DLIException;

    public abstract List<String> getLogByJobId(String jobId, int from, int size) throws DLIException;

    /**
     * Get a Spark job driver log when job is finished. If not finished will throw an exception.
     *
     * @param from  Start line of the log to be displayed. By default, the last 100 lines of the log
     *              are displayed. If a log file contains fewer than 100 lines, line 0 is the start line.
     * @param size  Number of obtained logs.
     * @param index When a submitted job is retried, multiple driver logs are generated.
     *              This parameter specifies the index number of the specified driver log.
     * @return the logs
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public abstract List<String> getDriverLog(int from, int size, int index) throws DLIException;

    public abstract List<String> getDriverLogByJobId(String jobId, int from, int size, int index) throws DLIException;

    protected void checkJobRunningState() throws DLIException {
        long startTime = System.currentTimeMillis();
        while (true) {
            SparkJobStatus jobStatus = getStatus();
            if (SparkJobStatus.SUCCESS.equals(jobStatus)) {
                return;
            }
            if (SparkJobStatus.DEAD.equals(jobStatus)) {
                LOGGER.error("The session has already exited");
                throw new DLIException("The session has already exited");
            }
            try {
                Thread.sleep(1000);
            } catch (InterruptedException e) {
                LOGGER.error("msg", e);
            }
            long endTime = System.currentTimeMillis();
            long timeout = endTime - startTime;
            if (getJobTimeout() != -1 && timeout >= getJobTimeout() * 1000) {
                LOGGER.error("Timed out for Job " + jobStatus + " status: " + timeout + "ms."
                    + "The Job id is " + getJobId());
                throw new DLIException("Timed out for Job " + jobStatus + " status: " + timeout + "ms."
                    + "The Job id is " + getJobId());
            }
        }
    }

    SparkJobStatus convertSessionState(String state) {
        if (state.equalsIgnoreCase("not_started")) {
            return SparkJobStatus.NOT_STARTED;
        }
        if (state.equalsIgnoreCase("starting")) {
            return SparkJobStatus.STARTING;
        }
        if (state.equalsIgnoreCase("idle")) {
            return SparkJobStatus.IDLE;
        }
        if (state.equalsIgnoreCase("busy")) {
            return SparkJobStatus.BUSY;
        }
        if (state.equalsIgnoreCase("running")) {
            return SparkJobStatus.RUNNING;
        }
        if (state.equalsIgnoreCase("shutting_down")) {
            return SparkJobStatus.SHUTTING_DOWN;
        }
        if (state.equalsIgnoreCase("error")) {
            return SparkJobStatus.ERROR;
        }
        if (state.equalsIgnoreCase("dead")) {
            return SparkJobStatus.DEAD;
        }
        if (state.equalsIgnoreCase("success")) {
            return SparkJobStatus.SUCCESS;
        }
        // Actually it can't reach this, or else it may be a huge bug.
        return null;
    }
}
