package com.huawei.dli.sdk;

import com.huawei.dli.restapi.ApiException;
import com.huawei.dli.restapi.model.PartitionInfo;
import com.huawei.dli.restapi.model.PartitionListInfo;
import com.huawei.dli.restapi.model.PartitionResult;
import com.huawei.dli.restapi.model.ResponseInfo;
import com.huawei.dli.sdk.common.PartitionSchema;
import com.huawei.dli.sdk.common.TableSchema;
import com.huawei.dli.sdk.common.TableType;
import com.huawei.dli.sdk.exception.DLIException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;

public class Table {

    private static final Logger LOGGER = LoggerFactory.getLogger(Table.class);

    private Database db;

    private String tableName;

    private String owner;

    private long createTime;

    private long lastAccessTime;

    private String location;

    private String dataLocation;

    private long tableSize;

    private TableType tableType;

    private String dataType;

    private List<String> partitionColumns;

    public Table(Database db, String tableName, String dataLocation, TableType tableType) {
        this.db = db;
        this.tableName = tableName;
        this.dataLocation = dataLocation;
        this.tableType = tableType;
    }

    @SuppressWarnings("checkstyle:ParameterNumber")
    public Table(Database db, String tableName, String owner, long createTime, long lastAccessTime,
        String location, String dataLocation, long tableSize,
        TableType tableType, List<String> partitionColumns, String dataType) {
        this.db = db;
        this.tableName = tableName;
        this.owner = owner;
        this.createTime = createTime;
        this.lastAccessTime = lastAccessTime;
        this.location = location;
        this.dataLocation = dataLocation;
        this.tableSize = tableSize;
        this.tableType = tableType;
        this.dataType = dataType;
        this.partitionColumns = partitionColumns;
    }

    public Table() {

    }

    public String getTableName() {
        return tableName;
    }

    public Database getDb() {
        return db;

    }

    public String getOwner() {
        return owner;
    }

    public long getCreateTime() {
        return createTime;
    }

    public long getLastAccessTime() {
        return lastAccessTime;
    }

    public String getLocation() {
        return location;
    }

    public String getDataLocation() {
        return dataLocation;
    }

    public long getTableSize() {
        return tableSize;
    }

    public TableType getTableType() {
        return tableType;
    }

    public String getDataType() {
        return dataType;
    }

    public List<String> getPartitionColumns() {
        return partitionColumns;
    }

    /**
     * Get describes the metadata information for the specified table.
     *
     * @return TableSchema
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public TableSchema getTableDetail() throws DLIException {
        return db.getTableDetail(this.tableName);
    }

    /**
     * Get describes the partition information for the table in specified offset and limit.
     *
     * @param offset get offset
     * @param limit  get limit
     * @return PartitionSchema
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public PartitionSchema getPartitionDetail(Integer offset, Integer limit) throws DLIException {
        PartitionSchema partitionSchema = new PartitionSchema();
        List<PartitionInfo> partitionList = new ArrayList<>();
        PartitionResult response = db.getClient().showPartitions(db.getDatabaseName(), tableName, offset, limit);
        if (response.getIsSuccess()) {
            PartitionListInfo partitions = response.getPartitions();
            partitionSchema.setTotalCount(partitions.getTotalCount());
            List<PartitionInfo> partitonInfos = partitions.getPartitionInfos();
            for (PartitionInfo pI : partitonInfos) {
                PartitionInfo partitionInfo = new PartitionInfo();
                partitionInfo.setPartitionName(pI.getPartitionName());
                partitionInfo.setCreateTime(pI.getCreateTime());
                partitionInfo.setLastAccessTime(pI.getLastAccessTime());
                partitionInfo.setLocations(pI.getLocations());
                partitionInfo.setLastDdlTime(pI.getLastDdlTime());
                partitionInfo.setTotalSize(pI.getTotalSize());
                partitionInfo.setNumRows(pI.getNumRows());
                partitionInfo.setNumFiles(pI.getNumFiles());
                partitionList.add(partitionInfo);
            }
            partitionSchema.setPartitionInfos(partitionList);
        }
        return partitionSchema;
    }

    /**
     * Get describes the partition information for the specified table.
     *
     * @return PartitionSchema
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public PartitionSchema getPartitions() throws DLIException {
        PartitionSchema partitionSchema = new PartitionSchema();
        try {
            // 此处取limit=1条数据是为了获取总的分区数，具体的分区信息通过多线程批量处理
            PartitionResult response = db.getClient()
                .getRestApi()
                .showPartitions(db.getProjectId(), db.getDatabaseName(),
                    tableName, db.getClient().getAuthToken(), 0, 1);
            if (response.getIsSuccess()) {
                Integer totalCount = response.getPartitions().getTotalCount();
                partitionSchema.setTotalCount(totalCount);
                List<PartitionInfo> partitionInfos = this.getPartitionDetail(0, totalCount).getPartitionInfos();
                partitionSchema.setPartitionInfos(partitionInfos);
                return partitionSchema;
            }
            LOGGER.error(response.getMessage());
            throw new DLIException(response.getMessage());
        } catch (ApiException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e.getMessage(), e.getCode(), e.getResponseBody());
        }
    }

    /**
     * Delete current table.
     *
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public void deleteTable() throws DLIException {
        if (db == null) {
            LOGGER.error("Database is null");
            throw new DLIException("Database is null");
        }
        try {
            ResponseInfo response = db.getClient().getRestApi().deleteTable(db.getProjectId(), db.getDatabaseName(),
                tableName, db.getClient().getAuthToken());
            if (!response.getIsSuccess()) {
                LOGGER.error(response.getMessage());
                throw new DLIException(response.getMessage());
            }
        } catch (ApiException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e.getMessage(), e.getCode(), e.getResponseBody());
        }

    }

    @Override
    public String toString() {
        return "Table{"
            + "tableName='" + tableName + '\''
            + ", owner='" + owner + '\''
            + ", createTime=" + createTime
            + ", lastAccessTime=" + lastAccessTime
            + ", location='" + location + '\''
            + ", dataLocation='" + dataLocation + '\''
            + ", tableSize=" + tableSize
            + ", tableType=" + tableType
            + ", partitionColumns=" + partitionColumns
            + '}';
    }
}
