package com.huawei.dli.sdk;

import com.huawei.dli.restapi.ApiException;
import com.huawei.dli.restapi.api.RestApi;
import com.huawei.dli.restapi.model.DownloadCreateJobResult;
import com.huawei.dli.restapi.model.EmptyBody;
import com.huawei.dli.sdk.authentication.Authentication;
import com.huawei.dli.sdk.common.Column;
import com.huawei.dli.sdk.common.DLIInfo;
import com.huawei.dli.sdk.common.PartitionSpec;
import com.huawei.dli.sdk.common.Row;
import com.huawei.dli.sdk.common.TableSchema;
import com.huawei.dli.sdk.exception.DLIException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;

public abstract class TunnelJob {
    private static final Logger logger = LoggerFactory.getLogger(TunnelJob.class);

    protected String bucketName;

    protected String objectPrefix;

    protected TableSchema tblSchema;

    protected List<Column> partCols;

    protected String nullValue = Row.NULL_VALUE_FOR_CSV;

    protected void prepare(Queue queue, String dbName, String tblName, String opType)
        throws DLIException {
        DLIClient client = queue.getClient();
        Database db = client.getDatabase(dbName);
        Table table = db.getTable(tblName);
        tblSchema = table.getTableDetail();
        partCols = tblSchema.getColumns().stream()
            .filter(Column::isPartitionColumn)
            .collect(Collectors.toList());
        bucketName = createTunnel(client, opType);
        objectPrefix = String.format("%s/%s/%s/%s",
            opType, table.getDb().getDatabaseName(), table.getTableName(), UUID.randomUUID().toString());
    }

    protected void prepare(Queue queue, TableSchema tblSchema) throws DLIException {
        this.tblSchema = tblSchema;
        partCols = tblSchema.getColumns().stream()
            .filter(Column::isPartitionColumn)
            .collect(Collectors.toList());
        bucketName = createTunnel(queue.getClient(), "export");
        objectPrefix = String.format("%s/exportDb/exportTbl/%s", "export", UUID.randomUUID().toString());
    }

    private String createTunnel(DLIClient dliClient, String opType) throws DLIException {
        Authentication tokenAuth = dliClient.getAuthentication();
        RestApi restApi = tokenAuth.getRestApi();
        DLIInfo info = dliClient.getDLIInfo();
        try {
            EmptyBody emptyBody = new EmptyBody();
            DownloadCreateJobResult result =
                restApi.createDownloadJob(info.getProjectId(), emptyBody, tokenAuth.getAuthToken());
            if (!result.getIsSuccess()) {
                logger.error(result.getMessage());
                throw new DLIException(result.getMessage());
            }
            return result.getBucketName();
        } catch (ApiException e) {
            logger.error("Failed to create tunnel for {} while call REST api", opType);
            throw new DLIException("Failed to create tunnel for " + opType + ", " + e.getMessage());
        }
    }

    protected boolean validPartSpec(PartitionSpec partSpec) {
        Set<String> partMetaKeys = partCols.stream().map(Column::getColumnName).collect(Collectors.toSet());
        Set<String> usedPartKeys = partSpec.keys();
        for (String key : usedPartKeys) {
            if (!partMetaKeys.contains(key)) {
                return false;
            }
        }
        return true;
    }

    public Row newRow() {
        return new Row(this.tblSchema);
    }

    public enum FileFormat {
        CSV("csv"), JSON("json"), AVRO("avro");

        private final String name;

        FileFormat(String name) {
            this.name = name;
        }

        public String getName() {
            return name;
        }
    }
}
