package com.huawei.dli.sdk;

import com.huawei.dli.sdk.common.Column;
import com.huawei.dli.sdk.common.DLIInfo;
import com.huawei.dli.sdk.common.JobResultInfo;
import com.huawei.dli.sdk.common.JobStatus;
import com.huawei.dli.sdk.common.PartitionSpec;
import com.huawei.dli.sdk.exception.DLIException;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringJoiner;

public class UploadJob extends TunnelJob {
    private static final Logger logger = LoggerFactory.getLogger(UploadJob.class);

    private static final int WRITER_BLOCK_SIZE = 32 * 1024 * 1024; // 32MB

    private final SQLJob job;

    private final FileFormat fileFormat;

    private boolean committed = false;

    private final List<Writer> writers = Collections.synchronizedList(new ArrayList<>());

    private int writerDataBlockSize = WRITER_BLOCK_SIZE;

    final Set<String> staticPartCols = new HashSet<>();

    @Deprecated
    public UploadJob(Queue queue, String dbName, String tableName) throws DLIException {
        this(queue, dbName, tableName, null, FileFormat.CSV);
    }

    /**
     * Create a upload job to upload data to DLI.
     * <p>
     * For overwrite it will keep same behaviors with Hive,
     * see https://support.huaweicloud.com/sqlref-spark-dli/dli_08_0095.html.
     *
     * @param queue       the queue to running the upload spark job
     * @param dbName      the database name which the data to upload
     * @param tableName   the table name which the data to upload
     * @param isOverwrite if use overwrite when upload data
     * @throws DLIException throw exception if create upload job failed
     */
    public UploadJob(Queue queue, String dbName, String tableName, Boolean isOverwrite) throws DLIException {
        this(queue, dbName, tableName, isOverwrite, FileFormat.CSV);
    }

    private UploadJob(Queue queue, String dbName, String tableName, Boolean isOverwrite, FileFormat fileFormat)
        throws DLIException {
        prepare(queue, dbName, tableName, "upload");
        String overwrite = "";
        if (isOverwrite != null && isOverwrite) {
            if (partCols.size() != 0) {
                throw new DLIException("Can not overwrite full table on a partitioned table, please check.");
            }
            overwrite = "OVERWRITE";
        }
        String importSql = String.format("LOAD DATA INPATH 'obs://%s/%s' %s INTO TABLE `%s`.`%s` "
                + "OPTIONS(data_type '%s', multiLine 'true', nullValue '%s') ",
            bucketName, objectPrefix, overwrite, dbName, tableName, fileFormat.getName(), nullValue);
        this.job = new SQLJob(queue, importSql);
        Map<String, Object> conf = new HashMap<>();
        conf.put("spark.sql.datasource.csv.treatEmptyStringAsNull", "false");
        conf.put("dli.sql.dynamicPartitionOverwrite.enabled", "true");
        List<Map<String, Object>> confList = new ArrayList<>();
        confList.add(conf);
        this.job.setConf(confList);
        this.fileFormat = fileFormat;
    }

    public UploadJob(Queue queue, String dbName, String tableName, String partitionSpec, Boolean isOverwrite)
        throws DLIException {
        this(queue, dbName, tableName, new PartitionSpec(partitionSpec), isOverwrite);
    }

    /**
     * Create a upload job to upload data to DLI.
     * If you specify the static partition, then this static partition data will be ignore.
     * For example: if have 3 column with 2 partition column(c1, pc1, pc2),
     * the partition spec is pc1=b1, and data value is a, b2, c, after upload the data is a, b1, c.
     * <p>
     * For overwrite it will keep same behaviors with Hive,
     * see https://support.huaweicloud.com/sqlref-spark-dli/dli_08_0095.html.
     *
     * @param queue         the queue to running the upload spark job
     * @param dbName        the database name which the data to upload
     * @param tableName     the table name which the data to upload
     * @param partitionSpec the partition spec for a partitioned table
     * @param isOverwrite   if use overwrite when upload data
     * @throws DLIException throw exception if create upload job failed
     */
    public UploadJob(Queue queue, String dbName, String tableName, PartitionSpec partitionSpec, Boolean isOverwrite)
        throws DLIException {
        this(queue, dbName, tableName, partitionSpec, isOverwrite, FileFormat.CSV);
    }

    private UploadJob(Queue queue, String dbName, String tableName, PartitionSpec partitionSpec, Boolean isOverwrite,
        FileFormat fileFormat) throws DLIException {
        prepare(queue, dbName, tableName, "upload");
        if (partCols.size() == 0) {
            throw new DLIException("It is not a partitioned table, please check.");
        }
        if (partitionSpec == null || !validPartSpec(partitionSpec)) {
            throw new DLIException("Invalid partition spec or there are unexpected partition column, please check.");
        }
        String overwrite = "";
        if (isOverwrite != null && isOverwrite) {
            overwrite = "OVERWRITE";
        }
        int inPartKeySize = partitionSpec.keys().size();
        StringJoiner partSpecClause = new StringJoiner(", ");
        int i = 0;
        for (Column partCol : partCols) {
            if (i < inPartKeySize) {
                String partValue = partitionSpec.get(partCol.getColumnName());
                if (partValue == null) {
                    throw new DLIException("Unexpected partition spec, please check");
                }
                partSpecClause.add(String.format("`%s`='%s'", partCol.getColumnName(), partValue));
                staticPartCols.add(partCol.getColumnName());
            }
            i++;
        }
        String importSql = String.format("LOAD DATA INPATH 'obs://%s/%s' %s INTO TABLE `%s`.`%s` PARTITION (%s) "
                + "OPTIONS(data_type '%s', multiLine 'true', nullValue '%s') ",
            bucketName, objectPrefix, overwrite, dbName, tableName, partSpecClause.toString(),
            fileFormat.getName(), nullValue);
        this.job = new SQLJob(queue, importSql);
        Map<String, Object> conf = new HashMap<>();
        conf.put("spark.sql.datasource.csv.treatEmptyStringAsNull", "false");
        conf.put("dli.sql.dynamicPartitionOverwrite.enabled", "true");
        List<Map<String, Object>> confList = new ArrayList<>();
        confList.add(conf);
        this.job.setConf(confList);
        this.fileFormat = fileFormat;
    }

    public int getWriterDataBlockSize() {
        return writerDataBlockSize;
    }

    public void setWriterDataBlockSize(int writerDataBlockSize) {
        this.writerDataBlockSize = writerDataBlockSize;
    }

    public void resetCommitted() {
        this.committed = false;
        this.job.setJobId(null);
    }

    /* This is a single-thread interface function */
    public synchronized Writer createWriter() throws DLIException {
        if (committed) {
            throw new DLIException("The job is committed");
        }

        DLIInfo dliInfo = this.job.getQueue().getClient().getDLIInfo();
        Writer writer;
        switch (fileFormat) {
            case CSV:
            case JSON:
                writer = new ObsWriter(
                    this,
                    dliInfo.getObsEndPoint(),
                    dliInfo.getAccessKey(),
                    dliInfo.getSecretKey(),
                    bucketName,
                    objectPrefix,
                    fileFormat);
                ((ObsWriter) writer).setDataBlockSize(this.writerDataBlockSize);
                break;
            case AVRO:
                writer = new AvroObsWriter(
                    this,
                    dliInfo.getObsEndPoint(),
                    dliInfo.getAccessKey(),
                    dliInfo.getSecretKey(),
                    bucketName,
                    objectPrefix);
                break;
            default:
                throw new DLIException("Unexpected inter file format: " + fileFormat.getName());
        }
        writers.add(writer);
        return writer;
    }

    /* This is a single-thread interface function */
    public synchronized void beginCommit() throws DLIException {
        this.beginCommit(null);
    }

    /* This is a single-thread interface function */
    public synchronized void beginCommit(String transactionId) throws DLIException {
        if (committed) {
            String msg = "The job is committed";
            logger.error(msg);
            throw new DLIException(msg);
        }

        if (this.writers.size() == 0) {
            String msg = "There is not writer create before, please check.";
            logger.error(msg);
            throw new DLIException(msg);
        }

        /* check flush status for writers */
        for (Writer writer : this.writers) {
            if (!writer.isClosed()) {
                String msg = "There is writer has not been closed.";
                logger.error(msg);
                throw new DLIException(msg);
            }
        }

        this.committed = true;
        if (StringUtils.isNotBlank(transactionId)) {
            this.job.setTransactionId(transactionId);
        }
        this.job.asyncSubmit();
    }

    public synchronized String getInternalJobId() throws DLIException {
        if (job != null) {
            return job.getJobId();
        }
        throw new DLIException("There is not job have been created, please check.");
    }

    @Deprecated
    public synchronized JobStatus getCommitStatus() throws DLIException {
        return JobStatus.fromName(getUploadStatus());
    }

    public synchronized String getUploadStatus() throws DLIException {
        if (job != null) {
            JobStatus status = job.getStatus();
            logger.info("Upload job internal status: {}", status.getName());
            return status.getName();
        }
        throw new DLIException("There is not job have been created, please check.");
    }

    public long getUploadResultCount() throws DLIException {
        if (job != null) {
            JobResultInfo jobResultInfo = job.getStatusResult(job.getJobId());
            if (jobResultInfo != null) {
                return jobResultInfo.getResultCount();
            }
            throw new DLIException("There is no result of job, please check job:" + job.getJobId());
        }
        throw new DLIException("There is not job have been created, please check.");
    }
}
