package com.huawei.dli.sdk.authentication;

import com.huawei.dli.restapi.ApiClient;
import com.huawei.dli.restapi.ApiException;
import com.huawei.dli.restapi.ApiResponse;
import com.huawei.dli.restapi.api.RestApi;
import com.huawei.dli.restapi.api.TokenAuthApi;
import com.huawei.dli.restapi.model.Auth;
import com.huawei.dli.restapi.model.Domain;
import com.huawei.dli.restapi.model.Identity;
import com.huawei.dli.restapi.model.Password;
import com.huawei.dli.restapi.model.Project;
import com.huawei.dli.restapi.model.Scope;
import com.huawei.dli.restapi.model.TokenAuthInfo;
import com.huawei.dli.restapi.model.TokenAuthResponse;
import com.huawei.dli.restapi.model.User;
import com.huawei.dli.sdk.common.DLIInfo;
import com.huawei.dli.sdk.exception.DLIException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.net.InetSocketAddress;
import java.net.Proxy;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.TimeZone;
import java.util.concurrent.TimeUnit;

import okhttp3.OkHttpClient;

public class Authentication {

    private static final String PROTOCOL = "https";

    private static final long DEFAULT_CONNECT_TIMEOUT = 300;

    private static final long DEFAULT_READ_TIMEOUT = 300;

    private static final String SERVICE_VERSION = "v1.0";

    private static final String TOKEN_EXPIRES_TIME_FORMAT = "yyyy-MM-dd'T'HH:mm:ss.SSS'Z'";

    private static final String TOKEN_EXPIRES_TIMEZONE = "UTC";

    private boolean needGetToken = true;

    private static final Logger LOGGER = LoggerFactory.getLogger(Authentication.class);

    private DLIInfo info;

    private String password;

    private String domainName;

    private String userName;

    private TokenAuthApi tokenApi;

    private RestApi restApi;

    private String authToken;

    private String projectId;

    private String serverEndPoint;

    private String iamEndPoint;

    private String serverName;

    private String accessKey;

    private String secretKey;

    private String securityToken;

    private String tokenExpiresTime;

    private AuthenticationMode authMode;

    public Authentication(AuthenticationMode mode, DLIInfo info) {
        this.info = info;
        userName = info.getUserName();
        domainName = info.getDomainName();
        password = info.getPassword();
        projectId = info.getProjectId();
        serverEndPoint = info.getServerEndPoint();
        iamEndPoint = info.getIamEndPoint();
        serverName = info.getServerName();
        accessKey = info.getAccessKey();
        secretKey = info.getSecretKey();
        securityToken = info.getSecurityToken();
        authMode = mode;
        restApi = new RestApi(new ApiClient());

        if (this.info.getToken() != null) {
            this.needGetToken = false;
            authToken = info.getToken();
        }

        if (AuthenticationMode.TOKEN.equals(authMode)) {
            buildTokenAuth();
        } else if (AuthenticationMode.AKSK.equals(authMode)) {
            buildAkskAuth();
        }

        if (info.getProxyHost() != null) {
            if (info.getProxyUser() != null) {
                setProxy(info.getProxyHost(), info.getProxyPort(), info.getProxyUser(), info.getProxyPwd());
            } else {
                setProxy(info.getProxyHost(), info.getProxyPort());
            }
        }
        initSwaggerClient();
    }

    public RestApi getRestApi() {
        return restApi;
    }

    public TokenAuthApi getTokenApi() {
        return tokenApi;
    }

    public String getAuthToken() throws DLIException {
        if (AuthenticationMode.AKSK.equals(authMode)) {
            return null;
        }

        if (!needGetToken) {
            return authToken;
        }

        if (isExpires(authToken)) {
            getToken();
        }
        return authToken;
    }

    /**
     * Check if token expires.
     * if authToken == null mean authenticationMode is AK/SK,return false;
     * if tokenExpiresTime < nowTime mean token had already expired,return true.
     *
     * @param authToken auth token
     * @return true/false
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    private boolean isExpires(String authToken) throws DLIException {
        if (null == authToken) {
            return true;
        }
        SimpleDateFormat format = new SimpleDateFormat(TOKEN_EXPIRES_TIME_FORMAT);
        format.setTimeZone(TimeZone.getTimeZone(TOKEN_EXPIRES_TIMEZONE));
        try {
            long expiresTime = format.parse(tokenExpiresTime).getTime();
            long nowTime = new Date().getTime();
            if (expiresTime < nowTime) {
                return true;
            }
        } catch (ParseException e) {
            LOGGER.error("Token expires time parse failed.", e);
            throw new DLIException("Token expires time parse failed.", e);
        }
        return false;
    }

    public String getProjectId() {
        return projectId;
    }

    public void updateAkSkAuth() {
        buildAkskAuth();
    }

    private void buildAkskAuth() {
        if (accessKey == null || accessKey.trim().isEmpty()) {
            LOGGER.error("accessKey can not be null");
            throw new IllegalArgumentException("accessKey can not be null");
        }
        if (secretKey == null || secretKey.trim().isEmpty()) {
            LOGGER.error("secretKey can not be null");
            throw new IllegalArgumentException("secretKey can not be null");
        }
        AkskAuthentication akskAuth = new AkskAuthentication(accessKey, secretKey, projectId);
        if (securityToken != null && !securityToken.isEmpty()) {
            akskAuth = new AkskAuthentication(accessKey, secretKey, securityToken, projectId);
        }
        restApi.getApiClient().addAuthentication("aksk", akskAuth);
    }

    private void buildTokenAuth() {
        tokenApi = new TokenAuthApi();
        ApiClient apiClient = new ApiClient();
        StringBuilder sb = new StringBuilder();
        if (iamEndPoint == null || iamEndPoint.trim().isEmpty()) {
            LOGGER.error("iamEndPoint can not be null");
            throw new IllegalArgumentException("iamEndPoint can not be null");
        }
        // iamEndPoint can be assign host or schema://host
        if (!iamEndPoint.startsWith("http")) {
            sb.append(PROTOCOL).append("://").append(iamEndPoint);
        } else {
            sb.append(iamEndPoint);
        }
        apiClient.setBasePath(sb.toString());
        tokenApi.setApiClient(apiClient);
    }

    public void setProxy(String proxyHost, int proxyPort) {
        Proxy proxy = new Proxy(Proxy.Type.HTTP, new InetSocketAddress(proxyHost, proxyPort));
        OkHttpClient newHttpClient = restApi.getApiClient().getHttpClient().newBuilder().proxy(proxy).build();
        restApi.getApiClient().setHttpClient(newHttpClient);

        if (AuthenticationMode.TOKEN.equals(authMode) && needGetToken) {
            OkHttpClient newHttpClient2 = tokenApi.getApiClient().getHttpClient().newBuilder().proxy(proxy).build();
            tokenApi.getApiClient().setHttpClient(newHttpClient2);
        }
    }

    public void setHttpClient(OkHttpClient httpClient) {
        restApi.getApiClient().setHttpClient(httpClient);

        if (AuthenticationMode.TOKEN.equals(authMode) && needGetToken) {
            tokenApi.getApiClient().setHttpClient(httpClient);
        }
    }

    public void setProxy(String proxyHost, int proxyPort, String username, String password) {
        java.net.Authenticator.setDefault(new ProxyAuthenticator(username, password));
        setProxy(proxyHost, proxyPort);
    }

    private void initSwaggerClient() {
        StringBuilder sb = new StringBuilder();
        if (serverEndPoint == null || serverEndPoint.trim().isEmpty()) {
            LOGGER.error("dliEndPoint can not be null");
            throw new IllegalArgumentException("dliEndPoint can not be null");
        }

        sb.append(PROTOCOL).append("://").append(serverEndPoint);
        if (info.getUrlPrefix() != null) {
            sb.append(info.getUrlPrefix());
        }
        restApi.getApiClient().setBasePath(sb.toString());
        OkHttpClient.Builder newHttpClientBuilder = restApi.getApiClient().getHttpClient().newBuilder();
        newHttpClientBuilder.connectTimeout(DEFAULT_CONNECT_TIMEOUT, TimeUnit.SECONDS);
        newHttpClientBuilder.readTimeout(DEFAULT_READ_TIMEOUT, TimeUnit.SECONDS);
        restApi.getApiClient().setHttpClient(newHttpClientBuilder.build());
    }

    private void getToken() throws DLIException {
        if (!needGetToken) {
            return;
        }

        try {
            if (userName == null || userName.trim().isEmpty()) {
                LOGGER.error("userName can not be null");
                throw new IllegalArgumentException("userName can not be null");
            }
            if (domainName == null || domainName.trim().isEmpty()) {
                LOGGER.error("domainName can not be null");
                throw new IllegalArgumentException("domainName can not be null");
            }
            if (password == null || password.trim().isEmpty()) {
                LOGGER.error("password can not be null");
                throw new IllegalArgumentException("password can not be null");
            }
            if (projectId == null || projectId.trim().isEmpty()) {
                LOGGER.error("projectId can not be null");
                throw new IllegalArgumentException("projectId can not be null");
            }
            ApiResponse<TokenAuthResponse> response =
                tokenApi.getTokenWithHttpInfo(assembleBody(userName, domainName, password, projectId));
            List<String> list = response.getHeaders().get("X-Subject-Token");
            authToken = list.get(0);
            tokenExpiresTime = response.getData().getToken().getExpiresAt();
        } catch (ApiException e) {
            LOGGER.error("msg", e);
            throw new DLIException(e.getMessage(), e.getCode(), e.getResponseBody());
        }
    }

    private static TokenAuthInfo assembleBody(String name, String domainName, String pwd, String projectId) {
        TokenAuthInfo tokenAuthInfo = new TokenAuthInfo();
        Auth auth = new Auth();
        Identity identity = new Identity();
        Password password = new Password();
        User user = new User();
        user.setName(name);
        user.setPassword(pwd);
        Domain domain = new Domain();
        domain.setName(domainName);
        user.setDomain(domain);
        password.setUser(user);
        identity.setPassword(password);
        List<String> list = new ArrayList<String>();
        list.add("password");
        identity.setMethods(list);
        auth.setIdentity(identity);
        Scope scope = new Scope();
        Project project = new Project();
        project.setId(projectId);
        scope.setProject(project);
        auth.setScope(scope);
        tokenAuthInfo.setAuth(auth);
        return tokenAuthInfo;
    }

    public void setAuthToken(String authToken) {
        this.authToken = authToken;
    }

    public void setAccessKey(String accessKey) {
        this.accessKey = accessKey;
    }

    public void setSecretKey(String secretKey) {
        this.secretKey = secretKey;
    }

    public void setSecurityToken(String securityToken) {
        this.securityToken = securityToken;
    }
}
