package com.huawei.dli.sdk.common;

import com.huawei.dli.sdk.common.type.DliArray;
import com.huawei.dli.sdk.common.type.DliMap;
import com.huawei.dli.sdk.common.type.DliStruct;
import com.huawei.dli.sdk.exception.DLIException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.Timestamp;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

public class Row {
    private static final Logger LOGGER = LoggerFactory.getLogger(Row.class);

    private static final ThreadLocal<DateFormat> DATE_FORMAT =
        ThreadLocal.withInitial(() -> new SimpleDateFormat("yyyy-MM-dd"));

    private static final ThreadLocal<DateFormat> TIMESTAMP_FORMAT =
        ThreadLocal.withInitial(() -> new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSSZZ"));

    public static final String NULL_VALUE_FOR_CSV = "\b\u00A0\uFEFF"; // \b is same as \u0008, backspace

    private final TableSchema schema;

    private List<Object> cols = new ArrayList<>();

    public Row(TableSchema schema) {
        this.schema = schema;
        for (int i = 0; i < schema.getColumns().size(); i++) {
            cols.add(null);
        }
    }

    public TableSchema getSchema() {
        return schema;
    }

    public List<Object> getCols() {
        return cols;
    }

    public void setCols(List<Object> cols) {
        this.cols = cols;
    }

    public void setBoolean(int index, Boolean data) throws DLIException {
        setColData(index, data, DataType.BOOLEAN);
    }

    public void setTinyint(int index, Byte tinyintData) throws DLIException {
        setColData(index, tinyintData, DataType.TINYINT);
    }

    // SMALLINT or SHORT
    public void setShort(int index, Short data) throws DLIException {
        int num = schema.getColumns().size();
        if (index > num - 1) {
            LOGGER.error("Columns length is " + num + " cols,index is greater than the length of columns");
            throw new DLIException("Columns length is " + num + " cols,index is greater than the length of columns");
        }
        Column column = schema.getColumn(index);
        if (!DataType.SHORT.equals(column.getType()) && !DataType.SMALLINT.equals(column.getType())) {
            LOGGER.error("Data type mismatch," + index + " col type is " + column.getType());
            throw new DLIException("Data type mismatch," + index + " col type is " + column.getType());
        }
        cols.set(index, data);
    }

    public void setInt(int index, Integer data) throws DLIException {
        setColData(index, data, DataType.INT);
    }

    // not BIGINT？
    public void setBigInteger(int index, BigInteger data) throws DLIException {
        setColData(index, data, DataType.BIGINT);
    }

    // BIGINT or LONG
    public void setLong(int index, Long data) throws DLIException {
        int num = schema.getColumns().size();
        if (index > num - 1) {
            LOGGER.error("Columns length is " + num + " cols,index is greater than the length of columns");
            throw new DLIException("Columns length is " + num + " cols,index is greater than the length of columns");
        }
        Column column = schema.getColumn(index);
        if (!DataType.LONG.equals(column.getType()) && !DataType.BIGINT.equals(column.getType())) {
            LOGGER.error("Data type mismatch," + index + " col type is " + column.getType());
            throw new DLIException("Data type mismatch," + index + " col type is " + column.getType());
        }
        cols.set(index, data);
    }

    public void setFloat(int index, Float data) throws DLIException {
        setColData(index, data, DataType.FLOAT);
    }

    public void setDouble(int index, Double data) throws DLIException {
        setColData(index, data, DataType.DOUBLE);
    }

    // DECIMAL(p, s)
    public void setBigDecimal(int index, BigDecimal data) throws DLIException {
        setColData(index, data, DataType.DECIMAL);
    }

    public void setString(int index, String data) throws DLIException {
        setColData(index, data, DataType.STRING);
    }

    public void setDate(int index, Date data) throws DLIException {
        setColData(index, data, DataType.DATE);
    }

    public void setTimestamp(int index, Timestamp data) throws DLIException {
        setColData(index, data, DataType.TIMESTAMP);
    }

    public void setStruct(int index, DliStruct data) throws DLIException {
        setColData(index, data, DataType.STRUCT);
    }

    public void setArray(int index, DliArray data) throws DLIException {
        setColData(index, data, DataType.ARRAY);
    }

    public void setMap(int index, DliMap data) throws DLIException {
        setColData(index, data, DataType.MAP);
    }

    private void setColData(int index, Object data, DataType dataType) throws DLIException {
        int num = schema.getColumns().size();
        if (index > num - 1) {
            LOGGER.error("Columns length is " + num + " cols,index is greater than the length of columns");
            throw new DLIException("Columns length is " + num + " cols,index is greater than the length of columns");
        }
        Column column = schema.getColumn(index);
        if (!dataType.equals(column.getType())) {
            LOGGER.error("Data type mismatch," + index + " col type is " + column.getType());
            throw new DLIException("Data type mismatch," + index + " col type is " + column.getType());
        }
        cols.set(index, data);
    }

    public Boolean getBoolean(int columnIdx) throws DLIException {
        DataType dataType = schema.getColumn(columnIdx).getType();
        if (DataType.BOOLEAN.getValue() != dataType.getValue()) {
            LOGGER.error("Data type mismatch," + columnIdx + " col type is " + dataType);
            throw new DLIException("Data type mismatch," + columnIdx + " col type is " + dataType);
        }
        return (Boolean) cols.get(columnIdx);
    }

    public Byte getTinyint(int columnIdx) throws DLIException {
        DataType dataType = schema.getColumn(columnIdx).getType();
        if (DataType.TINYINT.getValue() != dataType.getValue()) {
            LOGGER.error("Data type mismatch," + columnIdx + " col type is " + dataType);
            throw new DLIException("Data type mismatch," + columnIdx + " col type is " + dataType);
        }
        return (Byte) cols.get(columnIdx);
    }

    // SMALLINT or SHORT
    public Short getShort(int columnIdx) throws DLIException {
        DataType dataType = schema.getColumn(columnIdx).getType();
        if (DataType.SHORT.getValue() != dataType.getValue()
            && DataType.SMALLINT.getValue() != dataType.getValue()) {
            LOGGER.error("Data type mismatch," + columnIdx + " col type is " + dataType);
            throw new DLIException("Data type mismatch," + columnIdx + " col type is " + dataType);
        }
        return (Short) cols.get(columnIdx);
    }

    public Integer getInt(int columnIdx) throws DLIException {
        DataType dataType = schema.getColumn(columnIdx).getType();
        if (DataType.INT.getValue() != dataType.getValue()) {
            LOGGER.error("Data type mismatch," + columnIdx + " col type is " + dataType);
            throw new DLIException("Data type mismatch," + columnIdx + " col type is " + dataType);
        }
        return (Integer) cols.get(columnIdx);
    }

    // not BIGINT?
    public BigInteger getBigInteger(int columnIdx) throws DLIException {
        DataType dataType = schema.getColumn(columnIdx).getType();
        if (DataType.BIGINT.getValue() != dataType.getValue()
            && DataType.LONG.getValue() != dataType.getValue()) {
            LOGGER.error("Data type mismatch," + columnIdx + " col type is " + dataType);
            throw new DLIException("Data type mismatch," + columnIdx + " col type is " + dataType);
        }
        if (cols.get(columnIdx) == null) {
            return null;
        }
        String bigIntegerStr = String.valueOf(cols.get(columnIdx));
        return new BigInteger(bigIntegerStr);
    }

    // BIGINT or LONG
    public Long getLong(int columnIdx) throws DLIException {
        DataType dataType = schema.getColumn(columnIdx).getType();
        if (DataType.LONG.getValue() != dataType.getValue()
            && DataType.BIGINT.getValue() != dataType.getValue()) {
            LOGGER.error("Data type mismatch," + columnIdx + " col type is " + dataType);
            throw new DLIException("Data type mismatch," + columnIdx + " col type is " + dataType);
        }
        if (cols.get(columnIdx) == null) {
            return null;
        }
        String longStr = String.valueOf(cols.get(columnIdx));
        return Long.parseLong(longStr);
    }

    public Float getFloat(int columnIdx) throws DLIException {
        DataType dataType = schema.getColumn(columnIdx).getType();
        if (DataType.FLOAT.getValue() != dataType.getValue()) {
            LOGGER.error("Data type mismatch," + columnIdx + " col type is " + dataType);
            throw new DLIException("Data type mismatch," + columnIdx + " col type is " + dataType);
        }
        return (Float) cols.get(columnIdx);
    }

    public Double getDouble(int columnIdx) throws DLIException {
        DataType dataType = schema.getColumn(columnIdx).getType();
        if (DataType.DOUBLE.getValue() != dataType.getValue()) {
            LOGGER.error("Data type mismatch," + columnIdx + " col type is " + dataType);
            throw new DLIException("Data type mismatch," + columnIdx + " col type is " + dataType);
        }
        return (Double) cols.get(columnIdx);
    }

    // DECIMAL(p, s)
    public BigDecimal getBigDecimal(int columnIdx) throws DLIException {
        DataType dataType = schema.getColumn(columnIdx).getType();
        if (DataType.DECIMAL.getValue() != dataType.getValue()) {
            LOGGER.error("Data type mismatch," + columnIdx + " col type is " + dataType);
            throw new DLIException("Data type mismatch," + columnIdx + " col type is " + dataType);
        }
        return (BigDecimal) cols.get(columnIdx);
    }

    public String getString(int columnIdx) throws DLIException {
        DataType dataType = schema.getColumn(columnIdx).getType();
        if (DataType.STRING.getValue() != dataType.getValue()) {
            LOGGER.error("Data type mismatch," + columnIdx + " col type is " + dataType);
            throw new DLIException("Data type mismatch," + columnIdx + " col type is " + dataType);
        }
        return (String) cols.get(columnIdx);
    }

    public Date getDate(int columnIdx) throws DLIException {
        DataType dataType = schema.getColumn(columnIdx).getType();
        if (DataType.DATE.getValue() != dataType.getValue()) {
            LOGGER.error("Data type mismatch," + columnIdx + " col type is " + dataType);
            throw new DLIException("Data type mismatch," + columnIdx + " col type is " + dataType);
        }
        return (Date) cols.get(columnIdx);
    }

    public Timestamp getTimestamp(int columnIdx) throws DLIException {
        DataType dataType = schema.getColumn(columnIdx).getType();
        if (DataType.TIMESTAMP.getValue() != dataType.getValue()) {
            LOGGER.error("Data type mismatch," + columnIdx + " col type is " + dataType);
            throw new DLIException("Data type mismatch," + columnIdx + " col type is " + dataType);
        }
        return (Timestamp) cols.get(columnIdx);
    }

    public DliStruct getStruct(int columnIdx) throws DLIException {
        DataType dataType = schema.getColumn(columnIdx).getType();
        if (DataType.STRUCT.getValue() != dataType.getValue()) {
            LOGGER.error("Data type mismatch," + columnIdx + " col type is " + dataType);
            throw new DLIException("Data type mismatch," + columnIdx + " col type is " + dataType);
        }
        return (DliStruct) cols.get(columnIdx);
    }

    public DliArray getArray(int columnIdx) throws DLIException {
        DataType dataType = schema.getColumn(columnIdx).getType();
        if (DataType.ARRAY.getValue() != dataType.getValue()) {
            LOGGER.error("Data type mismatch," + columnIdx + " col type is " + dataType);
            throw new DLIException("Data type mismatch," + columnIdx + " col type is " + dataType);
        }
        return (DliArray) cols.get(columnIdx);
    }

    public DliMap getMap(int columnIdx) throws DLIException {
        DataType dataType = schema.getColumn(columnIdx).getType();
        if (DataType.MAP.getValue() != dataType.getValue()) {
            LOGGER.error("Data type mismatch," + columnIdx + " col type is " + dataType);
            throw new DLIException("Data type mismatch," + columnIdx + " col type is " + dataType);
        }
        return (DliMap) cols.get(columnIdx);
    }

    public String getRowString() throws DLIException {
        int size = schema.getColumns().size();
        StringBuffer sb = new StringBuffer();
        for (int index = 0; index < size; index++) {
            DataType type = schema.getColumn(index).getType();
            String str = getDataToString(type, index);
            str = str == null ? "" : str;
            if (index != size - 1) {
                sb.append(str).append(",");
            } else {
                sb.append(str).append("\n");
            }
        }
        return sb.toString();
    }

    public String getCsvString() throws DLIException {
        int size = schema.getColumns().size();
        StringBuffer sb = new StringBuffer();
        for (int index = 0; index < size; index++) {
            DataType type = schema.getColumn(index).getType();
            String str = getDataToString(type, index);

            if (str != null) {
                if (type.getValue() == DataType.STRING.getValue()) {
                    sb.append("\"").append(str).append("\"");
                } else {
                    sb.append(str);
                }
            }

            if (index != size - 1) {
                sb.append(",");
            } else {
                sb.append("\n");
            }
        }
        return sb.toString();
    }

    private String getDataToString(DataType type, int index) throws DLIException {
        switch (type) {
            case BOOLEAN:
                Boolean dataBoolean = getBoolean(index);
                return dataBoolean == null ? null : String.valueOf(dataBoolean);
            case TINYINT:
                Byte tinyintData = getTinyint(index);
                return tinyintData == null ? null : String.valueOf(tinyintData);
            case SMALLINT:
            case SHORT:
                Short dataSmallInt = getShort(index);
                return dataSmallInt == null ? null : String.valueOf(dataSmallInt);
            case INT:
                Integer dataInt = getInt(index);
                return dataInt == null ? null : String.valueOf(dataInt);
            case LONG:
                Long dataLong = getLong(index);
                return dataLong == null ? null : String.valueOf(dataLong);
            case BIGINT:
                BigInteger dataBigInt = getBigInteger(index);
                return dataBigInt == null ? null : String.valueOf(dataBigInt);
            case FLOAT:
                Float dataFloat = getFloat(index);
                return dataFloat == null ? null : String.valueOf(dataFloat);
            case DOUBLE:
                Double dataDouble = getDouble(index);
                return dataDouble == null ? null : String.valueOf(dataDouble);
            case DECIMAL:
                BigDecimal dataDecimal = getBigDecimal(index);
                return dataDecimal == null ? null : String.valueOf(dataDecimal);
            case STRING:
                String dataStr = getString(index);
                return dataStr == null ? null : dataStr
                    .replaceAll("\\\\", "\\\\\\\\")
                    .replaceAll("\"", "\\\\\"");
            case DATE:
                Date date = getDate(index);
                if (date == null) {
                    return null;
                }
                return DATE_FORMAT.get().format(date);
            case TIMESTAMP:
                Timestamp timestamp = getTimestamp(index);
                if (timestamp == null) {
                    return null;
                }
                return TIMESTAMP_FORMAT.get().format(timestamp);
            case STRUCT:
                DliStruct structData = getStruct(index);
                return structData == null ? null : structData.toString();
            case ARRAY:
                DliArray arrayData = getArray(index);
                return arrayData == null ? null : arrayData.toString();
            case MAP:
                DliMap mapData = getMap(index);
                return mapData == null ? null : mapData.toString();
            default:
                return null;
        }
    }
}
