package com.huawei.dli.sdk.common.type;

import com.huawei.dli.sdk.common.DataType;
import com.huawei.dli.sdk.common.DataTypeInfo;
import com.huawei.dli.sdk.common.DecimalTypeInfo;
import com.huawei.dli.sdk.common.SimplePrimitiveTypeInfo;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Locale;

public class TypeInfoParser {
    private static final Logger LOGGER = LoggerFactory.getLogger(TypeInfoParser.class);

    private ArrayList<String> tokens = new ArrayList<>();

    private String typeInfoName;

    private int index = 0;

    TypeInfoParser(String name) {
        this.typeInfoName = name.toUpperCase(Locale.US);
        this.tokenize(this.typeInfoName);
    }

    public static DataTypeInfo getTypeInfoFromTypeString(String name) {
        return new TypeInfoParser(name).parseTypeInfo();
    }

    private boolean isTypeInfoChar(char c) {
        return Character.isLetterOrDigit(c) || c == '_' || c == '.';
    }

    private void tokenize(String name) {
        if (name.startsWith("\"")) {
            name = name.substring(1);
        }
        if (name.endsWith("\"")) {
            name = name.substring(0, name.length() - 1);
        }
        int begin = 0;
        for (int end = 1; end <= name.length(); ++end) {
            if (end == name.length() || !this.isTypeInfoChar(name.charAt(end))
                || !this.isTypeInfoChar(name.charAt(end - 1))) {
                String token = name.substring(begin, end).trim();
                if (!token.isEmpty()) {
                    this.tokens.add(token);
                }
                begin = end;
            }
        }
    }

    private String peek() {
        if (this.index >= this.tokens.size()) {
            LOGGER.error("Parse type info failed, pls check: " + this.typeInfoName);
            throw new IllegalArgumentException("Parse type info failed, pls check: " + this.typeInfoName);
        } else {
            return (String) this.tokens.get(this.index++);
        }
    }

    private void expect(String expected) {
        if (!this.peek().equals(expected)) {
            LOGGER.error("Error parse type info: " + this.typeInfoName
                + ", expect '" + expected + "' in position: " + this.getPosition(this.index - 1));
            throw new IllegalArgumentException("Error parse type info: " + this.typeInfoName
                + ", expect '" + expected + "' in position: " + this.getPosition(this.index - 1));
        }
    }

    private int getPosition(int tokenIndex) {
        int pos = 0;
        for (int i = 0; i < tokenIndex; ++i) {
            pos += ((String) this.tokens.get(i)).length();
        }
        return pos;
    }

    private DataTypeInfo parseTypeInfo() {
        DataTypeInfo type = this.parseTypeInfoInternal();
        if (DataType.UNKNOWN.equals(type.getDataType())) {
            LOGGER.info("Parse type info failed, return UNKNOWN and not throw ex");
            return type;
        }
        if (this.index != this.tokens.size()) {
            LOGGER.error("Parse type info failed, pls check: " + this.typeInfoName);
            throw new IllegalArgumentException("Parse type info failed, pls check: " + this.typeInfoName);
        } else {
            return type;
        }
    }

    private int getInteger() {
        try {
            return Integer.parseInt(this.peek());
        } catch (NumberFormatException var2) {
            LOGGER.error("Error parse type info: " + this.typeInfoName
                + ", expect integer in position: " + this.getPosition(this.index - 1), var2);
            throw new IllegalArgumentException("Error parse type info: " + this.typeInfoName
                + ", expect integer in position: " + this.getPosition(this.index - 1), var2);
        }
    }

    private int[] getDecimalParams() {
        if (this.index < this.tokens.size() && ((String) this.tokens.get(this.index)).equals("(")) {
            this.expect("(");
            int precision = this.getInteger();
            this.expect(",");
            int scale = this.getInteger();
            this.expect(")");
            return new int[] {precision, scale};
        } else {
            return new int[0];
        }
    }

    private DataTypeInfo parseTypeInfoInternal() {
        DataType typeCategory = null;
        try {
            typeCategory = DataType.valueOf(this.peek());
        } catch (IllegalArgumentException e) {
            typeCategory = DataType.UNKNOWN;
        }
        switch (typeCategory) {
            case DECIMAL:
                return this.parseDecimalTypeInfo();
            case STRUCT:
                return this.parseStructTypeInfo();
            case ARRAY:
                return this.parseArrayTypeInfo();
            case MAP:
                return this.parseMapTypeInfo();
            default:
                return new SimplePrimitiveTypeInfo(typeCategory);
        }
    }

    private DataTypeInfo parseDecimalTypeInfo() {
        int[] params = this.getDecimalParams();
        return params.length == 0 ? new DecimalTypeInfo() : new DecimalTypeInfo(params[0], params[1]);
    }

    private DataTypeInfo parseArrayTypeInfo() {
        expect("<");
        DataTypeInfo elementTypeInfo = parseTypeInfoInternal();
        expect(">");
        return new SimpleArrayTypeInfo(elementTypeInfo);
    }

    private DataTypeInfo parseMapTypeInfo() {
        expect("<");
        DataTypeInfo keyTypeInfo = parseTypeInfoInternal();
        expect(",");
        DataTypeInfo valueTypeInfo = parseTypeInfoInternal();
        expect(">");
        return new SimpleMapTypeInfo(keyTypeInfo, valueTypeInfo);
    }

    private DataTypeInfo parseStructTypeInfo() {
        ArrayList<String> names = new ArrayList<String>();
        ArrayList<DataTypeInfo> typeInfos = new ArrayList<DataTypeInfo>();
        boolean first = true;
        while (true) {
            if (first) {
                expect("<");
                first = false;
            } else {
                String delimiter = peek();
                if (delimiter.equals(">")) {
                    break;
                } else {
                    index--;
                    expect(",");
                }
            }
            names.add(peek());
            expect(":");
            typeInfos.add(parseTypeInfoInternal());
        }
        return new SimpleStructTypeInfo(names, typeInfos);
    }
}
