package com.huawei.dli.sdk.common.type;

import com.huawei.dli.sdk.common.DataType;
import com.huawei.dli.sdk.common.DataTypeInfo;
import com.huawei.dli.sdk.common.DecimalTypeInfo;
import com.huawei.dli.sdk.common.SimplePrimitiveTypeInfo;

import java.util.HashMap;
import java.util.List;

public class TypeInfoUtils {
    public static final DataTypeInfo INT = new SimplePrimitiveTypeInfo(DataType.INT);

    public static final DataTypeInfo SMALLINT = new SimplePrimitiveTypeInfo(DataType.SMALLINT);

    public static final DataTypeInfo BIGINT = new SimplePrimitiveTypeInfo(DataType.BIGINT);

    public static final DataTypeInfo LONG = new SimplePrimitiveTypeInfo(DataType.LONG);

    public static final DataTypeInfo SHORT = new SimplePrimitiveTypeInfo(DataType.SHORT);

    public static final DataTypeInfo DOUBLE = new SimplePrimitiveTypeInfo(DataType.DOUBLE);

    public static final DataTypeInfo FLOAT = new SimplePrimitiveTypeInfo(DataType.FLOAT);

    public static final DecimalTypeInfo DECIMAL = new DecimalTypeInfo();

    public static final DataTypeInfo BOOLEAN = new SimplePrimitiveTypeInfo(DataType.BOOLEAN);

    public static final DataTypeInfo STRING = new SimplePrimitiveTypeInfo(DataType.STRING);

    public static final DataTypeInfo DATE = new SimplePrimitiveTypeInfo(DataType.DATE);

    public static final DataTypeInfo TIMESTAMP = new SimplePrimitiveTypeInfo(DataType.TIMESTAMP);

    public static final DataTypeInfo TINYINT = new SimplePrimitiveTypeInfo(DataType.TINYINT);

    private static HashMap<DataType, DataTypeInfo> typeInfoMap = new HashMap<DataType, DataTypeInfo>();

    static {
        typeInfoMap.put(INT.getDataType(), INT);
        typeInfoMap.put(SMALLINT.getDataType(), SMALLINT);
        typeInfoMap.put(BIGINT.getDataType(), BIGINT);
        typeInfoMap.put(LONG.getDataType(), LONG);
        typeInfoMap.put(SHORT.getDataType(), SHORT);
        typeInfoMap.put(DOUBLE.getDataType(), DOUBLE);
        typeInfoMap.put(FLOAT.getDataType(), FLOAT);
        typeInfoMap.put(DECIMAL.getDataType(), DECIMAL);
        typeInfoMap.put(BOOLEAN.getDataType(), BOOLEAN);
        typeInfoMap.put(STRING.getDataType(), STRING);
        typeInfoMap.put(DATE.getDataType(), DATE);
        typeInfoMap.put(TIMESTAMP.getDataType(), TIMESTAMP);
        typeInfoMap.put(TINYINT.getDataType(), TINYINT);
    }

    private TypeInfoUtils() {
    }

    /**
     * Get the data type info form a type string, e.g decimal(10, 0), struct&lt;age:int, name:string&gt;
     *
     * @param typeString The data type info in String format
     * @return DataTypeInfo
     */
    public static DataTypeInfo getDataTypeInfo(String typeString) {
        return TypeInfoParser.getTypeInfoFromTypeString(typeString);
    }

    /**
     * Get the data type info of a Primitive Type.
     *
     * @param dataType A DataType
     * @return DataTypeInfo
     */
    public static DataTypeInfo getPrimitiveTypeInfo(DataType dataType) {
        DataTypeInfo typeInfo = typeInfoMap.get(dataType);
        if (typeInfo != null) {
            return typeInfo;
        } else {
            // Not found in the cache. Must be parameterized types.
            throw new IllegalArgumentException("Error get PrimitiveTypeInfo instance for: " + dataType);
        }
    }

    /**
     * Get the data type info of a Decimal.
     *
     * @param precision The precision of the Decimal
     * @param scale     The scale of the Decimal
     * @return DecimalTypeInfo
     */
    public static DecimalTypeInfo getDecimalTypeInfo(int precision, int scale) {
        if ((precision == DecimalTypeInfo.DEFAULT_PRECISION) && (scale == DecimalTypeInfo.DEFAULT_SCALE)) {
            return DECIMAL;
        }
        return new DecimalTypeInfo(precision, scale);
    }

    /**
     * Get the data type info of a Struct Type.
     *
     * @param fieldNames     The list of the field names of the Struct Type
     * @param fieldTypeInfos The list of the field data type info ot the Struct type
     * @return StructTypeInfo
     */
    public static StructTypeInfo getStructTypeInfo(List<String> fieldNames, List<DataTypeInfo> fieldTypeInfos) {
        return new SimpleStructTypeInfo(fieldNames, fieldTypeInfos);
    }

    /**
     * Get the data type info of a Array Type.
     *
     * @param elementTypeInfo The element data type info of the Array Type
     * @return ArrayTypeInfo
     */
    public static ArrayTypeInfo getArrayTypeInfo(DataTypeInfo elementTypeInfo) {
        return new SimpleArrayTypeInfo(elementTypeInfo);
    }

    /**
     * Get the data type info of a Map Type.
     *
     * @param keyTypeInfo   The key type info of the Map Type
     * @param valueTypeInfo The value type info of the Map Type
     * @return MapTypeInfo
     */
    public static MapTypeInfo getMapTypeInfo(DataTypeInfo keyTypeInfo, DataTypeInfo valueTypeInfo) {
        return new SimpleMapTypeInfo(keyTypeInfo, valueTypeInfo);
    }
}
