package com.huawei.dli.sdk.common.type;

import com.huawei.dli.sdk.common.DataType;
import com.huawei.dli.sdk.common.DataTypeInfo;
import com.huawei.dli.sdk.exception.DLIException;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.Timestamp;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

public class TypeUtils {
    private static final String DATA_FORMAT = "yyyy-MM-dd";

    private static Map<DataType, Class> dataTypeMapper = new HashMap<DataType, Class>();

    static {
        dataTypeMapper.put(DataType.INT, Integer.class);
        dataTypeMapper.put(DataType.SMALLINT, Short.class);
        dataTypeMapper.put(DataType.BIGINT, BigInteger.class);
        dataTypeMapper.put(DataType.LONG, Long.class);
        dataTypeMapper.put(DataType.SHORT, Short.class);
        dataTypeMapper.put(DataType.DOUBLE, Double.class);
        dataTypeMapper.put(DataType.DECIMAL, BigDecimal.class);
        dataTypeMapper.put(DataType.BOOLEAN, Boolean.class);
        dataTypeMapper.put(DataType.STRING, String.class);
        dataTypeMapper.put(DataType.DATE, Date.class);
        dataTypeMapper.put(DataType.TIMESTAMP, Timestamp.class);
        dataTypeMapper.put(DataType.TINYINT, Byte.class);
        dataTypeMapper.put(DataType.STRUCT, DliStruct.class);
        dataTypeMapper.put(DataType.ARRAY, DliArray.class);
        dataTypeMapper.put(DataType.MAP, DliMap.class);
    }

    public static Class dliTypeToJavaType(DataType type) {
        if (dataTypeMapper.containsKey(type)) {
            return dataTypeMapper.get(type);
        }
        throw new IllegalArgumentException("Cannot get Java type for dli type: " + type);
    }

    /**
     * Make a DliStruct object form the JSON string
     *
     * @param typeInfo The data type info of the DliStruct object
     * @param data     The DliStruct data value in JSON string format
     * @return The object of DliStruct
     */
    public static DliStruct makeStruct(StructTypeInfo typeInfo, String data) {
        return new SimpleStruct(typeInfo, data);
    }

    /**
     * Make a DliArray object form the JSON string
     *
     * @param typeInfo The data type info of the DliArray object
     * @param data     The DliArray data value in JSON string format
     * @return The object of DliArray
     */
    public static DliArray makeArray(ArrayTypeInfo typeInfo, String data) {
        return new SimpleArray(typeInfo, data);
    }

    /**
     * Make a DliMap object form the JSON string
     *
     * @param typeInfo The data type info of the DliMap object
     * @param data     The DliMap data value in JSON string format
     * @return The object of DliMap
     */
    public static DliMap makeMap(MapTypeInfo typeInfo, String data) {
        return new SimpleMap(typeInfo, data);
    }

    /**
     * Transform the data value from string to real data type in java
     *
     * @param data          The data value in string type
     * @param valueTypeInfo The type info of the data
     * @return The data object in real data type
     * @throws DLIException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public static Object transform(String data, DataTypeInfo valueTypeInfo) throws DLIException {
        if (data == null) {
            throw new DLIException("The value is null, can not transform");
        }
        switch (valueTypeInfo.getDataType()) {
            case INT:
                return Integer.parseInt(data);
            case BIGINT:
                return new BigInteger(data);
            case LONG:
                return Long.parseLong(data);
            case SMALLINT:
            case SHORT:
                return Short.parseShort(data);
            case DOUBLE:
                return Double.parseDouble(data);
            case FLOAT:
                return Float.parseFloat(data);
            case STRING:
                return data;
            case BOOLEAN:
                return Boolean.parseBoolean(data);
            case DECIMAL:
                return new BigDecimal(data);
            case DATE:
                SimpleDateFormat format = new SimpleDateFormat(DATA_FORMAT);
                try {
                    return format.parse(data);
                } catch (ParseException e) {
                    throw new DLIException(e.getMessage());
                }
            case TIMESTAMP:
                return new Timestamp(Long.parseLong(data));
            case TINYINT:
                return Byte.parseByte(data);
            case STRUCT:
                return TypeUtils.makeStruct((StructTypeInfo) valueTypeInfo, data);
            case ARRAY:
                return TypeUtils.makeArray((ArrayTypeInfo) valueTypeInfo, data);
            case MAP:
                return TypeUtils.makeMap((MapTypeInfo) valueTypeInfo, data);
            default:
                return data;
        }
    }
}
