package com.huawei.dli.sdk.common.utils;

import java.sql.Timestamp;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.util.Date;
import java.util.TimeZone;

import static java.util.concurrent.TimeUnit.MILLISECONDS;

/**
 * TIPS: This is copy from Spark, details see [[org.apache.spark.sql.catalyst.util.DateTimeUtils]]
 */
public class DateTimeUtils {
    private static final long MICROS_PER_MILLIS = 1000L;

    private static final long MILLIS_PER_SECOND = 1000;

    private static final long MICROS_PER_SECOND = MICROS_PER_MILLIS * MILLIS_PER_SECOND;

    private static final long SECONDS_PER_DAY = 24 * 60 * 60;

    private static final long MILLIS_PER_DAY = SECONDS_PER_DAY * MILLIS_PER_SECOND;

    public static TimeZone defaultTimeZone() {
        return TimeZone.getDefault();
    }

    /**
     * Returns the number of days since epoch from java.util.Date.
     * @param date java.util.Date
     * @return the number of days since epoch
     */
    public static int fromJavaDate(Date date) {
        return millisToDays(date.getTime());
    }

    public static int millisToDays(long millisUtc) {
        return millisToDays(millisUtc, defaultTimeZone());
    }

    public static int millisToDays(long millisUtc, TimeZone timeZone) {
        long millisLocal = millisUtc + timeZone.getOffset(millisUtc);
        return (int) Math.floor((double) millisLocal / MILLIS_PER_DAY);
    }

    /**
     * Returns a java.util.Date from the number of days since epoch.
     * @param daysSinceEpoch the number of days since epoch
     * @return java.util.Date
     */
    public static Date toJavaDate(int daysSinceEpoch) {
        return new Date(daysToMills(daysSinceEpoch));
    }

    public static long daysToMills(int days) {
        return daysToMills(days, defaultTimeZone());
    }

    public static long daysToMills(int days, TimeZone timeZone) {
        long millisLocal = (long) days * MILLIS_PER_DAY;
        return millisLocal - getOffsetFromLocalMills(millisLocal, timeZone);
    }

    private static long getOffsetFromLocalMills(long millisLocal, TimeZone tz) {
        int guess = tz.getRawOffset();
        int offset = tz.getOffset(millisLocal - guess);
        if (offset != guess) {
            guess = tz.getOffset(millisLocal - offset);
            if (guess != offset) {
                LocalDate localDate = LocalDate.ofEpochDay(MILLISECONDS.toDays(millisLocal));
                LocalTime localTime = LocalTime.ofNanoOfDay(MILLISECONDS.toNanos(
                    Math.floorMod(millisLocal, MILLIS_PER_DAY)));
                LocalDateTime localDateTime = LocalDateTime.of(localDate, localTime);
                long millsEpoch = localDateTime.atZone(tz.toZoneId()).toInstant().toEpochMilli();

                guess = (int) (millisLocal - millsEpoch);
            }
        }
        return guess;
    }

    /**
     * Returns the number of micros since epoch from java.sql.Timestamp.
     * @param ts java.sql.Timestamp
     * @return the number of micros since epoch
     */
    public static long fromJavaTimestamp(Timestamp ts) {
        if (ts != null) {
            return ts.getTime() * 1000L + ((long) ts.getNanos() / 1000) % 1000L;
        } else {
            return 0L;
        }
    }

    /**
     * Returns a java.sql.Timestamp from number of micros since epoch.
     * @param us the number of micros since epoch
     * @return java.sql.Timestamp
     */
    public static Timestamp toJavaTimestamp(long us) {
        // setNanos() will overwrite the millisecond part, so the milliseconds should be
        // cut off at seconds
        long seconds = us / MICROS_PER_SECOND;
        long micros = us % MICROS_PER_SECOND;
        // setNanos() can not accept negative value
        if (micros < 0) {
            micros += MICROS_PER_SECOND;
            seconds -= 1;
        }
        Timestamp t = new Timestamp(seconds * 1000);
        t.setNanos((int) micros * 1000);
        return t;
    }
}
