package com.huawei.dli.sdk.common.utils;

import com.obs.services.ObsClient;
import com.obs.services.ObsConfiguration;
import com.obs.services.SecretFlexibleObsClient;
import com.obs.services.exception.ObsException;
import com.obs.services.model.GetObjectRequest;
import com.obs.services.model.ListObjectsRequest;
import com.obs.services.model.ObjectListing;
import com.obs.services.model.ObsObject;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

public class ObsClientUtils {
    private static final Logger logger = LoggerFactory.getLogger(ObsClientUtils.class);

    private static final int OBS_SOCKET_TIMEOUT = 3_000_000; // unit is ms

    private static final int OBS_CONNECTION_TIMEOUT = 1_000_000; // unit is ms

    private static final Object ACTIVE_CLIENT_LOCK = new Object();

    private final String endpoint;

    private String ak;

    private String sk;

    private ObsClient obsClient;

    private volatile SecretFlexibleObsClient secObsClient;

    public ObsClientUtils(String endpoint) {
        this.endpoint = endpoint;
    }

    public ObsClientUtils(String endpoint, String ak, String sk) {
        this.endpoint = endpoint;
        this.ak = ak;
        this.sk = sk;
    }

    private SecretFlexibleObsClient getOrCreateSecClient() {
        if (secObsClient == null) {
            synchronized (ACTIVE_CLIENT_LOCK) {
                if (secObsClient == null) {
                    secObsClient = new SecretFlexibleObsClient(initConf());
                }
            }
        }
        return secObsClient;
    }

    public ObsClient getObsClient() {
        if (this.obsClient == null) {
            this.obsClient = new ObsClient(this.ak, this.sk, initConf());
        }
        return this.obsClient;
    }

    private ObsConfiguration initConf() {
        ObsConfiguration config = new ObsConfiguration();
        config.setSocketTimeout(OBS_SOCKET_TIMEOUT);
        config.setConnectionTimeout(OBS_CONNECTION_TIMEOUT);
        config.setEndPoint(this.endpoint);
        return config;
    }

    public void closeObsClient() {
        if (this.obsClient != null) {
            try {
                this.obsClient.close();
            } catch (IOException e) {
                logger.error("close obs client failed", e);
            } finally {
                this.obsClient = null;
            }
        }
    }

    public List<ObsObject> listObjects(String ak, String sk, String bucketName, String objPrefix) {
        ListObjectsRequest listObjReq = new ListObjectsRequest();
        listObjReq.setBucketName(bucketName);
        listObjReq.setPrefix(objPrefix);
        listObjReq.setMaxKeys(1000);
        ObjectListing objListing;
        List<ObsObject> obsObjList = new ArrayList<>();
        long startTime = System.currentTimeMillis();
        do {
            objListing = getOrCreateSecClient().listObjects(listObjReq, ak, sk);
            obsObjList.addAll(objListing.getObjects().stream()
                .filter(object -> object.getMetadata().getContentLength() > 0)
                .collect(Collectors.toList()));
            listObjReq.setMarker(objListing.getNextMarker());
        } while (objListing.isTruncated());
        long endTime = System.currentTimeMillis();
        logger.info("get objects total cost: {}ms", endTime - startTime);

        return obsObjList;
    }

    public InputStream getObjectStream(String ak, String sk, String bucketName, String objKey) {
        GetObjectRequest getObjReq = new GetObjectRequest(bucketName, objKey);
        try {
            return getOrCreateSecClient().getObject(getObjReq, ak, sk).getObjectContent();
        } catch (ObsException e) {
            if (e.getResponseCode() == 404) {
                logger.warn("The obs object({}) to get is not fount", objKey);
                return null;
            } else {
                throw new RuntimeException("Failed to get obs object(" + objKey + ") content", e);
            }
        }
    }
}
